/*
   This file is part of the MoCA video library.
   Name: ActionDetectionTest.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/

// C++ includes
#include <string>
#include <iostream>
using namespace std;

// MoCA includes
#include <ActionDetection.h>
#include <IImageStream.h>
#include <IImageStream.h>
#include <FileList.h>
#include <MotionVector.h>
#include <VideoDecoder.h>

void usage()
{
	cerr << "Usage: ActionDetection [options] filename|path" << endl;
	cerr << ACTION_DETECTION_VERSION << endl;
	cerr << endl;
	cerr << "options:" << endl;
	cerr << " -recurse                      analyses all *.mpg files in path" << endl;
	cerr << " -aggregateimage aggregation   print aggregated action values based on" << endl;
	cerr << "                               image action" << endl;
	cerr << " -aggregatemotion aggregation  print aggregated action values based on" << endl;
	cerr << "                               motion action " << endl;
	cerr << endl;
	cerr << "valid values for aggregation: frames, seconds or minutes" << endl;
	cerr << endl;
	cerr << "Example:" <<  endl;
	cerr << "ActionDetection test.mpg" << endl;
	cerr << "analyse MPEG-file \"test.mpg\" and store values" << endl;
	cerr << endl;
	cerr << "ActionDetection -recurse /opt/movies/" << endl; 
	cerr << "analyse all MPEG-files in directory and subdirectory" << endl; 
	cerr << endl;
	cerr << "ActionDetection -aggregateimage seconds test.mpg" << endl; 
	cerr << "analyse \"test.mpg\" and print one aggregated " << endl; 
	cerr << "image-action value for each second" << endl; 
	cerr << endl;
	cerr << "ActionDetection -aggregatemotion minutes test.mpg  " << endl; 
	cerr << "analyse \"test.mpg\" and print one aggregated" << endl; 
	cerr << "motion-action value for each minute" << endl; 
	cerr << endl;
	exit(1);
} // usage


void printAction(ActionData &data, ieee64 frameRate, bool aggregateImage, bool aggregateMotion, bool printFrames, bool printSeconds, bool printMinutes)
{
	Signal<ieee64> dataValues;
	if (aggregateImage)	{
		cout << "image action: " << ends;
		dataValues=data.imgAction;
	} else {
		cout << "motion action: " << ends;
		dataValues=data.motionAction;
	} // else


	int32 frames = dataValues.getSize()[DIM_X];
	if (printFrames) {
		cout << "frame values" << endl;
		for (int32 i=0; i< frames; i++) {
			cout << dataValues(i) << endl;
		} // for
		cout << endl;
	} else if (printSeconds) {
		cout << "aggregated values based on seconds" << endl;
		ieee64 firstFrame=0.0;
		ieee64 lastFrame=frameRate;
		if (lastFrame > frames) lastFrame = frames;
		
		while (firstFrame< frames) {
			ieee64 sum=0.0;
			ieee64 n=0.0;
			for (int32 j=(int32) firstFrame; j<(int32)lastFrame; j++) {
				sum+=dataValues(j);
				n++;
			} // for
			if (n>0) {
				sum/=n;
				cout << sum << endl;
			} // if
			firstFrame+=frameRate;
			lastFrame+=frameRate;
			if (firstFrame>frames) firstFrame=frames;
			if (lastFrame>frames) lastFrame=frames;
		} // for
		cout << endl;
	} else if (printMinutes) {
		cout << "aggregated values based on minutes" << endl;
		ieee64 firstFrame=0.0;
		ieee64 lastFrame=frameRate*60.0;
		if (lastFrame > frames) lastFrame = frames;

		
		while (firstFrame< frames) {
			ieee64 sum=0.0;
			ieee64 n=0.0;
			for (int32 j=(int32) firstFrame; j<(int32)lastFrame; j++) {
				sum+=dataValues(j);
				n++;
			} // for
			if (n>0) {
				sum/=n;
				cout << sum << endl;
			} // if
			firstFrame+=frameRate*60.0;
			lastFrame+=frameRate*60.0;
			if (firstFrame>frames) firstFrame=frames;
			if (lastFrame>frames) lastFrame=frames;
		} // for
		cout << endl;
	}


	return;
}


int main(int argc, char* argv[])
{
    long int startTime = time(NULL);
		
    if (argc<2) usage();
	string filename =argv[argc-1];

	string s;
	bool recurse=false;
	bool aggregateImage=false;
	bool aggregateMotion=false;
	bool frames=false;
	bool seconds=false;
	bool minutes=false;
	for (int i=1; i<argc; i++) {
		s=argv[i];
		for (uint32 k=0; k<s.size(); k++)
		  s[k]=toupper(s[k]);
		
		if (s=="-RECURSE") recurse=true;
		else {
			if (s=="-AGGREGATEMOTION") {
				i++;
				s=argv[i];
				for (uint32 k=0; k<s.size(); k++)
					s[k]=toupper(s[k]);
				if (s=="FRAMES") frames=true;
				else if (s=="SECONDS") seconds=true;
				else if (s=="MINUTES") minutes=true;
				else usage();
				aggregateMotion=true;
			} else 
				if (s=="-AGGREGATEIMAGE") {
					i++;
					s=argv[i];
					for (uint32 k=0; k<s.size(); k++)
						s[k]=toupper(s[k]);
					if (s=="FRAMES") frames=true;
					else if (s=="SECONDS") seconds=true;
					else if (s=="MINUTES") minutes=true;
					else usage();
					aggregateImage=true;
				} else 
					if (i != argc-1) usage();
		} // else
	} // for
	
	ActionData data;
	if (!recurse) {
		ActionDetection action (filename);
		action.process (data);
		cout << data << endl;
		if (aggregateImage || aggregateMotion) {
			VideoDecoder reader (filename);
			ieee64 framerate=reader.getFrameRate();
			printAction(data, framerate, aggregateImage, aggregateMotion, frames, seconds, minutes);
		} // if
	} else {

		// read all MPEG files in directory
		vector<FileInfo> mpgFiles;
		FileList files(filename, true);    
		files.process("*.mpg", mpgFiles);

		// open mpeg files
	    vector<FileInfo>::iterator it;
		for (it = mpgFiles.begin(); it != mpgFiles.end(); it++) {

			cout << endl  << "processing file: "<< (*it).name << endl;
			ActionDetection action ((*it).name);
			action.process (data);
			cout << data << endl;
			if (aggregateImage || aggregateMotion) {
				VideoDecoder reader ((*it).name);
				ieee64 framerate=reader.getFrameRate();
				printAction(data, framerate, aggregateImage, aggregateMotion, frames, seconds, minutes);
			} // if
	} // for
		cout << endl << mpgFiles.size() << " files found." << endl;
	} // else

	cerr << endl << "time: " << (time(NULL)-startTime)  << " seconds" << endl;
	return 0;
} // main

