/*
   This file is part of the MoCA video library.
   Name: ObjectRecognitionTest.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


// C++ includes
#include <string>
#include <iostream>
using namespace std;

// MoCA includes
#include <ObjectMaskFeature.h>
#include <ObjectSegmentationFeature.h>
//#include <BackgroundRecognition.h>
#include <DrawLine.h>
#include <DrawIcon.h>
#include <IImageStream.h>
#include <RegionFeature.h>
#include <FormatedText.h>
#include <CssVideoAnalysis.h>
#include <CssDatabase.h>

//#include <CornerDetection.h>
#include <CameraTransformation.h>
#include <CameraTransformationPrecise.h>
#include <LogoDetection.h>
#include <CssVideoObject.h>
#include <OMpegVideo.h>
#include <VideoIndex.h>
#include <GeometricFeature.h>


string directory = "";

const string objectRecognitionVersion = "1.0";
void usage(bool extendedHelp=false)
{
	cerr << endl;
	cerr << "ObjectRecognition " << objectRecognitionVersion 
	 	 << "    12.Dec.2002   (c) Stephan Kopf, University of Mannheim" << endl;
	cerr << "-------------------------------------------------------------------------------" << endl;
	cerr << endl;
	cerr << "Usage: ObjectRecognition input-file output-file [options]" << endl;
	cerr << endl;
	
	if (extendedHelp) {
		cerr << "Options:" << endl;
		cerr << " -recurse           the input-file is the name of a directory. Objects are" << endl;
		cerr << "                    searched in all *.mpg files in this directory (and sub-" << endl;
		cerr << "                    directories). The parameter output-file is the " << endl;
		cerr << "                    destination directory. The prefix \"objects_\" is added" << endl;
		cerr << "                    to the filename of the XML file." << endl;
		//cerr << " -keepMetadata      metadata files are not deleted when program terminates" << endl;
		cerr << endl;
	} // if
	
	if (!extendedHelp) {
		cerr << "Description:" << endl;
		cerr << "The object recognition application analyses one MPEG-video, finds the shots" << endl;
		cerr << "of the video and analyses each shot. The algorithms search for moving " << endl;
		cerr << "objects in each shot. If a moving object is found, the application" << endl;
		cerr << "tries to recognize this object and stores the results as XML file." << endl;
		cerr << endl;
	} // if
	
	cerr << "Example:" <<  endl;
	cerr << "ObjectRecognition source.mpg dest.xml" << endl;
	cerr << "Analyses \"source.mpg\" and stores the detected objects to \"dest.xml\"." << endl;
	cerr << endl;
	exit(1);
	return;
} // usage

struct ObjectRecognitionParameters
{
	ObjectRecognitionParameters () 
	{
		recurse        = false;
		deleteMetadata = true;
		srcFilename    = "";
		VDXFilename    = "";
	}
	bool recurse;
	bool deleteMetadata;
	string srcFilename;
	string VDXFilename;
};

void parseParameters (int argc, char* argv[], ObjectRecognitionParameters &param)
{
	string s;
	if (argc < 2) usage();
	if (argc < 3) {
		s=argv[1];
		for (string::size_type k=0; k<s.size(); k++)
			s[k]=toupper(s[k]);
		if (s == "HELP") usage();
		if (s == "-HELP") usage(true);
		if (s == "--HELP") usage(true);
		if (s == "/?") usage(true);
		usage();
	} // if
	
	// set filenames
	param.srcFilename = argv[1];
	param.VDXFilename = argv[2];
	
	// read options
	string s1;
	for (int i=3; i<argc; i++) {
		s=argv[i];
		for (uint8 k=0; k<s.size(); k++)
			s[k]=toupper(s[k]);
		
		// recurse
		s1=s.substr(0,8);	
		if (s1 == "-RECURSE") {
			param.recurse = true;
			continue;
		} // if

		// recurse
		s1=s.substr(0,8);	
		
		// keep metadata
		//s1=s.substr(0,13);	
		//if (s1 == "-KEEPMETADATA") {
		//	param.deleteMetadata = false;
		//	continue;
		//} // if
		param.deleteMetadata = false;
		
		usage(false);
	} // for
	
	return;
} // parseParameters


struct CssShotData
{
	CssShotData (CssMaximumData &cmList, int32 frame, int32 shot)
	{
		data = cmList;
		this->frame  = frame;
		this->shot   = shot;
		databaseRef  = -1;
		databaseDist = 10000000;
	}

	CssMaximumData data;
	int32 frame;
	int32 shot;
	int32 databaseRef;
	int32 databaseDist;
};


void getTimeCode (ieee64 frameRate, int32 frameNo, string &ms)
{
	int32 h=(int32)(frameNo/(frameRate*3600));
	int32 m=(int32)((frameNo%((int32)frameRate*3600))/(frameRate*60));
	int32 s=(int32)((frameNo%((int32)frameRate*60))/frameRate);
	int32 f=(frameNo%(int32)frameRate);
	{
	  stringstream str;
	  if (h<10) str<<"0";
	  str << h << ":";
	  if (m<10) str<<"0";
	  str << m << ":";
	  if (s<10) str<<"0";
	  str << s << ":";
	  if (f<10) str<<"0";
	  str << f;
	  ms=str.str();
	}
	return;
} // getTimeCode



// optional function: files with additional metadata information (stored in VDX files extracted from a database)
void initializeGuid(string &pathFilename, string &guid, vector<FileInfo> &origVdxFiles)
{
	guid = "";

	CreateMocaPath mocaPath (pathFilename, false);
	string path = mocaPath.path;
	string file = mocaPath.file;

	int32 p = file.find_first_of('_');
	string prefix = file.substr(0,p);
	string searchID = prefix + "*.vdx";


	
	vector<FileInfo> files = origVdxFiles;
	FileList::select (searchID, files);


	// analyse each file
	vector<FileInfo>::iterator it;
	vector<FormatedTextItem> formatedTextData;
	FormatedTextItem item;
	for (it = files.begin(); it != files.end(); it++) {

		formatedTextData.clear();
		FormatedText ft((*it).name);
		ft.process (formatedTextData);
		string echoItem = "ECHOTYPE";
		ft.findData (echoItem, formatedTextData, item);

		// search for video level
		if ((item.status == 1) && (item.dataName == "'video'")) {
			echoItem = "DMGUID";
			ft.findData (echoItem, formatedTextData, item);
			if (item.status == 1) {
				guid = item.dataName;
				if (guid.substr(0,1) == "'") {
					guid = guid.substr(1);
				} // if
				int32 n=guid.size();
				if (guid.substr(n-1,n) == "'") {
					guid = guid.substr(0,n-1);
				} // if

			} // if
		} // if

	} // for


	return;
} // initializeGuid




void classifyObjects (ObjectRecognitionParameters &param, CssVideoObjectData &objects, bool storeObjects, vector<FileInfo> &vdxFiles)
{

	// initialize database
	vector<CssStoredDatabaseItem> database;
	vector<string> objectClassName;
	CssDatabaseSchemaType type = ECHO_DB_TYPE;
	CssDatabase cssDB (type);
	cssDB.getDatabase (database, objectClassName);


	// get video parameters
	ieee64 framerate=25.0;
	VideoSize vidSize;
	{
		VideoIndexData data;
		VideoIndex index(param.srcFilename);
		index.process (data);
		framerate = data.fps;
		vidSize = data.videoSize;
	}


	SignalSize objectClassSize (objectClassName.size()); 
	Signal<int32> resCounter (objectClassSize);
	Signal<ieee64> resDistance (objectClassSize);

	// aggregate the results
	for (int32 i=0; i<(int32)objects.data.size(); i++) {
		int32 firstShot = i;

		int32 lastShot  = i+1;
		while ((lastShot < (int32)objects.data.size()) && (objects.data[lastShot].shot == objects.data[firstShot].shot)) {
			lastShot++;
		} // while
		lastShot--;

		
		resCounter.clear(0);
		resDistance.clear(0.0); 

		// aggregate results for css data in this shot
		for (int32 j=firstShot; j<=lastShot; j++) {
			int32 dbRef = objects.data[j].databaseRef;
			if (dbRef != -1) {
				int32 dbDist= objects.data[j].databaseDist;
				int32 objId = database[dbRef].objectClassId;
				resCounter(objId)++;
				resDistance(objId)+= dbDist;
			} // if
		} // for
		for (int32 r=0; r<objectClassSize[DIM_X]; r++) {
			if (resCounter(r) > 0) {
				resDistance(r) /= resCounter(r);
			} // if
		} // for

		
		// find best matching object ID
		int32 bestID = 0;
		for (int32 r=1; r<objectClassSize[DIM_X]; r++) {
			if ((resCounter(r) >  resCounter(bestID)) || 
				(resCounter(r) == resCounter(bestID)) && (resDistance(r) <  resDistance(bestID))) {
				bestID = r;
			} // if
		} // for
		int32 bestRef  = -1;
		int32 bestDist = 10000000;
		for (int32 j=firstShot; j<=lastShot; j++) {
			int32 dbRef = objects.data[j].databaseRef;
			if (dbRef != -1) {
				int32 dbDist= objects.data[j].databaseDist;
				int32 objId = database[dbRef].objectClassId;
				if (objId == bestID) {
					bestRef = dbRef;
					bestDist = dbDist;
					break;
				} // if
			} // if
		} // for
		for (int32 j=firstShot; j<=lastShot; j++) {
			int32 dbRef = objects.data[j].databaseRef;
			if (dbRef != -1) {
				int32 dbDist= objects.data[j].databaseDist;
				int32 objId = database[dbRef].objectClassId;
				if ((objId == bestID) && (dbDist < bestDist)){
					bestRef = dbRef;
					bestDist = dbDist;
				} // if
			} // if
		} // for

		// object cannot be detected (count number of objects/other objects)
		int32 hits=0;
		int32 other=0;
		for (int32 j=firstShot; j<=lastShot; j++) {
			int32 dbRef = objects.data[j].databaseRef;
			if (dbRef != -1) {
                                //int32 dbDist= objects.data[j].databaseDist;
				int32 objId = database[dbRef].objectClassId;
				if (objId == bestID) {
					hits++;
				} else {
					other++;
					objects.data[j].databaseRef = -1;
				} // else
			} // if
		} // for


		bool itemRemoved = true;
		int32 removedObjects = 0;
		while (itemRemoved) {
			itemRemoved = false;

			// a simple tracking, get avg position and width
			int32 px=0;
			int32 py=0;
			int32 wx=0;
			int32 wy=0;
			int32 n=0;
			for (int32 j=firstShot; j<=lastShot; j++) {
				int32 dbRef = objects.data[j].databaseRef;
				if (dbRef != -1) {
					n++;
					px += (objects.data[j].px1+objects.data[j].px2)/2;
					py += (objects.data[j].py1+objects.data[j].py2)/2;
					wx += (objects.data[j].px2-objects.data[j].px1);
					wy += (objects.data[j].py2-objects.data[j].py1);
				} // if
			} // for
			if (n>1) {
				px/=n;
				py/=n;
				wx/=n;
				wy/=n;
			} // if

			// track the object
			int32 thPx = vidSize[DIM_X]/5;
			int32 thPy = vidSize[DIM_Y]/5;
			int32 thWx = vidSize[DIM_X]/4;
			int32 thWy = vidSize[DIM_Y]/4;
			for (int32 j=firstShot; j<=lastShot; j++) {
				if (itemRemoved) break;
				int32 dbRef = objects.data[j].databaseRef;
				if (dbRef != -1) {
					int32 firstObj = j;
					int32 secondObj = j+1;
					while ((secondObj <= lastShot) && (objects.data[secondObj].databaseRef < 0)) {
						secondObj++;
					} // while
					if (secondObj <= lastShot) {
						dbRef = objects.data[secondObj].databaseRef;
						if (dbRef != -1) {
							// compare both frames
							int32 px1 = (objects.data[firstObj].px1+objects.data[firstObj].px2)/2;
							int32 py1 = (objects.data[firstObj].py1+objects.data[firstObj].py2)/2;
							int32 wx1 = (objects.data[firstObj].px2-objects.data[firstObj].px1);
							int32 wy1 = (objects.data[firstObj].py2-objects.data[firstObj].py1);
							
							int32 px2 = (objects.data[secondObj].px1+objects.data[secondObj].px2)/2;
							int32 py2 = (objects.data[secondObj].py1+objects.data[secondObj].py2)/2;
							int32 wx2 = (objects.data[secondObj].px2-objects.data[secondObj].px1);
							int32 wy2 = (objects.data[secondObj].py2-objects.data[secondObj].py1);
							
							int diff=0;
							if (labs(px1-px2) > thPx) diff=1;
							if (labs(py1-py2) > thPy) diff=2;
							if (labs(wx1-wx2) > thWx) diff=3;
							if (labs(wy1-wy2) > thWy) diff=4;
							
							if (diff!=0) {
								// remove the frame, that does not fit
								int32 d1=0;
								int32 d2=0;
								switch (diff) {
								case 1:
									d1=labs (px1-px);
									d2=labs (px2-px);
									break;
								case 2:
									d1=labs (py1-py);
									d2=labs (py2-py);
									break;
								case 3:
									d1=labs (wx1-wx);
									d2=labs (wx2-wx);
									break;
								case 4:
									d1=labs (wy1-wy);
									d2=labs (wy2-wy);
									break;
								} // switch
								
								// remove object
								if (d1>d2) {
									objects.data[firstObj].databaseRef = -1;
								} else {
									objects.data[secondObj].databaseRef = -1;
								} // else
								itemRemoved = true;
								removedObjects++;
							} // if
						} // if
					} // if
				} // if
			} // for
		} // while
		int32 keptObjects = 0;
		for (int32 j=firstShot; j<=lastShot; j++) {
			int32 dbRef = objects.data[j].databaseRef;
			if (dbRef >= 0) keptObjects++;
		} // for


		// remove hists
		if (hits<other) {
			// remove hits
			for (int32 j=firstShot; j<=lastShot; j++) {
				objects.data[j].databaseRef = -1;
			} // for
		} // if
		if (hits<=1) {
			// remove hits
			for (int32 j=firstShot; j<=lastShot; j++) {
				objects.data[j].databaseRef = -1;
			} // for
		} // if

		
		i = lastShot;
	} // for


	
	// initialize variables
	string echoXmlFilename = param.VDXFilename;
	string echoMpegFilename = "";
	string tmpPath = "";
	{
		CreateMocaPath mocaPath (param.srcFilename, false);
		echoMpegFilename = mocaPath.file;
		tmpPath = mocaPath.path;
	}

    FileHandling::setCurrentDirectory (directory);
	string echoGuid = "";
	initializeGuid(param.srcFilename, echoGuid, vdxFiles);
    FileHandling::setCurrentDirectory (directory);


    int32 totalObjects = 0;
	VideoDecoder reader (param.srcFilename);
	Image<uint8> img;
	
	try {
		ofstream ofs (echoXmlFilename.c_str(), std::ios::out | std::ios::binary); 
		ofs << "<ObjectDetectionList>" << endl;
		ofs << "<GUID>" << echoGuid << "</GUID>" << endl;
		ofs << "<filename>" << echoMpegFilename << "</filename>" << endl;

		for (int32 i=0; i<(int32)objects.data.size(); i++) {

			int32 firstShot = i;
			int32 lastShot  = i+1;
			while ((lastShot < (int32)objects.data.size()) && (objects.data[lastShot].shot == objects.data[firstShot].shot)) {
				lastShot++;
			} // while
			lastShot--;


			// get parameters for each frame
			// initialize
			int32 firstFrame = objects.data[firstShot].frame;
			int32 lastFrame = objects.data[lastShot].frame;
			int32 nFrames = lastFrame - firstFrame + 1;
			SignalSize sigSize (nFrames, 6);
			Signal<int32> frameList (sigSize);
			for (int32 b=0; b<sigSize[DIM_BAND]; b++) {
				for (int32 x=0; x<sigSize[DIM_X]; x++) {
					frameList(x,b) = -2;
				} // for
			} // for


			// get reliability parameters
			int32 nObjects = 0;
			for (int32 j=firstShot; j<=lastShot; j++) {
				int32 dbRef = objects.data[j].databaseRef;
				if (dbRef >= 0) nObjects++;
			} // for
			ieee64 reliability = (8*nObjects);
			if (nObjects<2) reliability = 10;
			if (nObjects>=8) reliability = 100;

			// reliability based on tracking values
			int32 px=0;
			int32 py=0;
			int32 wx=0;
			int32 wy=0;
			int32 n=0;
			for (int32 j=firstShot; j<=lastShot; j++) {
				int32 dbRef = objects.data[j].databaseRef;
				if (dbRef != -1) {
					n++;
					px += (objects.data[j].px1+objects.data[j].px2)/2;
					py += (objects.data[j].py1+objects.data[j].py2)/2;
					wx += (objects.data[j].px2-objects.data[j].px1);
					wy += (objects.data[j].py2-objects.data[j].py1);
				} // if
			} // for
			if (n>1) {
				px/=n;
				py/=n;
				wx/=n;
				wy/=n;
			} // if



			// set calculated parameters
			for (int32 j=firstShot; j<=lastShot; j++) {
				int32 frame = objects.data[j].frame;
				frame -= firstFrame;

				// frameList(frame,0) : ref
				frameList(frame,0) = objects.data[j].databaseRef;
				//int32 tmp1 = frameList(frame,0);

				if (frameList(frame,0) >= 0) {

					// frameList(frame,1-4) : border
					frameList(frame,1) = objects.data[j].px1;
					frameList(frame,2) = objects.data[j].px2;
					frameList(frame,3) = objects.data[j].py1;
					frameList(frame,4) = objects.data[j].py2;


					// get reliability for frame based on object size/position
					int32 px1 = (objects.data[j].px1+objects.data[j].px2)/2;
					int32 py1 = (objects.data[j].py1+objects.data[j].py2)/2;
					int32 wx1 = (objects.data[j].px2-objects.data[j].px1);
					int32 wy1 = (objects.data[j].py2-objects.data[j].py1);
					ieee64 dx = ((ieee64)labs(px-px1)) / vidSize[DIM_X];
					ieee64 dy = ((ieee64)labs(py-py1)) / vidSize[DIM_Y];
					ieee64 dwx = ((ieee64)labs(wx-wx1)) / vidSize[DIM_X];
					ieee64 dwy = ((ieee64)labs(wy-wy1)) / vidSize[DIM_Y];

					ieee64 f = 1.0 - (dx+dy+dwx+dwy) / 2.0;
					if (f < 0.05) f = 0.05;
					if (f > 1.0) f = 1.0;


					// get reliability for frame based on matching distance
					ieee64 matchingDist = objects.data[j].databaseDist;
					ieee64 g = 1.0 - (matchingDist - 800) / 2000.0;
					if (g < 0.05) g = 0.05;
					if (g > 1.0)  g = 1.0;



					frameList(frame,5) = (int32)(reliability*f*g);
					if (frameList(frame,5) < 1) frameList(frame,5) = 1;
					if (frameList(frame,5) > 100) frameList(frame,5) = 100;
					//ieee64 tmp = frameList(frame,5);

				} // if
			} // for


			// interpolate unknown frames
			for (int32 j=0; j<nFrames; j++) {

				int32 ref = frameList(j,0);
				if (ref != -2) {
					int32 i=j+1;

					while ((i<nFrames) && (frameList(i,0) == -2)) {
						if (frameList(i,0) == -2) {
							frameList(i,0) = frameList(j,0);
							frameList(i,1) = frameList(j,1);
							frameList(i,2) = frameList(j,2);
							frameList(i,3) = frameList(j,3);
							frameList(i,4) = frameList(j,4);
							frameList(i,5) = frameList(j,5);
						} // if
						i++;
					} // while
					if (j<i-2) j=i-2;
				} // if
			} // for



			// store data
			for (int32 j=0; j<nFrames; j++) {
				
				int32 ref = frameList(j,0);
				
				if (ref >= 0) {
					int32 frame = firstFrame + j;
					string timecode;
					getTimeCode (framerate, frame, timecode);
					
					int32 position_x1 = frameList(j,1);
					int32 position_x2 = frameList(j,2);
					int32 position_y1 = frameList(j,3);
					int32 position_y2 = frameList(j,4);
					int32 reliability = frameList(j,5);
					
					//CssDatabaseItem item = DATABASE_CSS[ref];
					CssStoredDatabaseItem item = database[ref];
					int32 classID = item.objectClassId;
					string objectName = objectClassName [classID];
					//string objectName = DATABASE_OBJECT_CLASSES [classID];
					
					if ((objectName == "BOATS") || (objectName == "CARS") || (objectName == "PEOPLE")) {

						string objName = objectName;
						if (objectName == "PEOPLE") objName = "PERSON";
						
						ofs << "    <object>" << endl;
						
						ofs << "        <timecode>" << timecode << "</timecode>" << endl;
						ofs << "        <position_x1>" << position_x1 << "</position_x1>" << endl;
						ofs << "        <position_y1>" << position_y1 << "</position_y1>" << endl;
						ofs << "        <position_x2>" << position_x2 << "</position_x2>" << endl;
						ofs << "        <position_y2>" << position_y2 << "</position_y2>" << endl;
						ofs << "        <reliability>" << reliability << "</reliability>" << endl;
						ofs << "        <name>" << objName << "</name>" << endl;
						
						ofs << "    </object>" << endl;
						
						totalObjects++;
						
						if (storeObjects) {
							
							// store frames (*=BOATS, **=CARS, ***=PEOPLE)
							// set filename
							string imgFilename = "";
							string imgOrigFilename = "";
							{
								CreateMocaPath mocaPath (param.srcFilename, false);
								string path = mocaPath.path;
								string fn   = mocaPath.file;
								fn = fn.substr(0,fn.size()-4);
								
								char buf[32];
								sprintf (buf,"%06li",frame);
								string sNumber = buf;
								imgFilename     = path +"object_" +  fn + "_" + sNumber + ".png";
								imgOrigFilename = path +"img_"    +  fn + "_" + sNumber + ".png";
							}
							// read image
							img = reader.getImage(frame);
							
							GrayscaleFilter<uint8> gray;
							gray.process (img);
							
							//cerr << "store image: " << imgOrigFilename << endl;
							//WriteImage (img, imgOrigFilename);
							
							
							DrawRectangle rect (Color::WHITE);
							ImagePosition p1(position_x1, position_y1);
							ImagePosition p2(position_x2, position_y2);
							rect.process(p1,p2,img);
							
							DrawIcon draw (Color::WHITE);
							if (objectName == "BOATS") {
							    draw.process (ImagePosition (5,img.getSize()[DIM_Y]-28), ICON_BOAT, img);
							} // if
							if (objectName == "CARS") {
							    draw.process (ImagePosition (5,img.getSize()[DIM_Y]-25), ICON_CAR, img);
							} // if
							if (objectName == "PEOPLE") {
							    draw.process (ImagePosition (5,img.getSize()[DIM_Y]-40), ICON_PEOPLE, img);
							} // if
							
							WriteImage (img, imgFilename);
						} // if (storeObjects)
						
					} // if
				} // if
			} // for

			i = lastShot;
		} // for

		ofs << "</ObjectDetectionList>" << endl;
		
		ofs.close();
	} catch (...) {
		cerr << endl << "ObjectRecognition: could not store XML data file!" << endl << endl;
	} // catch

	cerr << "number of objects: " << totalObjects << endl;
	return;
} // classifyObjects


void searchObjectsInFile (ObjectRecognitionParameters &param, vector<FileInfo> &vdxFiles)
{

    FileHandling::setCurrentDirectory (directory);

	try {
		// validate filenames
		FileHandling::convertToValidUnixFilename (param.srcFilename);
		FileHandling::convertToValidUnixFilename (param.VDXFilename);
		
		cerr << endl;
		cerr << "source MPEG file: " << param.srcFilename << endl;
		cerr << "XML object file : " << param.VDXFilename << endl;
		cerr << endl;

		// start process: object masks
		{
			CssVideoObjectData objects;
			{
				CssVideoObject cssObjects (param.srcFilename);
				cssObjects.process (objects);
			}

			bool storeObjects = false;
			classifyObjects (param, objects, storeObjects, vdxFiles);


#if 0
			ObjectMaskData objectMasksData;
			int minimumShotLength = 12;
			bool preciseCalculation=false;
			ObjectMask objMask(param.srcFilename, minimumShotLength, preciseCalculation);
			objMask.process(objectMasksData);
			
			//objMask.validateResults(objectMasksData);


#endif

		}

#if 0
		VideoDecoder reader (param.srcFilename);
		int32 frames = reader.getSize()[DIM_Z];
		ieee64 vidLength  = (frames / reader.getFrameRate())/60;
		ieee64 calcLength = (time(NULL) - start)/60.0;
		cerr << endl 
			 << "-----------------------------------------" << endl 
			 <<"calculation    : " <<  (int32)calcLength << " min" << endl
			 <<"video          : " <<  (int32)vidLength << " min" << endl
			 <<"factor realtime: " <<  (int32)(calcLength / vidLength) << endl
			 << "-----------------------------------------" << endl;

		char beep = 7; cout << beep; cout << beep; cout << beep;
#endif

	} catch (...) {
		cerr << endl << "error in ObjectRecognitionTest::searchObjectsInFile !" << endl;
	}


	return;
} // searchObjectsInFile 


#include <VideoEncoder.h>


int main(int argc, char* argv[])
{

    FileHandling::getCurrentDirectory (directory);


	// analyse parameters
	ObjectRecognitionParameters param;
	parseParameters (argc, argv, param);


	
	if (!param.recurse) {
		
		// read all VDX files in directory
		vector<FileInfo> vdxFiles;

		searchObjectsInFile (param, vdxFiles);		
		
	} else {
		
		// set source path name
		string fullPathSrc = "";
		{
			CreateMocaPath mocaPath1 (param.srcFilename, false);
			if (mocaPath1.pathFile.length() == mocaPath1.path.length()) {
				fullPathSrc = mocaPath1.path;
			} else {
				string s = param.srcFilename + '/';
				CreateMocaPath mocaPath2 (s, false);
				fullPathSrc = mocaPath2.path;
			} // else
		}
		
		
		// set destination path name
		string fullPathDest = "";
		{
			CreateMocaPath mocaPath1 (param.VDXFilename, false);
			if (mocaPath1.pathFile.length() == mocaPath1.path.length()) {
				fullPathDest = mocaPath1.path;
			} else {
				string s = param.VDXFilename + '/';
				CreateMocaPath mocaPath2 (s, false);
				fullPathDest = mocaPath2.path;
			} // else
		}
		cerr << endl;
		cerr << "source path      : " << fullPathSrc << endl;
		cerr << "destination path : " << fullPathDest << endl;
		
		// read all MPEG files in directory
		vector<FileInfo> mpgFiles;
		{
		    FileList files(fullPathSrc, false);    
		    files.process("*.mpg", mpgFiles);
		    //files.unselect("abstract*.mpg", mpgFiles);
		}


		// read all VDX files in directory
		vector<FileInfo> vdxFiles;
		{
		    FileList filesVdx(fullPathSrc, false);
		    string searchID = "*.vdx";
		    filesVdx.process(searchID, vdxFiles);
		}


		// open mpeg files
		vector<FileInfo>::iterator it;
		for (it = mpgFiles.begin(); it != mpgFiles.end(); it++) {

			// set filenames
			param.srcFilename = (*it).name;
			param.VDXFilename = fullPathDest + (*it).shortName;
			param.VDXFilename = param.VDXFilename.substr(0, param.VDXFilename.size()-3);
			param.VDXFilename += "xml";
			
			

			searchObjectsInFile (param, vdxFiles);		
		} // for
		cout << endl << mpgFiles.size() << " files found." << endl;
	} // else
	
	cerr << endl;
	cerr << "Moving objects analysed: " << param.VDXFilename << endl;

#if 0
	cerr << endl << "time: " << (time(NULL)-startTime)  << " seconds" << endl;
#endif
	return 0;
} // main

