/*
   This file is part of the MoCA video library.
   Name: VideoAbstract.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


// C++ includes
#include <string>
#include <iostream>
using namespace std;

// MoCA includes
#include <VideoAbstractFeature.h>


#undef DISPLAY_TIME
//#define DISPLAY_TIME 1
#ifdef DISPLAY_TIME
#include<time.h>
static double videoSeconds = 0;
#endif

#if 0
void usage()
{
	cerr << "Usage: VideoAbstract [options] filename|directory" << endl;
	cerr << "read a MPEG-video \"file.mpg\" and save the video abstract \"abstract_file.mpg\" " << endl;
	cerr << endl;
	cerr << "options:" << endl;
	cerr << " -recurse  create one abstract for each *.mpg file in the specified directory" << endl;
	cerr << " -length   length of the video abstract in minutes" << endl;
	cerr << " -percent  length of the video abstract in percent of the original video" << endl;
	cerr << " -faces    find faces in the video (very slow)" << endl;
	cerr << " -create   in very short videos no abstract is created. This" << endl;
	cerr << "           options forces the creation of the abstract" << endl;
	cerr << endl;
	cerr << "Example:" <<  endl;
	cerr << "VideoAbstract test.mpg" << endl;
	cerr << "Analyses \"test.mpg\" and creates the MPEG video \"abstract_test.mpg\" " << endl;
	cerr << "The default length of the abstract is 20 percent of the original video length." << endl;
	cerr << endl;
	cerr << "VideoAbstract -length 1.5 test1.mpg" << endl;
	cerr << "Analyses \"test1.mpg\" and creates the MPEG video \"abstract_test1.mpg\" " << endl;
	cerr << "The length of the abstract is 90 seconds (1.5 minutes)." << endl;
	cerr << endl;
	cerr << "VideoAbstract -recurse -percent 12.5 D:/videos/" << endl;
	cerr << "Analyses all videos in directory. The length of each abstract is " << endl;
	cerr << "12.5 percent length of the original video." << endl;
	cerr << endl;
	exit(1);
} // usage
#endif

const string videoAbstractingVersion = "1.0";
void usage(bool extendedHelp)
{
	cerr << endl;
	cerr << "VideoAbstract " << videoAbstractingVersion 
		 << "    27.Nov.2002   (c) Stephan Kopf, University of Mannheim" << endl;
	cerr << "---------------------------------------------------------------------------" << endl;
	cerr << endl;
	cerr << "Usage: VideoAbstract input-file output-file [options]" << endl;
	cerr << endl;
	cerr << "Options:" << endl;
	cerr << " AVDocument=filename   AVDocument of the ECHO metadata schema" << endl;
	cerr << " VideoLevel=filename   VideoLevel of the ECHO metadata schema" << endl;
	cerr << " VDX=filename          VDX data of the ECHO metadata schema" << endl;
	cerr << " length=<float>        length of the abstract [in percent]" << endl;
	cerr << endl;
	if (extendedHelp) {
		cerr << " minutes=<float>       absolute length of the abstract [in minutes]" << endl;
		cerr << " recurse               the input-file is the name of a directory. Abstracts are" << endl;
		cerr << "                       created of all *.mpg files in this directory (and sub-" << endl;
		cerr << "                       directories). The parameter output-file is the " << endl;
		cerr << "                       destination directory. The prefix \"abstract_\" is added" << endl;
		cerr << "                       to the filename of the abstract." << endl;
		cerr << " nofaces               the abstract is created without analysing faces (faster)" << endl;
		cerr << " keepMetadata          metadata files are not deleted when program terminates" << endl;
		cerr << " keepAllData           metadata files and temporary files are not deleted " << endl;
		cerr << " store-abs             requires \"keepmetadata\", ignore the output-path and" << endl;
		cerr << "                       store all abstracts as *.abs-files in the same directory" << endl;
		cerr << " noAbstract            calculate metadata, but do not create the abstract" << endl;
		cerr << " help                  display this information" << endl;
		cerr << endl;
		cerr << endl;
	} // if
	if (!extendedHelp) {
		cerr << "Description:" << endl;
		cerr << "The video abstracting application analyses one MPEG-video and creates a short" << endl;
		cerr << "version of this video. The most important clips of the original video are" << endl;
		cerr << "selected for the video abstract. The application detects and analyses camera" << endl;
		cerr << "shots, scenes, histograms, faces, action, and the audio track. " << endl;
		cerr << endl;
	cerr << "Example:" <<  endl;
	cerr << "VideoAbstract source.mpg dest.mpg" << endl;
	cerr << "Analyses \"source.mpg\" and creates the abstract \"dest.mpg\"." << endl;
	cerr << endl;
	cerr << "VideoAbstract source.mpg C:\\videos\\dest.mpg length=30" << endl;
	cerr << "The length of the abstract is 30% the length of the original MPEG-video." << endl;
	cerr << endl;
	cerr << "VideoAbstract src.mpg dest.mpg  AVDocument=C:\\data\\av.xml VDX=C:\\data\\vdx.xml" << endl;
	cerr << "Additional text files with metadata information are analysed. The format" << endl;
	cerr << "of these files is based on the ECHO metadata schema." << endl;
	cerr << endl;
	} // if
	exit(1);
	return;
} // usage


void parseParameters (int argc, char* argv[], VideoAbstractParameters &param)
{
	string s;
	if (argc < 2) usage(false);
	if (argc < 3) {
		s=argv[1];
		for (string::size_type k=0; k<s.size(); k++)
			s[k]=toupper(s[k]);
		if (s == "HELP") usage(true);
		if (s == "-HELP") usage(true);
		if (s == "--HELP") usage(true);
		if (s == "/?") usage(true);
		usage(false);
	} // if

	// set filenames
	param.srcFilename  = argv[1];
	param.destFilename = argv[2];
	param.deleteMetadata = true;

	// read options
	string s1;
	for (int i=3; i<argc; i++) {
		s=argv[i];
		for (uint8 k=0; k<s.size(); k++)
		  s[k]=toupper(s[k]);
		
		// AVDocument
		s1=s.substr(0,11);	
		if (s1 == "AVDOCUMENT=") {
			s=argv[i];
			param.AVDocumentFilename = s.substr(11);
			continue;
		} // if

		// VideoLevel
		s1=s.substr(0,11);	
		if (s1 == "VIDEOLEVEL=") {
			s=argv[i];
			param.videoLevelFilename = s.substr(11);
			continue;
		} // if

		// VideoLevel
		s1=s.substr(0,4);	
		if (s1 == "VDX=") {
			s=argv[i];
			param.VDXFilename = s.substr(4);
			continue;
		} // if

		// length (percent)
		s1=s.substr(0,7);	
		if (s1 == "LENGTH=") {
			s1 = s.substr(7);
			param.lengthPercent = atof(s1.c_str());
			continue;
		} // if

		// minutes
		s1=s.substr(0,8);	
		if (s1 == "MINUTES=") {
			s1 = s.substr(8);
			param.lengthMinutes = atof(s1.c_str());
			continue;
		} // if

		// recurse
		s1=s.substr(0,7);	
		if (s1 == "RECURSE") {
			param.recurse = true;
			continue;
		} // if

		// faces
		s1=s.substr(0,7);	
		if (s1 == "NOFACES") {
			param.faces = false;
			continue;
		} // if

		// keep metadata
		s1=s.substr(0,12);	
		if (s1 == "KEEPMETADATA") {
			param.deleteMetadata = false;
			continue;
		} // if

		// keep metadata
		s1=s.substr(0,11);	
		if (s1 == "KEEPALLDATA") {
			param.deleteMetadata = false;
			param.deleteTemporaryData = false;
			continue;
		} // if

		// no abstract
		s1=s.substr(0,10);	
		if (s1 == "NOABSTRACT") {
			param.createAbstract = false;
			continue;
		} // if

		// length (percent)
		s1=s.substr(0,9);	
		if (s1 == "STORE-ABS") {
			param.storeExtensionAbs = true;
			continue;
		} // if

		usage(false);
	} // for

	return;
} // parseParameters





void abstractFile (string &srcFilename, string &destFilename, VideoAbstractParameters &param)
{

	try
	{
		// validate filenames
		FileHandling::convertToValidFilename (srcFilename);
		FileHandling::convertToValidFilename (destFilename);
		FileHandling::convertToValidFilename (param.AVDocumentFilename);
		FileHandling::convertToValidFilename (param.VDXFilename);
		FileHandling::convertToValidFilename (param.videoLevelFilename);
		
				
		cerr << endl;
		cerr << "source file   : " << srcFilename << endl;
		cerr << "abstract file : " << destFilename << endl;
		if (param.AVDocumentFilename != "") {
			cerr << "AV-Document   : " << param.AVDocumentFilename << endl;
		} // if
		if (param.videoLevelFilename != "") {
			cerr << "Video-Document: " << param.videoLevelFilename << endl;
		} // if
		if (param.VDXFilename != "") {
			cerr << "VDX-Document  : " << param.VDXFilename << endl;
		} // if
		if (param.lengthMinutes != -1.0) cerr << "length       : " << param.lengthMinutes << " minutes" << endl;
		if (param.lengthPercent != -1.0) cerr << "length       : " << param.lengthPercent << " percent" << endl;
		cerr << endl;

#ifdef DISPLAY_TIME
		MpegIndexData tmpData;
		MpegIndex tmpIdx (srcFilename);
		tmpIdx.process (tmpData);
		videoSeconds += tmpData.videoSize[DIM_Z]/tmpData.frameRate;
#endif

		// start process
		VideoAbstract abstract (srcFilename, destFilename, param);
		abstract.process();
    } 
	catch (MocaException e) 
	{
		string errMessage = e.what();
        cerr << e.what() << endl;

		string::size_type pos = errMessage.find_first_of("The length of the source video is too short");
		if (pos != string::npos) {

			// video is too short
			if (!param.recurse) {
				exit(2);
			} // if
		}
    } // catch
	catch (...)
	{
        exit (1);
	}
	

	return;
} // abstractFile 



int main(int argc, char* argv[])
{
#ifdef DISPLAY_TIME
    long int startTime = time(NULL);
#endif


	// analyse parameters
	VideoAbstractParameters param;
	parseParameters (argc, argv, param);
		
	if (!param.recurse) {

		abstractFile (param.srcFilename, param.destFilename, param);		

	} else {

		// set source path name
		{
			CreateMocaPath mocaPath1 (param.srcFilename, false);
			if (mocaPath1.pathFile.length() == mocaPath1.path.length()) {
				param.srcFilename = mocaPath1.path;
			} else {
				string s = param.srcFilename + '/';
				CreateMocaPath mocaPath2 (s, false);
				param.srcFilename = mocaPath2.path;
			} // else
		}
		// set destination path name
		{
			CreateMocaPath mocaPath1 (param.destFilename, false);
			if (mocaPath1.pathFile.length() == mocaPath1.path.length()) {
				param.destFilename = mocaPath1.path;
			} else {
				string s = param.destFilename + '/';
				CreateMocaPath mocaPath2 (s, false);
				param.destFilename = mocaPath2.path;
			} // else
		}

		cerr << endl;
		cerr << "source path      : " << param.srcFilename << endl;
		cerr << "destination path : " << param.destFilename << endl;
		if (param.storeExtensionAbs) {
		    cerr << "ignore destination path, store abstracts as \"abs\"-files instead." << endl;
		} // if
	       

		// read all MPEG files in directory
		vector<FileInfo> mpgFiles;
		FileList files(param.srcFilename, true);    
		files.process("*.mpg", mpgFiles);
		files.unselect("abstract*.mpg", mpgFiles);

		// set destination path
		CreateMocaPath mocaPath (param.destFilename, false);
		string destPath = mocaPath.path; 
	
		// open mpeg files
		vector<FileInfo>::iterator it;
		for (it = mpgFiles.begin(); it != mpgFiles.end(); it++) {

			// set filenames
			string scrFilename  = (*it).name;
			string destFilename = destPath + "abstract_" + (*it).shortName;
			if (param.storeExtensionAbs) {
			    destFilename = scrFilename.substr(0,scrFilename.size()-4) + ".abs";
			} // if


			abstractFile (scrFilename, destFilename, param);		
		} // for
		cout << endl << mpgFiles.size() << " files found." << endl;
	} // else
	
	cerr << endl;
	cerr << "Video abstract created: " << param.destFilename << endl;

#ifdef DISPLAY_TIME
	long int totalTime = (time(NULL)-startTime);
 	long int h=totalTime/3600;
	long int min=(totalTime%3600)/60;
	long int sec=totalTime%60;
	cerr << endl << "total time: " << ends;
	if (h<10) cerr << "0"; cerr << h << ":";
	if (min<10) cerr << "0"; cerr << min << ":";
	if (sec<10) cerr << "0"; cerr << sec << ends;

	cerr << "     video length: " << ends;
	totalTime = (int32)videoSeconds;
 	h=totalTime/3600;
	min=(totalTime%3600)/60;
	sec=totalTime%60;
	if (h<10) cerr << "0"; cerr << h << ":";
	if (min<10) cerr << "0"; cerr << min << ":";
	if (sec<10) cerr << "0"; cerr << sec << endl;
#endif
	return 0;
} // main


