/*
   This file is part of the MoCA video library.
   Name: VideoStatisticTest.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#include <VideoStatistic.h>
#include <FileList.h>


void usage()
{
	cerr << "Usage: VideoStatistic [options] filename|path" << endl;
	cerr << VIDEO_STATISTIC_VERSION << endl;
	cerr << endl;
	cerr << "options:" << endl;
	cerr << " -recurse                analyses all *.mpg files in path" << endl;
	cerr << " -aggregate TYPE VALUE   print aggregated action values based on TYPE" << endl;
	cerr << endl;
	cerr << "TYPE:  AVG_Y, AVG_U, AVG_V      : average Y-,U-, or V-values" << endl;
	cerr << "       SIGMA_Y, SIGMA_U, SIGMA_V: standard deviation of plane" << endl;
	cerr << "       CONTRAST_MAX             : maximum contrast of frames " << endl;
	cerr << "       CONTRAST_HIST            : contrast based on histogram values" << endl;
	cerr << "       CONTRAST_SCALED          : contrast based on scaled images" << endl;
	cerr << "                                  the contrast values are scaled to (0..100)" << endl;
	cerr << "VALUE: aggregation based on FRAMES, SECONDS or MINUTES" << endl;
	cerr << endl;
	cerr << endl;
	cerr << "Example:" <<  endl;
	cerr << "VideoStatistic test.mpg" << endl;
	cerr << "analyse MPEG-file \"test.mpg\" and store values" << endl;
	cerr << endl;
	cerr << "VideoStatistic -recurse /opt/movies/" << endl; 
	cerr << "analyse all MPEG-files in directory and subdirectory" << endl; 
	cerr << endl;
	cerr << "VideoStatistic -aggregate AVG_Y SECONDS test.mpg" << endl; 
	cerr << "analyse \"test.mpg\" and print the aggregated AVG_Y value for each second" << endl; 
	cerr << endl;
	cerr << "VideoStatistic -aggregate CON_HIST MINUTES test.mpg" << endl; 
	cerr << "analyse \"test.mpg\" and print one aggregated contrast " << endl; 
	cerr << "value based on histograms (CON_HIST) for each minute" << endl; 
	cerr << endl;
	exit(1);
} // usage


#include<IImageStream.h>
#include<GrayscaleFilter.h>
#include<ImageStatistic.h>


void printAction(VideoStatisticData &data, ieee64 frameRate, int32 aggregateType, int32 aggregateValue)
{
	Signal<ieee64> dataValues;
	SignalSize size (data.imgData.size());
	dataValues.resize(size);

	switch (aggregateType) {
	case 0: 
		for (int32 i=0; i<size[DIM_X]; i++)
			dataValues(i) = data.imgData[i].avg[0];
		break;
	case 1: 
		for (int32 i=0; i<size[DIM_X]; i++)
			dataValues(i) = data.imgData[i].avg[1];
		break;
	case 2: 
		for (int32 i=0; i<size[DIM_X]; i++)
			dataValues(i) = data.imgData[i].avg[2];
		break;
	case 3: 
		for (int32 i=0; i<size[DIM_X]; i++)
			dataValues(i) = data.imgData[i].sigma[0];
		break;
	case 4: 
		for (int32 i=0; i<size[DIM_X]; i++)
			dataValues(i) = data.imgData[i].sigma[1];
		break;
	case 5: 
		for (int32 i=0; i<size[DIM_X]; i++)
			dataValues(i) = data.imgData[i].sigma[2];
		break;
	case 6: 
		for (int32 i=0; i<size[DIM_X]; i++)
			dataValues(i) = data.imgData[i].contrastMaxMin;
		break;
	case 7: 
		for (int32 i=0; i<size[DIM_X]; i++)
			dataValues(i) = data.imgData[i].contrastHistogram;
		break;
	case 8: 
		for (int32 i=0; i<size[DIM_X]; i++)
			dataValues(i) = data.imgData[i].contrastScale;
		break;
	default:
		cout << "unknown option";
	} // switch


	// print data
	int32 frames = dataValues.getSize()[DIM_X];
	switch (aggregateValue) {
	case 0: 
		cout << "frame values" << endl;
		for (int32 i=0; i< frames; i++) {
			cout << dataValues(i) << endl;
		} // for
		cout << endl;
		break;
	case 1: 
		{
			cout << "aggregated values based on seconds" << endl;
			ieee64 firstFrame=0.0;
			ieee64 lastFrame=frameRate;
			if (lastFrame > frames) lastFrame = frames;
			
			while (firstFrame< frames) {
				ieee64 sum=0.0;
				ieee64 n=0.0;
				for (int32 j=(int32) firstFrame; j<(int32)lastFrame; j++) {
					sum+=dataValues(j);
					n++;
				} // for
				if (n>0) {
					sum/=n;
					cout << sum << endl;
				} // if
				firstFrame+=frameRate;
				lastFrame+=frameRate;
				if (firstFrame>frames) firstFrame=frames;
				if (lastFrame>frames) lastFrame=frames;
			} // for
			cout << endl;
		}
		break;
	case 2: 
		{
			cout << "aggregated values based on minutes" << endl;
			ieee64 firstFrame=0.0;
			ieee64 lastFrame=frameRate*60.0;
			if (lastFrame > frames) lastFrame = frames;
			
			while (firstFrame< frames) {
				ieee64 sum=0.0;
				ieee64 n=0.0;
				for (int32 j=(int32) firstFrame; j<(int32)lastFrame; j++) {
					sum+=dataValues(j);
					n++;
				} // for
				if (n>0) {
					sum/=n;
					cout << sum << endl;
				} // if
				firstFrame+=frameRate*60.0;
				lastFrame+=frameRate*60.0;
				if (firstFrame>frames) firstFrame=frames;
				if (lastFrame>frames) lastFrame=frames;
			} // for
			cout << endl;
		}
		break;
	default:
		cout << "unknown option";
		exit(1);
	} // switch

	return;
}


int main(int argc, char* argv[])
{

	bool recurse=false;
	string filename="";
	string s = "";

	if (argc<2) usage();
	filename=argv[argc-1];

	bool aggregate      =false;
	int32 aggregateType=-1;  // AVG_Y=0, AVG_U=1, AVG_V=2, SIGMA_Y=3, SIGMA_U=4, SIGMA_V=5, 
	                         // CONTRAST_MAX=6, CONTRAST_HIST=7, CONTRAST_SCALED=8
	int32 aggregateValue=-1; // FRAMES=0, SECONDS=1, MINUTES=2

	for (int i=1; i<argc; i++) {
		s=argv[i];
		for (uint32 k=0; k<s.size(); k++)
		  s[k]=toupper(s[k]);
		
		if (s=="-RECURSE") recurse=true;
		else {
			if (s=="-AGGREGATE") {
				aggregate=true;

				// read second parameter
				i++;
				if (i>=argc) usage();
				s=argv[i];
				for (uint32 k=0; k<s.size(); k++)
					s[k]=toupper(s[k]);
				if (s=="AVG_Y")                aggregateType=0;
				else if (s=="AVG_U")           aggregateType=1;
				else if (s=="AVG_V")           aggregateType=2;
				else if (s=="SIGMA_Y")         aggregateType=3;
				else if (s=="SIGMA_U")         aggregateType=4;
				else if (s=="SIGMA_V")         aggregateType=5;
				else if (s=="CONTRAST_MAX")    aggregateType=6;
				else if (s=="CONTRAST_HIST")   aggregateType=7;
				else if (s=="CONTRAST_SCALED") aggregateType=8;
				else usage();

				// read third parameter
				i++;
				if (i>=argc) usage();
				s=argv[i];
				for (uint32 k=0; k<s.size(); k++)
					s[k]=toupper(s[k]);
				if (s=="FRAMES")       aggregateValue=0;
				else if (s=="SECONDS") aggregateValue=1;
				else if (s=="MINUTES") aggregateValue=2;
				else usage();

			} else 
				if (i != argc-1) usage();
		} // else
	} // for

	cout << "analyse: " << filename << endl;
	cout << "parameters:" << endl;
	if (recurse) cout << "  -recurse directory" << endl;
	if (aggregate) {
		cout << "  -aggregate ";

		switch (aggregateType) {
		case 0: 
			cout << "AVG_Y based on ";
			break;
		case 1: 
			cout << "AVG_U based on ";
			break;
		case 2: 
			cout << "AVG_V based on ";
			break;
		case 3: 
			cout << "SIGMA_Y based on ";
			break;
		case 4: 
			cout << "SIGMA_U based on ";
			break;
		case 5: 
			cout << "SIGMA_V based on ";
			break;
		case 6: 
			cout << "CONTRAST_MAX based on ";
			break;
		case 7: 
			cout << "CONTRAST_HIST based on ";
			break;
		case 8: 
			cout << "CONTRAST_SCALED based on ";
			break;
		default:
			cout << "unknown option";
			exit(1);
		} // switch

		switch (aggregateValue) {
		case 0: 
			cout << "frames" << endl;
			break;
		case 1: 
			cout << "seconds" << endl;
			break;
		case 2: 
			cout << "minutes" << endl;
			break;
		default:
			cout << "unknown option";
			exit(1);
		} // switch
	} // if
	cout << endl;


	VideoStatisticData data;
	if (!recurse) {
		cout << "processing file: "<< filename << endl;
		VideoStatistic vs (filename);
 		vs.process(data);
		cout << data;

		if (aggregate) {
			VideoDecoder reader (filename);
			ieee64 framerate=reader.getFrameRate();
			printAction(data, framerate, aggregateType, aggregateValue);
		} // if

	} else {

		vector<FileInfo> mpgFiles;
		FileList fl(filename, true);    
		fl.process("*.mpg", mpgFiles);

		// open mpeg files
	    vector<FileInfo>::iterator it;
		for (it = mpgFiles.begin(); it != mpgFiles.end(); it++) {
			cout << endl  << "processing file: "<< (*it).name << endl;
			try {
				VideoStatistic vs ((*it).name);
				vs.process(data);
				cout << data << endl;

				if (aggregate) {
					VideoDecoder reader (filename);
					ieee64 framerate=reader.getFrameRate();
					printAction(data, framerate, aggregateType, aggregateValue);
				} // if

			}  catch (MocaException e) {
		        cout << "VideoStatistic " << e.what() << endl;
				exit (-1);
		    } // try
		} // for
		cout << endl << endl << mpgFiles.size() << " files found." << endl;
	} // else

	return 0;
} // main




