/*
   This file is part of the MoCA video library.
   Name: CssCreateData.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file



   \author    Stephan Kopf

              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim
              68131 Mannheim
              Germany
              email: kopf@informatik.uni-mannheim.de

   (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt

*/


// Moca includes
#include <DrawDot.h>
#include <Image.h>
#include <CssImage.h>
#include <FileList.h>
#include <IImageStream.h>
#include <ScaleFilter.h>
#include <ImageEdgePosition.h>
#include <ContourEdgeList.h>
#include <ContourParametrization.h>
#include <OImageStream.h>
#include <CssMaximaFeature.h>
#include <OTagStream.h>
#include <InverseFilter.h> 
#include <GrayscaleFilter.h> 
#include <CropFilter.h>
#include <RegionFeature.h>
//#include <ContourEdgeList.h>
#include <BinarizeFilter.h>
#include <InverseFilter.h>
#include <CssDatabase.h>


const uint8 BACKGROUND_COLOR = 255;
const uint8 FOREGROUND_COLOR = 0;

/*
	read all "*segmented*" files in directory,
	calculate contour, save contour
	write css image and css data in tag file  
*/


void usage()
{
	cerr << "CssCreateData" << endl;
	cerr << "version 1.0.3 (c) 2005 University of Mannheim" << endl;
	cerr << endl;
	cerr << "Usage: CssCreateData  directory fileExtension [options]" << endl;
	cerr << "   directory       Directory of the images, which should be transformed to CSS" << endl;
	cerr << "                   images. A white or black background is required." << endl;
	cerr << "   fileExtension   Analyse all files with this file extension," << endl;
	cerr << "                   e. g. \"*.jpg\" analyses all files with the extension jpg" << endl;
	cerr << "options:" << endl;
	
	cerr << "   -extended        apply the extended css algorithm (mirror contour) " << endl;
	cerr << "   -crop left right top bottom" << endl;
	cerr << "                    Remove noise at the borders of the images. The default crop value is 1. " << endl;
	cerr << "                    Example: -crop 10 10 15 0 " << endl;
	cerr << "                    removes: 10 pixel from the left and right border," << endl;
	cerr << "                             15 pixel from the top border, and" << endl;
	cerr << "                             no pixel from the bottom. " << endl;
	cerr << "   -headerfile db-filename   creates a header filename with all CSS objects. " << endl;
	cerr << "   -scale <int>              scale the original image by an integer value. " << endl;
	cerr << endl;
	cerr << "Example:" <<  endl;
	cerr << "CssCreateData D:/css/sequence/changick/  *segmented.jpg  -crop 0 0 10 10 -scale 4 " << endl;
	exit(1);
} // usage



void processFile(string &filename, int32 cropLeft, int32 cropRight, int32 cropTop, int32 cropBottom, int32 scale, bool extended)
{

	// variables
    string inputFile   = "";				// name of each input file
    string tmpFile     = "";				// temp filename
    string contourFile = "";				// output: filename of contour
    string cssFile     = "";				// output: filename of css image
    string tagFile     = "";				// output: filename of tag file

    string xContourFile= "";				// output: filename of extended contour
    string xCssFile    = "";				// output: filename of extended css image
    string xTagFile    = "";				// output: filename of extended tag file
    Image<uint8> img;
    Image <uint8> dest, destExt;
    ImageSize srcSize;

	
	// print filename
	cout << "process file: " << filename << endl;

	
	// define filenames
	inputFile = filename;
	tmpFile = inputFile;
	int p = tmpFile.find_last_of('/') +1;
	tmpFile.insert(p,"result/");
	p = tmpFile.find_last_of('/') +1;
	string path = tmpFile.substr (0,p);
	string fn = tmpFile.substr (p);
	
	// set contour filename
	string contourPath = path + "contour/";
	contourFile = contourPath + fn;
	p = contourFile.find_last_of('.');
	contourFile.insert(p,"_CONTOUR");
	p = contourFile.find_last_of('.') +1;
	contourFile.replace(p, 3, "pgm");


	// set inverted /extended) contour filename
	string xContourPath = path + "contourX/";
	xContourFile = xContourPath + fn;
	p = xContourFile.find_last_of('.');
	xContourFile.insert(p,"_CONTOURX");
	p = xContourFile.find_last_of('.') +1;
	xContourFile.replace(p, 3, "pgm");

	// set css filename
	string cssPath = path + "css/";
	cssFile = cssPath + fn;
	p = cssFile.find_last_of('.');
	cssFile.insert(p,"_CSS");
	p = cssFile.find_last_of('.') +1;
	cssFile.replace(p, 3, "pgm");

	// set extended css filename
	string xCssPath = path + "cssX/";
	xCssFile = xCssPath + fn;
	p = xCssFile.find_last_of('.');
	xCssFile.insert(p,"_CSSX");
	p = xCssFile.find_last_of('.') +1;
	xCssFile.replace(p, 3, "pgm");


	// set css tag filename
	tagFile = cssFile;
	p = tagFile.find_last_of('.') +1;
	tagFile.replace(p, 3, "tag");

	// set css tag filename
	xTagFile = xCssFile;
	p = xTagFile.find_last_of('.') +1;
	xTagFile.replace(p, 3, "xtag");
	

	// create path, if it does not exist
	FileHandling::createDirectory (path,false);
	FileHandling::createDirectory (contourPath,false);
	FileHandling::createDirectory (cssPath,false);
	if (extended) {
		FileHandling::createDirectory (xContourPath,false);
		FileHandling::createDirectory (xCssPath,false);
	} // if



	
	// read image
	IImageStream<uint8> iis(inputFile);
	iis.read(img);

	// resize input image
	if (scale>1) {
		ScaleFilter<uint8> sf;
		sf.expand(img, scale, scale);
	} // if

 	srcSize = img.getSize();


	// convert image to grayscale and binarize image
	{
		GrayscaleFilter<uint8> gray;
		gray.process(img);
	}


	// get the background color of the image (this should be 0 or 255)
	uint8 backgroundColor=0;
	uint8 foregroundColor=255;
	{
		Signal<int32> colors (SignalSize(256));
		colors.clear(0);
		for (int32 y=0; y<srcSize[DIM_Y]; y++) {
			colors(img(0,y))++;
			colors(img(srcSize[DIM_X]-1,y))++;
		} // for
		for (int32 x=0; x<srcSize[DIM_X]; x++) {
			colors(img(x,0))++;
			colors(img(x, srcSize[DIM_Y]-1))++;
		} // for
		int32 maxCountPos=0;
		for (int32 i=1; i<256; i++) {
			if (colors(maxCountPos)<colors(i)) maxCountPos=i;
		} // for
		if (maxCountPos<0) maxCountPos=0;
		if (maxCountPos>255) maxCountPos=255;
		backgroundColor=(uint8)maxCountPos;
		if (backgroundColor == 255) {
			foregroundColor = 0;
		} else {
			foregroundColor = 255;
		} // else
	}

#if 0
	// debug
	{
		WriteImage (img, "D:/obj01.png");
	}
#endif

	
	// set border of image to background color (use crop values)
	{
		if ((cropLeft!=0) || (cropRight!=0) || (cropTop!=0) || (cropBottom!=0)) {
			for (int32 y=0; y<srcSize[DIM_Y]; y++) {
				for (int32 x=0; x<cropLeft; x++) {
					img(x,y) = backgroundColor;
				} // for
				for (int32 x=srcSize[DIM_X]-cropRight; x<srcSize[DIM_X]; x++) {
					img(x,y) = backgroundColor;
				} // for
			} // for
			for (int32 x=0; x<srcSize[DIM_X]; x++) {
				for (int32 y=0; y<cropTop; y++) {
					img(x,y) = backgroundColor;
				} // for
				for (int32 y=srcSize[DIM_Y]-cropBottom; y<srcSize[DIM_Y]; y++) {
					img(x,y) = backgroundColor;
				} // for
			} // for
		} // if
	}

	// enlarge image with border
	srcSize[DIM_X] += 6;
	srcSize[DIM_Y] += 6;
	srcSize[DIM_BAND] = 1;
	{
		Image<uint8> tmp (srcSize);
		tmp.clear(backgroundColor);
		SampleBlockPosition p (3,3,0,0);
		tmp.copy (img, p);
		img = tmp;
	}


#if 0
	// debug
	WriteImage (img, "D:/obj02.png");
	exit(1);
#endif


	// set all pixels to foregroundColor, that are not background pixels
	for (int32 y=0; y<srcSize[DIM_Y]; y++) {
		for (int32 x=0; x<srcSize[DIM_X]; x++) {
			if (img(x,y) != backgroundColor) {
				img(x,y) = foregroundColor;
			} // if
		} // for
	} // for

#if 0
	// debug
	WriteImage (img, "D:/obj02.png");
#endif


	// locate largest object (not background)
	{
		RegionFeature<uint8> rf;
		Image<uint32> regions;
		Image<uint8> objectImage;
		bool objectFound;
		rf.process(img, objectImage, objectFound);
		img = objectImage;
	}

#if 0
	// debug
	WriteImage (img, "D:/obj03.png");
#endif
	


	// invert image, if background color of object does not match CSS_BACKGROUND_COLOR
	{
		backgroundColor = img(0,0);
		if (backgroundColor != BACKGROUND_COLOR) {
			InverseFilter invF;
			invF.process(img);
			backgroundColor = img(0,0);
		} // if
	}
	
#if 0
	// debug
	WriteImage (img, "D:/obj03.png");
	exit(1);
#endif
	
		
	// get, parametrize and save contour
	vector<ImageEdgePosition> iepList;
	ContourEdgeList cel(BACKGROUND_COLOR);
	cel.process (img, iepList);
	
	vector<ImagePosition> ipList;
	ContourParametrization cp(ContourParametrization::DEFAULT_CONTOUR_SIZE);
	cp.process(iepList, ipList);

#if 0
	// debug: add noise and save contour
	{
		const int v = 3;
		srand( (unsigned)time (NULL));
		for (int32 i=0; i<ipList.size(); i++) {
			ipList[i][DIM_X] += (rand()%v) - (rand()%v);
			ipList[i][DIM_Y] += (rand()%v) - (rand()%v);
		} // for
		
		// write contour
		Image<uint8> tmp = img;
		cp.drawContourLines (ipList, tmp, BACKGROUND_COLOR);
		WriteImage (tmp, "D:/noise.png");
		exit(1);
	}
#endif
	

	// write contour
	cp.drawContourLines (ipList, img, BACKGROUND_COLOR);
	//cp.drawContourDots (ipList, img, BACKGROUND_COLOR);
	WriteImage (img, contourFile);
#if 0
	// set image to background color
	for (int32 x=0; x<srcSize[DIM_X]; x++) {
		for (int32 y=0; y<srcSize[DIM_Y]; y++) {
			img(x,y) = BACKGROUND_COLOR;
		} // for
	} // for
	vector<ImagePosition>::iterator ipIt;
	DrawDot dot;
	for (ipIt= ipList.begin(); ipIt != ipList.end(); ipIt++) {
		dot.process((*ipIt), Color(FOREGROUND_COLOR), img);
	} // for
	WriteImage (img, contourFile);
#endif

	// calculate and save css-image
	CssImage css;
	if (extended) {
		Image<uint8> debugContour;
		css.process(ipList, dest, destExt, debugContour);
		WriteImage (debugContour, xContourFile);
	} else {
		css.process(ipList, dest);
	} // else


#if 0
	// debug
	// save modified CSS image (print horizontal lines, scale CSS, ...)

	// store normal css image
	{
		cerr << "DEBUG: CssCreateData: processFile()" << endl;
		// initialize images
		ImageSize orgCssSize = dest.getSize();
		ImageSize saveCssSize = orgCssSize;
		saveCssSize[DIM_X]++;
		
		// scaling factor
		int scaleFactor=3;
		
		// horizontal lines in CSS image
		int32 lp1=0;  
		int32 lp2=100;
		int32 lp3=10000;
		int32 lp4=10000;
		int32 lp5=10000; // 10000 = no line
		
		// initialize new image
		Image<uint8> saveCss (saveCssSize);
		for (int32 y=0; y<saveCssSize[DIM_Y]; y++) {
			for (int32 x=0; x<saveCssSize[DIM_X]; x++) {
				saveCss(x,y) = 255;
			} // for
		} // for
		
		// copy css to image
		for (int32 y=0; y<orgCssSize[DIM_Y]; y++) {
			for (int32 x=0; x<orgCssSize[DIM_X]; x++) {
				saveCss(x+1,y-1) = dest(x,y);
			} // for
		} // for
		
		// add horizontal lines to CSS image
		for (int32 y=0; y<saveCssSize[DIM_Y]; y++) {
			if ((y==saveCssSize[DIM_Y]-1-lp1) || 
				(y==saveCssSize[DIM_Y]-1-lp2) || 
				(y==saveCssSize[DIM_Y]-1-lp3) || 
				(y==saveCssSize[DIM_Y]-1-lp4) || 
				(y==saveCssSize[DIM_Y]-1-lp5)) {
				for (int32 x=0; x<saveCssSize[DIM_X]; x++) {
					saveCss(x,y) = 0;
				} // for
			} // if
		} // for
		
		// add vertical lines to CSS image
		for (int32 y=0; y<saveCssSize[DIM_Y]; y++) {
			saveCss(0,y) = 0;
		} // for
		
		
		// scale
		if (scaleFactor>1) {
			ImageSize tmpSize(saveCssSize);
			tmpSize[DIM_Y] = (tmpSize[DIM_Y]+scaleFactor-1) / scaleFactor +1;
			
			// initialize tmp image
			Image<uint8> tmp (tmpSize);
			for (int32 y=0; y<tmpSize[DIM_Y]; y++) {
				for (int32 x=0; x<tmpSize[DIM_X]; x++) {
					tmp(x,y) = 255;
				} // for
			} // for
			
			// copy saveCss to tmp
			for (int32 y=0; y<saveCssSize[DIM_Y]; y++) {
				for (int32 x=0; x<saveCssSize[DIM_X]; x++) {
					if (saveCss(x,y)!=255) {
						int32 py=y/scaleFactor;
						if (tmp(x,py)>saveCss(x,y)) tmp(x,py)=saveCss(x,y);
					} // if
				} // for
			} // for
			saveCss=tmp;
		} // if
		
		cerr << "store modified CSS image to D:/modifiedCSS.png" << endl;
		{
			OImageStream<uint8> ois("D:/css.png");
			ois.write(saveCss);
		}
	}
	
	// store extended (mapped) css image
	if (extended) {
		dest = destExt;
		
		// initialize images
		ImageSize orgCssSize = dest.getSize();
		ImageSize saveCssSize = orgCssSize;
		saveCssSize[DIM_X]++;
		
		// scaling factor
		int scaleFactor=3;
		
		// horizontal lines in CSS image
		int32 lp1=0;  
		int32 lp2=10000;
		int32 lp3=10000;
		int32 lp4=10000;
		int32 lp5=10000; // 10000 = no line
		
		// initialize new image
		Image<uint8> saveCss (saveCssSize);
		for (int32 y=0; y<saveCssSize[DIM_Y]; y++) {
			for (int32 x=0; x<saveCssSize[DIM_X]; x++) {
				saveCss(x,y) = 255;
			} // for
		} // for
		
		// copy css to image
		for (int32 y=0; y<orgCssSize[DIM_Y]; y++) {
			for (int32 x=0; x<orgCssSize[DIM_X]; x++) {
				saveCss(x+1,y-1) = dest(x,y);
			} // for
		} // for
		
		// add horizontal lines to CSS image
		for (int32 y=0; y<saveCssSize[DIM_Y]; y++) {
			if ((y==saveCssSize[DIM_Y]-1-lp1) || 
				(y==saveCssSize[DIM_Y]-1-lp2) || 
				(y==saveCssSize[DIM_Y]-1-lp3) || 
				(y==saveCssSize[DIM_Y]-1-lp4) || 
				(y==saveCssSize[DIM_Y]-1-lp5)) {
				for (int32 x=0; x<saveCssSize[DIM_X]; x++) {
					saveCss(x,y) = 0;
				} // for
			} // if
		} // for
		
		// add vertical lines to CSS image
		for (int32 y=0; y<saveCssSize[DIM_Y]; y++) {
			saveCss(0,y) = 0;
		} // for
		
		
		// scale
		if (scaleFactor>1) {
			ImageSize tmpSize(saveCssSize);
			tmpSize[DIM_Y] = (tmpSize[DIM_Y]+scaleFactor-1) / scaleFactor +1;
			
			// initialize tmp image
			Image<uint8> tmp (tmpSize);
			for (int32 y=0; y<tmpSize[DIM_Y]; y++) {
				for (int32 x=0; x<tmpSize[DIM_X]; x++) {
					tmp(x,y) = 255;
				} // for
			} // for
			
			// copy saveCss to tmp
			for (int32 y=0; y<saveCssSize[DIM_Y]; y++) {
				for (int32 x=0; x<saveCssSize[DIM_X]; x++) {
					if (saveCss(x,y)!=255) {
						int32 py=y/scaleFactor;
						if (tmp(x,py)>saveCss(x,y)) tmp(x,py)=saveCss(x,y);
					} // if
				} // for
			} // for
			saveCss=tmp;
		} // if
		
		cerr << "store modified CSS image to D:/modifiedCSS.png" << endl;
		{
			OImageStream<uint8> ois("D:/mapped.png");
			ois.write(saveCss);
		}
	}
	exit(1);
#endif


	// store css image
	WriteImage (dest, cssFile);
	if (extended) {
		WriteImage (destExt, xCssFile);
	} // if

	// get css maxima
	{
		CssMaximumData cmList;
		CssMaximaFeature cmf;
		cmf.process (dest, cmList, ipList);
		
		OTagStream ots (tagFile);
		ots << cmList;		
		ots.close();
	}

	// get extended css maxima
	if (extended) {
		CssMaximumData cmList;
		CssMaximaFeature cmf;
		cmf.process (destExt, cmList, ipList);
		
		OTagStream ots (xTagFile);
		ots << cmList;		
		ots.close();
	} // if

	return;
}

void parse (int argc, char* argv[], string &srcPath, string &fileExt, 
			int &cropLeft, int &cropRight, int &cropTop, int &cropBottom, 
			string &fnHeader, int &scale, bool &extended)
{
	if (argc < 3) usage();
	srcPath = argv [1];
	fileExt = argv [2];

	for (int32 i=3; i<argc; i++) {
		string s=argv[i];
		FileHandling::upper(s);
		
		if (s == "-CROP") {
			if (i+4>=argc) usage();
			cropLeft  = atoi(argv[i+1]);
			cropRight = atoi(argv[i+2]);
			cropTop   = atoi(argv[i+3]);
			cropBottom= atoi(argv[i+4]);
			i+=4;
		} else if (s == "-SCALE") {
			if (i+1>=argc) usage();
			scale = atoi(argv[i+1]);
			i+=1;
		} else if (s=="-HEADERFILE") {
			if (i+1>=argc) usage();
			fnHeader = argv[i+1];
			i+=1;
		} else if (s=="-EXTENDED") {
			if (i>=argc) usage();
			extended = true;
		} else {
			usage();
		} // else
	} // for


	return;
} // parse

int main(int argc, char* argv[])
{   
	
	// read srcPath
    string srcPath="";	// path for pgm-images
	string fileExt="";  // input files, e. g. "*segmented.png"
	int cropLeft  = 0;	// crop border
	int cropRight = 0;
	int cropTop   = 0;
	int cropBottom= 0;
	string fnHeader=""; // filename of a header file
	int scale = 1;		// scale input image
	bool extended = false;
	parse (argc, argv, srcPath, fileExt, cropLeft, cropRight, cropTop, cropBottom, fnHeader, scale, extended);
	if (fnHeader.size() > 0) {
		cerr << "create header file: " << fnHeader << endl;
	} // if
	

	// real all "*segmented.png"-files
    cout << "read files..." << endl;
    vector<FileInfo> fileList;
    FileList nfl(srcPath, true);    
	nfl.process(fileExt, fileList);
    //nfl.show(fileList);
    cout << fileList.size() << " files found" << endl;
    

    // create css image for each file in list
    cout << "calculate css data..." << endl;
    int counter=0;					// number of iterations
    vector<FileInfo>::iterator it;
    for (it = fileList.begin(); it != fileList.end(); it++) {
        counter++;

		processFile((*it).name, cropLeft, cropRight, cropTop, cropBottom, scale, extended);

	} // for        

    cout << endl << srcPath << " ("<< counter << " objects)" << endl;

    if (fnHeader.size() > 0) {
	CssDatabaseSchemaType type = ECHO_DB_TYPE;
	CssDatabase css (type);
	css.createDatabase (fnHeader, srcPath);
    } // if

    return 0;
} // main

