/*
   This file is part of the MoCA video library.
   Name: GeometricFeatureCreateData.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file



   \author    Stephan Kopf

              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim
              68131 Mannheim
              Germany
              email: kopf@informatik.uni-mannheim.de

   (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt

*/


// Moca includes
#include <GeometricFeature.h>
#include <IImageStream.h>
#include <FileList.h>



const uint8 BACKGROUND_COLOR = 255;
const uint8 FOREGROUND_COLOR = 0;

/*
	read all "*segmented*" files in directory,
	calculate contour, save contour
	write css image and css data in tag file  
*/


void usage()
{
	cerr << "GeometricFeatureCreateData" << endl;
	cerr << "version 0.2.1 (c) 2005 University of Mannheim" << endl;
	cerr << endl;
	cerr << "Usage: GeometricFeatureCreateData  directory fileExtension [options]" << endl;
	cerr << "   directory       Directory of the images, which should be analysed." << endl;
	cerr << "                   A white or black background is required." << endl;
	cerr << "   fileExtension   Analyse all files with this file extension," << endl;
	cerr << "                   e. g. \"*.jpg\" analyses all files with the extension jpg" << endl;
	cerr << "options:" << endl;
	
	cerr << "   -crop left right top bottom" << endl;
	cerr << "                    Remove noise at the borders of the images. " << endl;
	cerr << "                    The default crop value is 4. " << endl;
	cerr << "                    Example: -crop 10 10 15 0 " << endl;
	cerr << "                    removes: 10 pixel from the left and right border," << endl;
	cerr << "                             15 pixel from the top border, and" << endl;
	cerr << "                             no pixel from the bottom. " << endl;
	cerr << endl;
	cerr << "Example:" <<  endl;
	cerr << "GeometricFeatureCreateData ./  img*.jpg  -crop 0 0 10 10 " << endl;
	exit(1);
} // usage



void processFile(string &filename, int32 cropLeft, int32 cropRight, int32 cropTop, int32 cropBottom)
{

	// variables
    string inputFile   = "";				// name of each input file
    string tmpFile     = "";				// temp filename
    string tagFile     = "";				// output: filename of tag (data) file
    string objectFile = "";				// output: filename of object

    Image<uint8> img;
    Image <uint8> dest;
    ImageSize srcSize;

	
	// print filename
	//cout << "process file: " << filename << endl;

	
	// define filenames
	inputFile = filename;
	tmpFile = inputFile;
	int p = tmpFile.find_last_of('/') +1;
	tmpFile.insert(p,"result/");
	p = tmpFile.find_last_of('/') +1;
	string path = tmpFile.substr (0,p);
	string fn = tmpFile.substr (p);
	

	// set data filename
	string geometricPath = path + "geometric/";
	tagFile = geometricPath + fn;
	p = tagFile.find_last_of('.');
	tagFile.insert(p,"_GEOMETRIC");
	p = tagFile.find_last_of('.') +1;
	tagFile.replace(p, 3, "tag");

	objectFile = tagFile;
	p = objectFile.find_last_of('.') +1;
	objectFile.replace(p, 3, "pgm");


	// create path, if it does not exist
	FileHandling::createDirectory (path,false);
	FileHandling::createDirectory (geometricPath,false);

	
	// read image
	ReadImage (inputFile, img);
 	srcSize = img.getSize();


	// convert image to grayscale and binarize image
	{
		GrayscaleFilter<uint8> gray;
		gray.process(img);
	}


	// get the background color of the image (this should be 0 or 255)
	uint8 backgroundColor=0;
	uint8 foregroundColor=255;
	{
		Signal<int32> colors (SignalSize(256));
		colors.clear(0);
		for (int32 y=0; y<srcSize[DIM_Y]; y++) {
			colors(img(0,y))++;
			colors(img(srcSize[DIM_X]-1,y))++;
		} // for
		for (int32 x=0; x<srcSize[DIM_X]; x++) {
			colors(img(x,0))++;
			colors(img(x, srcSize[DIM_Y]-1))++;
		} // for
		int32 maxCountPos=0;
		for (int32 i=1; i<256; i++) {
			if (colors(maxCountPos)<colors(i)) maxCountPos=i;
		} // for
		if (maxCountPos<0) maxCountPos=0;
		if (maxCountPos>255) maxCountPos=255;
		backgroundColor=(uint8)maxCountPos;
		if (backgroundColor == 255) {
			foregroundColor = 0;
		} else {
			foregroundColor = 255;
		} // else
	}


	
	// set border of image to background color (use crop values)
	{
		if ((cropLeft!=0) || (cropRight!=0) || (cropTop!=0) || (cropBottom!=0)) {
			for (int32 y=0; y<srcSize[DIM_Y]; y++) {
				for (int32 x=0; x<cropLeft; x++) {
					img(x,y) = backgroundColor;
				} // for
				for (int32 x=srcSize[DIM_X]-cropRight; x<srcSize[DIM_X]; x++) {
					img(x,y) = backgroundColor;
				} // for
			} // for
			for (int32 x=0; x<srcSize[DIM_X]; x++) {
				for (int32 y=0; y<cropTop; y++) {
					img(x,y) = backgroundColor;
				} // for
				for (int32 y=srcSize[DIM_Y]-cropBottom; y<srcSize[DIM_Y]; y++) {
					img(x,y) = backgroundColor;
				} // for
			} // for
		} // if
	}

	// enlarge image with border
	srcSize[DIM_X] += 6;
	srcSize[DIM_Y] += 6;
	srcSize[DIM_BAND] = 1;
	{
		Image<uint8> tmp (srcSize);
		tmp.clear(backgroundColor);
		SampleBlockPosition p (3,3,0,0);
		tmp.copy (img, p);
		img = tmp;
	}


	// set all pixels to foregroundColor, that are not background pixels
	for (int32 y=0; y<srcSize[DIM_Y]; y++) {
		for (int32 x=0; x<srcSize[DIM_X]; x++) {
			if (img(x,y) != backgroundColor) {
				img(x,y) = foregroundColor;
			} // if
		} // for
	} // for


	// locate largest object (not background)
	{
		RegionFeature<uint8> rf;
		Image<uint32> regions;
		Image<uint8> objectImage;
		bool objectFound;
		rf.process(img, objectImage, objectFound);
		img = objectImage;
	}



	// invert image, if background color of object does not match CSS_BACKGROUND_COLOR
	{
		backgroundColor = img(0,0);
		if (backgroundColor != BACKGROUND_COLOR) {
			InverseFilter invF;
			invF.process(img);
			backgroundColor = img(0,0);
		} // if
	}

	WriteImage (img, objectFile);

	int32 minX=10000;
	int32 maxX=-1;
	int32 minY=10000;
	int32 maxY=-1;
	int32 w=img.getSize()[DIM_X];
	int32 h=img.getSize()[DIM_Y];
	for (int32 y=0; y<h; y++) {
		for (int32 x=0; x<w; x++) {
			if (img(x,y) != BACKGROUND_COLOR) {
				if (minX>x) minX=x;
				if (minY>y) minY=y;
				if (maxX<x) maxX=x;
				if (maxY<y) maxY=y;
			} // if
		} // for
	} // for
	if (maxX-minX > 0.90*w) {
		cerr << filename << "   ...width?" << endl;
	} else if (maxY-minY > 0.90*h) {
		cerr << filename << "   ...height?" << endl;
	} else if (maxX-minX >0.70*w && maxY-minY > 0.70*h) {
		cerr << filename << "   ...size?" << endl;
	} // else


	
	// get geometric features
	GeometricFeatureData data;
	GeometricFeature gf;
	gf.process (img, data);

	// store data
	OTagStream ots (tagFile);
	ots << data;		
	ots.close();
		
	return;
}

void parse (int argc, char* argv[], string &srcPath, string &fileExt, 
			int &cropLeft, int &cropRight, int &cropTop, int &cropBottom)
{
	if (argc < 3) usage();
	srcPath = argv [1];
	fileExt = argv [2];

	for (int32 i=3; i<argc; i++) {
		string s=argv[i];
		FileHandling::upper(s);
		
		if (s == "-CROP") {
			if (i+4>=argc) usage();
			cropLeft  = atoi(argv[i+1]);
			cropRight = atoi(argv[i+2]);
			cropTop   = atoi(argv[i+3]);
			cropBottom= atoi(argv[i+4]);
			i+=4;
		} else {
			usage();
		} // else
	} // for


	return;
} // parse

int main(int argc, char* argv[])
{   
	
	// read srcPath
    string srcPath="";	// path for pgm-images
	string fileExt="";  // input files, e. g. "*segmented.png"
	int cropLeft  = -1;	// crop border
	int cropRight = -1;
	int cropTop   = -1;
	int cropBottom= -1;
	parse (argc, argv, srcPath, fileExt, cropLeft, cropRight, cropTop, cropBottom);
	if (cropLeft == -1) {
		cropLeft  = 4;
		cropRight = 4;
		cropTop   = 4;
		cropBottom= 4;
	} // if
	
 
	// real all "*segmented.png"-files
    cout << "read files..." << endl;
    vector<FileInfo> fileList;
    FileList fl(srcPath, true);    
	fl.process(fileExt, fileList);
    //fl.show(fileList);
    cout << fileList.size() << " files found" << endl;
    

    // create geometric data for each file in list
    cerr << "calculate geometric features..." << endl;
    for (int32 c=0; c<fileList.size(); c++) {
		processFile(fileList[c].name, cropLeft, cropRight, cropTop, cropBottom);
	} // for        
    cerr << endl << srcPath << " ("<< fileList.size() << " objects)" << endl;

    return 0;
} // main

