/*
   This file is part of the MoCA video library.
   Name: Png2Avi.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

  
	
	  \author    Stephan Kopf
	  
		Lehrstuhl fuer Praktische Informatik IV
		Universitaet Mannheim
		68131 Mannheim
		Germany
		email: kopf@informatik.uni-mannheim.de
		
		  (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
		  
*/

// C++ includes
#include <iomanip>
#include <sstream>
#include <iostream>
using namespace std;

// MoCA includes
#include<VideoIndex.h>
#include<VideoDecoder.h>
#include<OMpegVideo.h>
#include<VideoEncoder.h>
#include<IImageStream.h>
#include<CropFilter.h>


const string ext="png";

void show()
{
	string ext1=ext;
	string ext2=ext;
	FileHandling::upper(ext1);
	ext2[0] = ext1[0];

	cerr << "read all " << ext1 << " files in directory and store them as AVI video" << endl << endl;
	cerr << "Usage: " << endl;
	cerr << ext2 << "2Avi path  dest.avi [options] " << endl;
	cerr << endl;
	cerr << "options: " << endl;
	cerr << "  -fps rate                      set frame rate (default: 25.0)" << endl;
	cerr << "  -codec type                    set video codec (default: MPEG4)" << endl;
	cerr << "                                 codecs: WMV1, WMV2, MPEG1, MPEG2, MPEG4" << endl;
	cerr << "  -multiframe number             duplicate frames (default: 0)" << endl;
	cerr << "  -crop top bottom left right    crop borders (default: 0 0 0 0)" << endl;
	cerr << "  -autocrop                      crop to valid video size (default: no) " << endl;
	cerr << "  -bitrate <number>              valid bitrates: 4..16000 (in K)" << endl;

	cerr << endl;
	cerr << "example: "<< ext2 << "2Avi ./ dest.avi -codec MPEG1 -fps 15.0 -multiframe 2" << endl;
	exit(-1);
	return;
}

void parseParameters (int argc, char* argv[], EncoderCodecs &codec, ieee64 &framerate, int32 &bitrate, 
					  int &multiframe, int &left, int &right, int &top, int &bottom, bool &autocrop)
{
	left   = 0;
	right  = 0;
	top    = 0;
	bottom = 0;
	bitrate = -1;
	autocrop = false;
	ieee64 fps = 25.0;
	int multi = 0;
	
	string codS = "MPEG4";
	for (int i=3; i<argc; i++) {
		string s=argv[i];
		for (uint32 k=0; k<s.size(); k++)
			s[k]=toupper(s[k]);
		
		if (s=="-CODEC") {
			i++;
			s=argv[i];
			for (uint32 k=0; k<s.size(); k++)
				s[k]=toupper(s[k]);
			codS = s;
		} else if (s=="-FPS") {
			i++;
			s=argv[i];
			fps = atof (s.c_str());
		} else if (s=="-BITRATE") {
			i++;
			s=argv[i];
			bitrate = atoi (s.c_str());
		} else if (s=="-MULTIFRAME") {
			i++;
			s=argv[i];
			multi = atoi (s.c_str());
		} else if (s=="-AUTOCROP") {
			autocrop = true;
		} else if (s=="-CROP") {
			i++; s=argv[i];
			top = atoi (s.c_str());
			i++; s=argv[i];
			bottom = atoi (s.c_str());
			i++; s=argv[i];
			left = atoi (s.c_str());
			i++; s=argv[i];
			right = atoi (s.c_str());
		} else {
			cerr << "unknown option!" << endl;
			show();
		} // else
	} // for
	
	
	// set codec for encoder
	EncoderCodecs cod = CODEC_MPEG4;
	if (codS != "") {
		if (codS == "WMV1")
			cod = CODEC_WMV_1;
		
		else if (codS == "WMV2")
			cod = CODEC_WMV_2;
		
		else if (codS == "MPEG1")
			cod = CODEC_MPEG1;
		
		else if (codS == "MPEG2")
			cod = CODEC_MPEG2;
		
		else if (codS == "MPEG4")
			cod = CODEC_MPEG4;
		
		else {
			cerr << "unknown video codec!" << endl;
			show();
			exit(1);
		} // else
	} // if
	
	
	codec = cod;
	framerate = fps;
	multiframe = multi;
	return;
}

int main(int argc, char* argv[])
{
	
    try {
		
		if (argc < 3) {
			show();
		} // if
		string srcPath = argv[1];
		string destFile = argv[2];
		
		EncoderCodecs codec; 
		ieee64 framerate;
		int32 bitrate;
		int multiframe;
		int left,right, top, bottom;
		bool autocrop;
		parseParameters (argc, argv, codec, framerate,bitrate, multiframe, left,right, top, bottom, autocrop);
		if (multiframe<0) multiframe=0;
		if (bitrate<4) bitrate=-1;
		if (bitrate>16000) bitrate=-1;
		
		// read all files in directory
		vector<FileInfo> files;
		try {
			FileList fl(srcPath, false);    
			string fe = "*."; fe += ext;
			fl.process(fe, files);
			fl.sort(files);
		} catch (MocaException e) {
			cout << "Exception in FileList()  " << e.what() << endl;
			exit (-1);
		} // try
		
		
		if (files.size()==0) {
			cerr << "no files found in directory!" << endl;
			exit(-1);
		} // if
		
		
		// check autocrop
		if (autocrop) {
			Image<uint8> img;
			ImageSize tmpSize;
			CropFilter<uint8> crop (left,right,top,bottom);
			vector<FileInfo>::iterator it = files.begin();
			try {
				// open image
				ReadImage ((*it).name, img);
				crop.process(img);
				tmpSize = img.getSize();
			} catch (MocaException e) {
				cout << "Exception in ReadFirstImage()  " << e.what() << endl;
				exit (-1);
			} // try
			int dx = tmpSize[DIM_X] - ((tmpSize[DIM_X] / 16) * 16);
			int dy = tmpSize[DIM_Y] - ((tmpSize[DIM_Y] / 16) * 16);
			right  += dx;
			bottom += dy;
		} // if


		// get image size
		CropFilter<uint8> crop (left,right,top,bottom);
		Image<uint8> img;
		ImageSize size;
		vector<FileInfo>::iterator it = files.begin();
		try {
			// open image
			ReadImage ((*it).name, img);
			crop.process(img);
			size = img.getSize();
		} catch (MocaException e) {
			cout << "Exception in ReadFirstImage()  " << e.what() << endl;
			exit (-1);
		} // try
		
		
		
		try {
			// open writer
			VideoEncoderOptions encoderOpt;
			encoderOpt.pathFilename = destFile;
			encoderOpt.encCodec = codec;
			if (bitrate!=-1) {
				encoderOpt.bitrate=bitrate;
			} // if
			encoderOpt.silent=false;
			encoderOpt.framerate=framerate;
			encoderOpt.imgCodec = CODEC_RGB;
			encoderOpt.size = size;
			VideoEncoder writer(encoderOpt);
			
			
			// save files
			for (it = files.begin(); it != files.end(); it++) {
				
				// open image
				try {
					ReadImage ((*it).name, img);
					crop.process(img);
				} catch (MocaException e) {
					cout << "Exception in ReadImage()  " << e.what() << endl;
					exit (-1);
				} // try
				
				
				
				// write image
				try {
					for (int32 i=0; i<=multiframe; i++) {
						writer.write(img);
					} // for
				} catch (MocaException e) {
					cout << "Exception in ReadImage()  " << e.what() << endl;
					exit (-1);
				} // try
				cerr << "process image: " << (*it).name << endl;
			} // for
			cerr << endl;
			writer.close();
			cerr << "video stored!" << endl<<endl;
		} catch (MocaException e) {
			cout << "Exception in OpenEncoder()   " << e.what() << endl;
			exit (-1);
		} // try
		
		
    } catch (MocaException e) {
		string ext1=ext;
		string ext2=ext;
		FileHandling::upper(ext1);
		ext2[0] = ext1[0];
        cout << "Exception in " <<ext2 << "2Avi " << e.what() << endl;
        exit (-1);
    } // try
	
	return 0;
} // main

