/*
   This file is part of the MoCA video library.
   Name: CameraBackgroundFeature.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/

// this class analyses the frames in a video and creates a background mosaic image of 
// the frames
#ifndef CAMERA_BACKGROUND_FEATURE_H
#define CAMERA_BACKGROUND_FEATURE_H


#define DEBUG_CAMERA_BACKGROUND_FEATURE
//#undef DEBUG_CAMERA_BACKGROUND_FEATURE


// C++ includes
#include <iostream>
using namespace std;


// MoCA includes
#include <Image.h>
#include <Video.h>
#include <CameraTransformation.h>
#include <CornerDetection.h>
#include <ClipFilter.h>
#include <CameraModel.h>
#include <CameraTransformationPrecise.h>
#include <DifferenceFilter.h>
#include <MarkovFilter.h>
#include <DrawRectangle.h>
#include <OpeningFilter.h>
#include <ClosingFilter.h>
#include <OYuvVideo.h>
#include <MedianFilter.h>



/*
	Constructor: initializes variables

	process(): 
	constructs the background image
*/
class CameraBackgroundFeature
{
public:
	
	CameraBackgroundFeature(const VideoSize &srcVideoSize, 
				int32 bgSizeX, int32 bgSizeY, int32 bgShiftX, int32 bgShiftY, 
				bool calculateObjects, bool precise, bool estimateObjectLocation, 
				DefaultObjectPosition objPos=POSITION_UNKNOWN);
	~CameraBackgroundFeature() {};
	

	// parameters:
	//    vid          : source video
	//    motionMaskVid: mask of the video with objects (object is white)
	//    bgImg        : background image
	void process (const Video<uint8> &vid, Video<uint8> &motionMaskVid, Image<uint8> &bgImg, vector<TransformationData> &vidTrans);

	// store images and object mask video
	void storeData (const Video<uint8> &motionMaskVid, const Image<uint8> &bgImg, string &fnPrefix);
	void addObject (const Image<uint8> &img, const Image<uint8> &mask, Image<uint8> &bgImg, TransformationData &trans, string &fnPrefix);



	// calculates the median of a video, 
	// if the pixel==255 (background), the pixel is ignored
	void bgMedian(int32 lastFrameToAnalyse, Video<uint8> &bgVid, Image<uint8> &backgrd);

private:
	void calcBackground (Video<uint8> &vid, Video<uint8> &bgVid, Image<uint8> &bgImg, int32 bgSizeX, int32 bgSizeY, int32 &bgShiftX, int32 &bgShiftY);
	void findBgVidBorders (const Video<uint8> &bgVid, int32 &minX, int32 &maxX, int32 &minY, int32 &maxY);
	void findBgImgBorders (Image<uint8> &bgImg, int32 &minX, int32 &maxX, int32 &minY, int32 &maxY);

	bool getTransformation (const Video<uint8> &vid, Video<uint8> &tmpBgVid, vector<TransformationData> &vidTrans);
	bool cropBackground (const Video<uint8> &tmpBgVid, Video<uint8> &bBgVid, vector<TransformationData> &vidTrans);
	bool getBackgroundImage (const Video<uint8> &bgVid, Video<uint8> &maskVid, Image<uint8> &newBgImg, bool estimateObjectLocation);
#if 0
	bool getPreciseTransformation (const Video<uint8> &vid, Video<uint8> &maskVid, 
		                           Video<uint8> &tmpBgVid, Image<uint8> &bgImg, vector<TransformationData> &vidTrans);
#endif
	void calcObjects (const Video<uint8> &vid, const Video<uint8> &bgVid, const Image<uint8> &bgImg, const vector<TransformationData> &vidTrans, Video<uint8> &objMaskVid);
	
	void getCorners (const Image<uint8> &img, vector<CornerPosition>& cornersFloat, vector<CornerPositionInt>& cornersInt);

	// variables
	int32 bgSizeX;
	int32 bgSizeY;
	int32 bgShiftX;
	int32 bgShiftY;

	VideoSize srcVidSize;
	ImageSize srcImgSize;

	bool calculateObjects;
	bool precise;
	bool estimateObjectLocation;

	// use additional knowledge about object
	DefaultObjectPosition defaultObjPos;
	int32 defaultPx0;
	int32 defaultPx1;
	int32 defaultPy0;
	int32 defaultPy1;
};

#endif
