/*
   This file is part of the MoCA video library.
   Name: CameraModel.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef CAMERA_MODEL_H
#define CAMERA_MODEL_H

// C++ includes
#include <iomanip>
#include <sstream>
#include <iostream>
using namespace std;


// MoCA includes
#include <MocaEnum.h>
#include <CreateMocaPath.h>
#include <WriteMocaData.h>
#include <VideoDecoder.h>
#include <DrawLine.h>
#include <OImageStream.h>
#include <CornerDetection.h>
#include <ClipFilter.h>
#include <CropFilter.h>
#include <LinearEstimatorFeature.h>
#include <DrawCross.h>
#include <GrayscaleFilter.h>
#include <ImageStatistic.h>
#include <CameraTransformation.h>
#include <CameraTransformationPrecise.h>
#include <BlockMotionFeature.h>
 


// external C includes
extern "C" {
#include <SadFeature_mmx.h>
}


static const string CAMERA_MODEL_VERSION  = "Version-0.72";
static const string CAMERA_MODEL_FILENAME = "cameramodel.idx";

//const string WINDOWS_CAMERA_MODEL_EXECUTABLE = "cameramodel.exe";
//const string LINUX_CAMERA_MODEL_EXECUTABLE   = "/usr/local/bin/cameramodel";


// stores one match of corners
struct CameraModelVectors
{
	CameraModelVectors (ieee64 x0, ieee64 y0, ieee64 x1, ieee64 y1,
		                bool outlier, ieee64 error ) {
		this->x0 = x0;
		this->y0 = y0;
		this->x1 = x1;
		this->y1 = y1;

		this->outlier = outlier;
		matchingError = error;
	}

	// position of corners
	ieee64 x0, y0, x1, y1; 

	bool outlier;
	ieee64 matchingError;
};



struct CameraModelItem
{


	CameraModelItem () {
		clear();
	} 

	void clear()
	{
		frame = 0;
		numberCornerMatchesAvailable = 0;
		maxVectorLength  = 0;
		valid = 1;
		lastCorners = 0;
		currCorners = 0;
		numberCornerMatched = 0;

		transData.clear();
		storeDebug = false;
		return;
	}


	void activateDebug() 
	{
		storeDebug = true;
		return;
	}

    CameraModelItem& operator=(const CameraModelItem &other) {
		if (this != &other) {
			frame = other.frame;
			numberCornerMatchesAvailable = other.numberCornerMatchesAvailable;
			maxVectorLength = other.maxVectorLength;
			valid = other.valid;
			lastCorners = other.lastCorners;
			currCorners = other.currCorners;
			numberCornerMatched = other.numberCornerMatched;

			transData = other.transData;
			storeDebug = other.storeDebug;
		} // if
        return *this;
    } // operator=

	bool isValid () const
	{
		if (valid == 0) return true;
		return false;
	}


	// transform (oldX,oldY) to new position (newX/newY) 
	inline void getVectors(const ieee64 oldX, const ieee64 oldY, ieee64 &newX, ieee64 &newY)
	{
		//newX = (oldX*a00 + oldY*a01 + tx) / (oldX*px + oldY*py + 1.0);
		//newY = (oldX*a10 + oldY*a11 + ty) / (oldX*px + oldY*py + 1.0);
		newX = (oldX*transData.a00 + oldY*transData.a01 + transData.tx) / (oldX*transData.px + oldY*transData.py + 1.0);
		newY = (oldX*transData.a10 + oldY*transData.a11 + transData.ty) / (oldX*transData.px + oldY*transData.py + 1.0);
		return;
	}

	// use inverse transform to transform (oldX,oldY) to new position (newX/newY) 
	inline void getInverseVectors(const ieee64 oldX, const ieee64 oldY, ieee64 &newX, ieee64 &newY) const
	{
		TransformationData p;
		PerspectiveTransformation trans;
		TransformationData::processInverse (transData, p);

		newX = (oldX* p.a00 + oldY* p.a01 + p.tx) / (oldX* p.px + oldY* p.py + 1.0);
		newY = (oldX* p.a10 + oldY* p.a11 + p.ty) / (oldX* p.px + oldY* p.py + 1.0);
		return;
	}



	// apply transformation to an image
	void applyTransformation (Image<uint8> &srcDest)
	{
		PerspectiveTransformation trans;
		if (srcDest.getSize()[DIM_BAND] == 1) {
			trans.processImg (transData, srcDest);
		} else {
			trans.processColorImg (transData, srcDest);
		} // else
	}

	// apply transformation to an image
	void applyTransformation (const Image<uint8> &src, Image<uint8> &dest, bool linearInterpolation=true)
	{
		PerspectiveTransformation trans;
		if (src.getSize()[DIM_BAND] == 1) {
			trans.processImg (transData, src, dest, linearInterpolation);
		} else {
			trans.processColorImg (transData, src, dest, linearInterpolation);
		} // else
	}

	// transform (oldX,oldY) to new position (newX/newY) 
	inline void getMaxVectorLength (const int32 width, const int32 height, ieee64 &length)
	{
		length = 0.0;
		ieee64 oldX = 0.0;
		ieee64 oldY = 0.0;
		ieee64 newX, newY;
		getVectors(oldX, oldY, newX, newY);
		ieee64 dx = (newX-oldX);
		ieee64 dy = (newY-oldY);
		ieee64 d = dx*dx + dy*dy;
		if (length<d) length = d;
		
		oldY = width-1;
		getVectors(oldX, oldY, newX, newY);
		dx = (newX-oldX);
		dy = (newY-oldY);
		d = dx*dx + dy*dy;
		if (length<d) length = d;
		
		oldX = height-1;
		getVectors(oldX, oldY, newX, newY);
		dx = (newX-oldX);
		dy = (newY-oldY);
		d = dx*dx + dy*dy;
		if (length<d) length = d;

		oldY = 0.0;
		getVectors(oldX, oldY, newX, newY);
		dx = (newX-oldX);
		dy = (newY-oldY);
		d = dx*dx + dy*dy;
		if (length<d) length = d;
		if (length>0) {
			length = sqrt (length);
		} // if
		return;
	}

	void drawVector (ieee64 px, ieee64 py, Image<uint8> &img, Color c)
	{
		ieee64 nx, ny;
		getInverseVectors (px, py, nx, ny);
		ImagePosition p0((int32)px,(int32)py);
		ImagePosition p1((int32)nx,(int32)ny);

		DrawLine line (c);
		line.process (p0,p1,img);
		return;
	}


	void drawVectorImage (Image<uint8> &img)
	{
		int32 nx=img.getSize()[DIM_X];
		int32 ny=img.getSize()[DIM_Y];
		for (int32 y=8; y<ny; y+=16) {
			for (int32 x=8; x<nx; x+=16) {
				drawVector (x, y, img, Color::WHITE);
			} // for
		} // for
		return;
	}




	int32 frame;	                    // frame number
	int32 numberCornerMatchesAvailable; // number of possible matches (candidates for matches, correlation)
	int32 lastCorners;					// number of corners in last frame
	int32 numberCornerMatched;			// number of matched corners (real matches)

	int32 currCorners;      // number of corners in current frame
	ieee64 maxVectorLength;  // maximum length of motion vector with parameters specified in transData
	int32 valid;    
		// 0=valid, 
		// 1=error during decoding, 
		// 2=no corners detected (no camera model available)
	    // 3=no correlations detected (no camera model available)
	    // 4=parameters calculated, but the motion vectors do not make sense (e.g. large object visible)
	    // 5=parameters calculated, but the motion vectors of 2 following frames are very different
	    // 6=parameters calculated, but the motion vectors of 2 following frames are different (may not be an error)


	TransformationData transData;
	//ieee64 a00, a01, a10, a11;   // affine transformation
	//ieee64 tx, ty;               // pan, tilt (horizontal and vertical movement)
	//ieee64 px, py;               // perscpective parameters
	//
	//        a00*x + a01*y + tx           a10*x + a11*y + ty 
	//  x' = --------------------    y' = --------------------  
	//        px*x + py*y + 1              px*x + py*y + 1
	//

	// store corners, outliers, etc
	bool storeDebug;
	vector<CameraModelVectors> vectorData;
};

std::ostream& operator<< (ostream&,    const CameraModelItem&);
OTagStream&   operator<< (OTagStream&, const CameraModelItem&);
ITagStream&   operator>> (ITagStream&, CameraModelItem&);




struct CameraModelData
{
	CameraModelData () 
	{
		version    = CAMERA_MODEL_VERSION;
		filename   = "";
		cropleft   = 8;
		cropright  = 8;
		croptop    = 8;
		cropbottom = 8;
		data.clear();
	} 
    CameraModelData& operator=(const CameraModelData &other) {
		if (this != &other) {
			version    = other.version;
			filename   = other.filename;
			cropleft   = other.cropleft;
			cropright  = other.cropright;
			croptop    = other.croptop;
			cropbottom = other.cropbottom;
			data       = other.data;
		} // if
        return *this;
    } // operator=
	
	string version;
	string filename;
	int32 cropleft;
	int32 cropright;
	int32 croptop;
	int32 cropbottom;

	vector<CameraModelItem> data;
};

std::ostream& operator<< (ostream&,    const CameraModelData&);
OTagStream&   operator<< (OTagStream&, const CameraModelData&);
ITagStream&   operator>> (ITagStream&, CameraModelData&);


// required for matching corners
struct matchCorners
{
  int16 last,curr;
  int32 score;
};

int 
inline compMatchCorners (const void *m1, const void * m2)
{
  matchCorners* ma1 = (matchCorners*)m1;
  matchCorners* ma2 = (matchCorners*)m2;

  return ma1->score - ma2->score;
}




/*
Constructor: initializes variables

	process(): 
	if status file does not exist:
	calculate the camera model parameters from mpeg-file, save status file
	else:
	read camera operation from status file
*/
class CameraModel
{
public:
	

	// analyse two images 
	CameraModel(bool normalizeImages=false, bool noCrop=false, DefaultObjectPosition objPos=POSITION_UNKNOWN,
		        int32 defaultMaxMotionVectorLength = 30, bool calcTranslationOnly=false)
	{
		calcTranslation=calcTranslationOnly;
		defaultObjPos = objPos;
		defaultPx0 = -1;
		defaultPx1 = -1;
		defaultPy0 = -1;
		defaultPy1 = -1;
		defaultMotionVectorLength = defaultMaxMotionVectorLength;
		useDefaultMotion = false;

		normalize = normalizeImages;
		pathFilename = "";
		path = "";

		cropleft=8;
		cropright=8;
		croptop=8;
		cropbottom=8;
		if (noCrop) {
			cropleft=0;
			cropright=0;
			croptop=0;
			cropbottom=0;
		}
		processShot = true;
	}

	// analyse AVI-/MPEG-video
	CameraModel(const string &pathFilename, int32 cropleft=8, int32 cropright=8, int32 croptop=8, int32 cropbottom=8,
			    int32 defaultMaxMotionVectorLength = 30, bool calcTranslationOnly=false);
	~CameraModel() {};

	//calculate camera model data
	void process (CameraModelData&);
	void processOneShot (const Video<uint8> &vid, CameraModelData& cm);


	// calculate camera parameters for two frames
	void processTwoFrames (Image<uint8> &curr, Image<uint8> &last, CameraModelItem& item); 



#if 0
	void calcFrames (CameraModelItem &item, const Image<uint8> &curr, const Image<uint8> &last,
					 vector<CornerPosition> &currcorners, vector<CornerPositionInt> &currcorners_int,
					 const vector<CornerPosition> &lastcorners, const vector<CornerPositionInt> &lastcorners_int);

	void calcFramesWithMask (CameraModelItem &item, const Image<uint8> &curr, const Image<uint8> &last,
					vector<CornerPosition> &currcorners, vector<CornerPositionInt> &currcorners_int,
					const Image<uint8> &mask,
					const vector<CornerPosition> &lastcorners, 
					const vector<CornerPositionInt> &lastcorners_int);
#endif

	// get corners in image
	void getCorners (const Image<uint8> &img, vector<CornerPosition>& cornersFloat, vector<CornerPositionInt>& cornersInt);

	// calculate parameters of camera model for two frames (based on corners in the images)
	void getCameraModelItem (const vector<CornerPosition> &currcorners, vector<CornerPositionInt> &currcorners_int,
							 const vector<CornerPosition> &lastcorners, const vector<CornerPositionInt> &lastcorners_int,
							 const Image<uint8> &curr, const Image<uint8> &last, 
							 CameraModelItem &item);

	void setDefaultMotion (TransformationData &item)
	{
		useDefaultMotion = true;
		defaultMotion = item;
	}


private:
	void getCameraValues();
	void readValues (CameraModelData& cod);
	void validateCameraParameters (CameraModelData& cm);
	void validateCameraParametersTranslation (CameraModelData& cm);
	void calculateCameraParameters (CameraModelData& cm);
	void calculateTranslationParameters (CameraModelData& cm); 
	void calculateTranslation (Image<uint8> &refImg, Image<uint8> &img, ieee64 &tx, ieee64 &ty);


	int32 cropleft;
	int32 cropright;
	int32 croptop;
	int32 cropbottom;

	bool normalize; // normalize luminance

	//MpegIndexData index;
	VideoSize size;
	string pathFilename;
	string path;


	// process two frames or one shot (no AVI-/MPEG-video)
	bool processShot;
	DefaultObjectPosition defaultObjPos;
	int32 defaultPx0;
	int32 defaultPx1;
	int32 defaultPy0;
	int32 defaultPy1;

	// set maximum length of motion vectors
	int32 defaultMotionVectorLength;
	
	// get translation only 
	bool calcTranslation;

	// specify default motion information (useful for panorama images)
	bool useDefaultMotion;
	TransformationData defaultMotion;

};

#endif
