/*
   This file is part of the MoCA video library.
   Name: CameraParameterEstimation.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: beier@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* -*- C++ -*-

   ParameterEstimation.h

   Author:    Markus Beier
   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim, Germany
   email: beier@pi4.informatik.uni-mannheim.de

   $Id:
*/

#ifndef CAMERA_PARAMETER_ESTIMATION
#define CAMERA_PARAMETER_ESTIMATION


#include <math.h>
#include <iostream>
#include <MocaTypes.h>
#include <Image.h>
#include <GetType.h>
#include <SvdFit.h>
#include <fstream>

enum ParameterModel {SIX_PARAMETER_COMMON, /*!< with this the parameters will be estimated using the common six parameter model */
                     SIX_PARAMETER_PERSPECTIVE, /*!< with this the parameters will be estimated using the perspective six parameter model */
                     EIGHT_PARAMETER_PERSPECTIVE /*!< with this the parameters will be estimated using the perspective eight parameter model */
};

//! Class to estimate the parameters of the camera operation belonging to given motion vectors
/*!
  This class will estimate the parameters of the camera operation.
 */


class CameraParameterEstimation
{
public:
    //! Default contructor
    CameraParameterEstimation();

    //! Constructor setting optional parameters
    /*!
      For a detailed description of the parameters see the documentation of the respective methods
      /sa setIterationsTo
      /sa setSigmaRangeTo
      /sa setSigmaThreshold
     */
    CameraParameterEstimation(uint8 iterations, ieee64 multiplier, ieee64 sigmaThreshold);

    //! processing the Vectors to calculate the parameters
    /*!
        \param srcVectors of type Image<ieee64> is an image containing the found motion vector.
        \param parameters of type vector<ieee64> is the structure wherein the resulting data will be stored
        CAUTION: The calculated parameters refer to a normalized image. To calculate vectors
        following the model you have to recalculate all coordinates relative to the center of the of the image.
        Example: Coordinate (0,0) becomes (-CenterX, CenterY)
        \param blocksize of type uint8 contains the size of the makroblocks used for motion vector search
        \param model of Type ParameterModel indicates the affine model to use for calculation of the parameters
        This mehthod calculates the parameters of the camera operation according to the model specified in the parameter ParameterModel.
        The method returns the value of the standard deviation of all found vector to their model-estimated counterparts.
    */
    ieee64 process( Image<ieee64>& srcVectors, Vector<ieee64>& parameters, uint8 blocksize = 1, ParameterModel model = EIGHT_PARAMETER_PERSPECTIVE);

    //! set the number of iteration to a custom value (default is 2)
    /*!
      \param iteration of type uint8 indicates the number of iterations used for elimination of vectors with a high deviation
      To clear the result from falsely detected vectors the method process calculates a standard deviation for the srcVectors and the vectors calculated with the
      parameters etimated in the first step. Then all vectors will be excluded from further calculation which have a deviation from the model-estimated vectors above a certain range depending of the
      standard deviation. So with this value you can set the number of iteration used for eliminating deviated vectors.
      By default this value is set to 2. Keep in mind, that the value has to be greater than 1 (with 0 iteration no parameters would be estimated).
      /sa setSigmaRangeTo
    */
    void setIterationsTo(uint8 iterations);

    //! set the multiplicator for the standard deviation of motion vectors (default = 2)
    /*!
      \param multiplier of type ieee64 is multiplicated with the standard deviation to set the range for the actual deviation of a vector for not being deleted
      To clear the result from falsely detected vectors the method process calculates a standard deviation for the srcVectors and the vectors calculated with the
      parameters etimated in the first step. Then all vectors will be excluded from further calculation which have a deviation from the model-estimated vectors above
      the value calculated with the parameter of this method multiplicated with the standard deviation.
      So with this value you can set the range within vectors will be considered as good.
      By default this value is set to 2.
    */
    void setSigmaRangeTo(ieee64 multiplier);

    //! set the threshold for the standrd deviation below which exclusion of vector is stopped
    /*!
      With this method you can set the threshold for the deviation. The optimization is stopped if either the number of optimization steps set with setIterationsTo is reached or the standard deviation sinks below this threshold.
      So you can prevent "optimization" of yet acceptable results improving the performance.
      \param sigmaThreshold of type ieee64 sets the threshold value. Its unit is pixel.
      /sa setSigmaRangeTo
    */
    void setSigmaThreshold(ieee64 sigmaThreshold);

    //! returns the the norm of the distance between the given vector and the estimated vector for each vector in the image
    /*!
      /param devImg of type Image<ieee64> is a image with one band where the norm values will be filled in. It will be structured like the input vector image.
    */
    void getModelVectors(Image<ieee64>& mdlvec);

    ieee64 getError();

private:
    ImageSize size;
    Image<ieee64> modelVectors;
    Image<ieee64> normDev;
    uint8 iterations;
    ieee64 multiplier;
    ieee64 sigmaThreshold;
    ieee64 chisq;

    void cleanVectors( Image<ieee64>& src , Image<ieee64>& mdl, Image<ieee64>& Deviation, ieee64 maxDev);
    int16 killVectors( Image<ieee64>& vectors, Image<ieee64>& Deviation, ieee64 maxDev);
    ieee64 stdDeviation( Image<ieee64>& vector1, Image<ieee64>& vector2, Image<ieee64>& normDev );

    // with this mehtod you can print the vectors to the file end.F which is psflow compliant
    void printSlice( Image<ieee64>& srcVectors, char* filename );
};



#endif
