/*
   This file is part of the MoCA video library.
   Name: CameraTransformationPrecise.h
   Copyright (C) 2004 Dirk Farin, Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Dirk Farin, Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef CAMERA_TRANSFORMATION_PRECISE_H
#define CAMERA_TRANSFORMATION_PRECISE_H

// C++ includes
#include <iostream>
using namespace std;


// MoCA includes
#include <Image.h>
#include <CameraTransformation.h>



/*
	2 images and a rough camera transformation is given

	The Class calculates the exact camera transformation, so that the images match best.

  
	Constructor: initializes variables

	process(): 
	calculate the exact camera transformation parameters 
*/
enum CameraTransformationMode {
	CAMERA_TRANSFORM_TRANSLATORIAL= 1,
	CAMERA_TRANSFORM_AFFINE       = 2,
	CAMERA_TRANSFORM_PERSPECTIVE  = 3
};

class CameraTransformationPrecise
{
public:
	
	CameraTransformationPrecise(const CameraTransformationMode mode = CAMERA_TRANSFORM_PERSPECTIVE)
	{
		ref_w = 0;
		ref_h = 0;
		img_w = 0;
		img_h = 0;

		lambda=0.001;
		minerror_valid=false;
		minerror = 0;
		isError = false;

		switch (mode) {
		case CAMERA_TRANSFORM_TRANSLATORIAL: 
			affine=false; 
			persp=false;
			break;
		case CAMERA_TRANSFORM_AFFINE: 
			affine=true; 
			persp=false;
			break;
		case CAMERA_TRANSFORM_PERSPECTIVE: 
			affine=true; 
			persp=true;
			break;
		} // switch
	}

	~CameraTransformationPrecise() {};
	
	void process (const Image<uint8> &referenceImage, const Image<uint8> &img, const Image<uint8> &mask,	 	
				  const TransformationData &srcCameraParameters,  // initial camera parameters
				  TransformationData &destCameraParameters, ieee64 &minErr);      // precise camera parameters

	void process (const Image<uint8> &referenceImage, const Image<uint8> &img, const Image<uint8> &mask, 	 	
				  TransformationData &srcDestCameraParameters, ieee64 &minErr)
	{
		TransformationData tmp = srcDestCameraParameters;
		process (referenceImage, img, mask, tmp, srcDestCameraParameters, minErr);

		return;
	}

private:
	void transformationStep (TransformationData &trans);
	ieee64 calcError(const ieee64 *s);
	ieee64 refGetPixel (ieee64 y,ieee64 x);
	void refGetPixelGrads (ieee64 y,ieee64 x, ieee64& err,ieee64& I1x, ieee64& I1y);
	void calcAll (const ieee64* state);



	inline bool AtOptimum() const 
	{ 
		return lambda >= 1e6; 
	}


	inline TransformationData valueToTrans(const double* s)
	{
		TransformationData t;
		t.a00 =s[0];
		t.a01 =s[1];
		t.tx  =s[2];
		t.a10 =s[3];
		t.a11 =s[4];
		t.ty  =s[5];
		t.px  =s[6];
		t.py  =s[7];
		return t;
	}
	inline void delparam(int32 n,gsl_matrix* m,gsl_vector* v)
	{
		for (int32 i=0;i<8;i++) {
			gsl_matrix_set(m,i,n,0);
			gsl_matrix_set(m,n,i,0);
		} // for
		
		gsl_matrix_set(m,n,n,1);
		gsl_vector_set(v,n,0);
		return;
	}



	bool affine,persp;

	int32 ref_w;
	int32 ref_h;
	int32 img_w;
	int32 img_h;
	Image<uint8> img;
	Image<uint8> refImg;
	Image<uint8> mask;

	ieee64 lambda;
	ieee64 minerror;
	bool   minerror_valid;
	bool   isError;

	ieee64 gradient[8];
	ieee64 hesse[8][8];

};
#endif
