/*
   This file is part of the MoCA video library.
   Name: CcrFeature.h
   Copyright (C) 2004 Markus Beier, Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: beier@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  \author    Markus Beier, Stephan Kopf
  
   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: beier@pi4.informatik.uni-mannheim.de
   
	(c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	
*/


#ifndef CCR_FEATURE_H
#define CCR_FEATURE_H


//C++ includes
#include <vector>
#include <string>
#include <sstream>
#include <iostream>
using namespace std;

// MoCA includes
#include <MocaTypes.h>
#include <VideoDecoder.h>
#include <MocaTypes.h>
#include <CannyFilter.h>
#include <DilationFilter.h>
#include <GrayscaleFilter.h>
#include <BinarizeFilter.h>
#include <InverseFilter.h>
#include <CropFilter.h>
#include <OImageStream.h>
#include <CameraModel.h>
#include <CameraTransformation.h>
#include <EdgeDetectionFilter.h>




//! Class to calculate CCR (corner change ratio)
/*!
  This class will calculate the CCR values of the frames of a video.
  The difference of consecutive CCR values will be calculated and 
  stored in a vector provided as parameter.
*/

const static string CCR_FEATURE_VERSION  ="Version-0.02";
const static string CCR_FEATURE_FILENAME ="ccr.idx";



struct CcrData
{
	CcrData () 
	{
		version = CCR_FEATURE_VERSION;
		filename = "";

		size              = 0;
		minCcrValue       = 1e10;
		maxCcrValue       = -1 ;
		avgCcrValue       = 0;
		ccrFilename       = "";
		pinFilename       = "";
		poutFilename      = "";
		ccr.clear();
		pin.clear();
		pout.clear();
	} 
    CcrData& operator=(const CcrData &other) {
		if (this != &other) {
			version            = other.version;
			filename           = other.filename;

			size               = other.size;
			minCcrValue        = other.minCcrValue;
			maxCcrValue        = other.maxCcrValue;
			avgCcrValue        = other.avgCcrValue;
			ccrFilename        = other.ccrFilename;
			pinFilename        = other.pinFilename;
			poutFilename       = other.poutFilename;
			ccr                = other.ccr;
			pin                = other.pin;
			pout               = other.pout;
			mocaPath           = other.mocaPath;
			pathFilename       = other.pathFilename;
		} // if
        return *this;
    } // operator=
	
	
	ieee64 minCcrValue;
	ieee64 maxCcrValue;
	ieee64 avgCcrValue;

	string ccrFilename;
	string pinFilename;
	string poutFilename;

	int32 size;
    Signal<int32> ccr;
    Signal<int32> pin;
    Signal<int32> pout;

	string version;
	string filename;
	string mocaPath;
	string pathFilename;
};

std::ostream& operator<< (ostream&,    const CcrData&);
OTagStream&   operator<< (OTagStream&, const CcrData&);
ITagStream&   operator>> (ITagStream&, CcrData&);




class CcrFeature
{
public:
	static const string DEFAULT_CCR_FEATURE_FILE_NAME;

	
    //! Contructor initializing the class with a filename of the MPEG file 
    /*!
	\param fullname of Type string& provides the filename of the mpeg file to be processed
    */
    CcrFeature (string &filename)
	{
		pathFilename = filename;
		VideoDecoder reader (pathFilename, CODEC_Y);
		size = reader.getSize();

		CreateMocaPath pathReader (pathFilename, false);
		mocaPath = pathReader.destinationPath;
	}
	
    //! Default destructor
    ~CcrFeature() {};
	
	
    //! Calculating the CCR differences of consecutive frames in the 
	//! given video, store the CCR values in a file and then returning these values.
    /*!
	\param item of Type CcrData& is the structure wherein the resulting data will be stored. 
    */
	void process (CcrData& ccr);

	
private:
	void readValues (CcrData& ccr);

private:

	string pathFilename;
	string mocaPath;
	VideoSize size;

}; //class CcrFeature


#endif
