/*
   This file is part of the MoCA video library.
   Name: ContourFeature.cpt
   Copyright (C) 2004 Oliver Schuster

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: schuster@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file



   \author    Oliver Schuster

              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim
              68131 Mannheim
              Germany
              email: schuster@informatik.uni-mannheim.de

   (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt

   $Id: ContourFeature.cpt,v 1.1 2004/02/17 11:52:58 kopf Exp $

*/

#include <ContourFeature.h>

ContourFeature::ContourFeature()
{
} // ContourFeature::ContourFeature


ContourFeature::~ContourFeature()
{
} // ContourFeature::ContourFeature


/*!

  The filter itself

  \param src The src image. It is assumed that it contains a single black contour on a white background.

 */
void ContourFeature::process(Image<uint8> &src, ContourFeatureData &data)
{

    // get size of source image and create a visited image, initialise
    // it with maximum values
    ImageSize size = src.getSize();
    size[DIM_BAND] = 1;
    Image<uint8> visited(size);
    visited.clear (255); 

    // find starting pixel by starting in the first column and
    // traversing each column from top to bottom

    Pixel<uint8> startPixel;

    for (int x = 0; x < size[DIM_X] && startPixel.values.empty(); x++) {
        for (int y = 0; y < size[DIM_Y] && startPixel.values.empty(); y++) {
            // find black pixel
            if (src(x, y, 0) == 0) {
                startPixel = ImagePosition(x, y);
                // mark startpixel as 0
                startPixel.values.push_back(0);
            } // if
        } // for
    } // for


    // list of pixels on contour, result at the end
    list<Pixel<uint8> > contour;

    // list of crossings, positions where more than one neighbour is
    // eligible
    list<Crossing<uint8> > clist;

    // Variable to hold next pixel to be processed, start with
    // starting pixel
    Pixel<uint8> nextPixel = startPixel;

    bool foundStartAgain = false;

    // main loop, until all pixels of contour are found
    do {

        // set next pixel to actual pixel being processed, set it to
        // visited and add it to contour list
        Pixel<uint8> actPixel = nextPixel;
        visited.setPixel(actPixel);
        contour.push_back(actPixel);

        // TODO: for debugging only
        //cout << actPixel << endl;

        // get neighbours of actual pixel, 8-neighbourship
        list<Pixel<uint8> > neighbours = src.getNeighbours(actPixel, 8);

        // remove background neighbours (being non-black) and neighbours already visited
        for (list<Pixel<uint8> >::iterator it = neighbours.begin(); it != neighbours.end();) {
            if ((*it).values[0] != 0 || visited.getPixel(*it).values[0] == 0) {
                it = neighbours.erase(it);
            } else {
                it++;
            } // if
        } // for

        // if only one neighbour exist, it is the next pixel -> continue
        if (neighbours.size() == 1) {
            nextPixel = neighbours.front();
            continue;
        } else if (neighbours.empty()){

            // else if no more neighbours exist, check if we are close
            // to the starting pixel (i.e. a neighbour of it)

            list<Pixel<uint8> > neighbours = src.getNeighbours(actPixel, 8);

            for (list<Pixel<uint8> >::iterator it = neighbours.begin(); it != neighbours.end(); it++) {
                if ((*it).getX() == startPixel.getX()
                    && (*it).getY() == startPixel.getY()) {
                    foundStartAgain = true;

                    // yes, we are a neighbour of the starting pixel,
                    // check if all pixels of the contour are found.

                    for (int x = 0; x < size[DIM_X] && foundStartAgain; x++) {
                        for (int y = 0; y < size[DIM_Y] && foundStartAgain; y++) {
                            if (src(x, y, 0) == 0 && visited(x, y, 0) != 0) {
                                // nope, at least one pixel is not in the list
                                foundStartAgain = false;
                            } // if
                        } // for
                    } // for
                    break;

                } // if
            } // for

            // no further neighbours exist, walk the same way back...
            if (!foundStartAgain) {
                // TODO: debugging, temporary:
                cerr << "Not all pixels visited, but nevertheless found a contour..." << endl;
                cerr << "Backing up, trying again" << endl;

                Crossing<uint8> x;

                do {
                    x = clist.back();
                    clist.pop_back();
                } while (visited(x.p.getX(), x.p.getY(), 0) == 0);

                //cerr << "Backing up to" << endl;
                //cerr << x.p << " " << *(x.i) << endl;

                for (list<Pixel<uint8> >::iterator it = x.i; it != contour.end(); it++) {
                    // TODO: replace 255 with some getMax function
                    (*it).values[0] = 255;
                    visited.setPixel(*it);
                    //cerr << "removed pixel " << (*it) << endl;
                } // for

                contour.erase(x.i, contour.end());

                nextPixel = x.p;

            } // if
        } else {
            // more than one valid neighbour exists!

            // choose one neighbour to follow
            nextPixel = neighbours.front();
            neighbours.pop_front();

            // for all other neighbours
            for (list<Pixel<uint8> >::iterator it = neighbours.begin(); it != neighbours.end(); it++) {
                // build struct containing pos in contour and pixel

                list<Pixel<uint8> >::iterator cit = contour.end();
                cit--;
                Crossing<uint8> x(*it, cit);

                //cerr << "Found additional neighbour" << endl;
                //cerr << *it << " " << *cit << endl;

                // add struct to crossinglist
                clist.push_back(x);

            } // for
        } // if


    } while(!foundStartAgain);

    cout << "The following pixels are not, but should be, part of the contour: " << endl;
    for (int x = 0; x < size[DIM_X]; x++) {
        for (int y = 0; y < size[DIM_Y]; y++) {
            // find black pixel
            if (src(x, y, 0) == 0 && visited(x, y, 0) != 0) {
                cout << x << " " << y << endl;
            } // if
        } // for
    } // for

	data.data = contour;
	return;
} // process
