/*
    CornerDetection.cpp     Harris corner detector
    Copyright (C) 2004 Dirk Farin

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#include <CornerDetection.h>



void 
CornerDetection::filterCorners(vector<CornerPosition>& corners,int left,int top,int right,int bottom)
{
	vector<CornerPosition> tmp = corners;
	corners.clear();
	for (int32 i=0; i<(int32)tmp.size(); i++) {
		
		int32 x = (int)tmp[i].x;
		int32 y = (int)tmp[i].y;
		
		if (!(x<left || x>right || y<top || y>bottom)) { 
			corners.push_back (tmp[i]);
		} // if
	} // for
	return;
} // filterCorners

void 
CornerDetection::filterCornersInt(vector<CornerPositionInt>& corners,int left,int top,int right,int bottom)
{
	vector<CornerPositionInt> tmp = corners;
	corners.clear();
	for (int32 i=0; i<(int32)tmp.size(); i++) {
		
		int32 x = (int)tmp[i].x;
		int32 y = (int)tmp[i].y;
		
		if (!(x<left || x>right || y<top || y>bottom)) { 
			corners.push_back (tmp[i]);
		} // if
	} // for
	return;
} // filterCorners



void 
CornerDetection::removeCornersRect(vector<CornerPosition>& corners,int left,int top,int right,int bottom)
{
	vector<CornerPosition> tmp = corners;
	corners.clear();
	for (int32 i=0; i<(int32)tmp.size(); i++) {
		
		int32 x = (int)tmp[i].x;
		int32 y = (int)tmp[i].y;
		
		if (!(x>=left && x<=right && y>=top && y<=bottom)) { 
			corners.push_back (tmp[i]);
		} // if
	} // for
	return;
} // removeCornersRect

void 
CornerDetection::removeCornersRectInt(vector<CornerPositionInt>& corners,int left,int top,int right,int bottom)
{
	vector<CornerPositionInt> tmp = corners;
	corners.clear();
	for (int32 i=0; i<(int32)tmp.size(); i++) {
		
		int32 x = (int)tmp[i].x;
		int32 y = (int)tmp[i].y;
		
		if (!(x>=left && x<=right && y>=top && y<=bottom)) { 
			corners.push_back (tmp[i]);
		} // if
	} // for
	return;
} // removeCornersRectInt



// remove corner, if (mask(x,y) == 255)
void 
CornerDetection::filterCornersInt(vector<CornerPositionInt>& corners, const Image<uint8> &mask)
{
	vector<CornerPositionInt> tmp = corners;
	corners.clear();
	for (int32 i=0; i<(int32)tmp.size(); i++) {
		
		int32 x = tmp[i].x;
		int32 y = tmp[i].y;
		if (mask(x,y) != 255) {
		    corners.push_back (tmp[i]);
		} // if

	} // for
	return;
} // filterCorners


void 
CornerDetection::cornersToInt(const vector<CornerPosition> &corners, vector<CornerPositionInt> &dest)
{
	dest.clear();
	vector<CornerPosition>::const_iterator it; 
	for (it=corners.begin(); it!=corners.end(); it++) {
		
		int32 x = (int32)(*it).x;
		int32 y = (int32)(*it).y;
		
		CornerPositionInt p (x,y);
		dest.push_back (p);
	} // for
	return;
} // cornersToInt



void 
CornerDetection::process (const Image<uint8>& img, vector<CornerPosition>& corners)
{
	
	ImageSize size = img.getSize();
	const int width  = size[DIM_X];
	const int height = size[DIM_Y];
	
	smooth_input = img;
	CornerBiFilter (smooth_input, d_steps_input);
	
	xx.resize(size);
	yy.resize(size);
	xy.resize(size);


	int16* pxx = xx.getSamplePtr(0,0);  
	int16* pyy = yy.getSamplePtr(0,0);  
	int16* pxy = xy.getSamplePtr(0,0);  
	for (int x=0;x<width;x++) {
		*pxx = 0; pxx++;  
		*pyy = 0; pyy++;  
		*pxy = 0; pxy++;  
	} // for
	pxx = xx.getSamplePtr(0,height-1);  
	pyy = yy.getSamplePtr(0,height-1);  
	pxy = xy.getSamplePtr(0,height-1);  
	for (int x=0;x<width;x++) {
		*pxx = 0; pxx++;  
		*pyy = 0; pyy++;  
		*pxy = 0; pxy++;  
	} // for


	for (int y=1;y<height-1;y++) {
		//xx(0,y) = 0;   xx(width-1,y) = 0;
		//yy(0,y) = 0;   yy(width-1,y) = 0;
		//xy(0,y) = 0;   xy(width-1,y) = 0;
		pxx = xx.getSamplePtr(0,y);  
		*pxx = 0; pxx++;  
		pyy = yy.getSamplePtr(0,y);  
		*pyy = 0; pyy++;  
		pxy = xy.getSamplePtr(0,y);  
		*pxy = 0; pxy++;  


		uint8* smoothPtrX0 = smooth_input.getSamplePtr(0,y);   // start with index 0 
		uint8* smoothPtrX2 = smooth_input.getSamplePtr(2,y);   // start with index 2 
		for (int x=1;x<width-1;x++) {

			//int16 dh = (smooth_input(x+1,y) - smooth_input(x-1,y))>>1;
			int16 dh = ((*smoothPtrX2) - (*smoothPtrX0))>>1;
			smoothPtrX0++;
			smoothPtrX2++;

			int16 dv = (smooth_input(x,y+1) - smooth_input(x,y-1))>>1;
			
			//xx(x,y) = dh*dh;
			//yy(x,y) = dv*dv;
			//xy(x,y) = dh*dv;
			*pxx = dh*dh;   pxx++; 
			*pyy = dv*dv;   pyy++; 
			*pxy = dh*dv;   pxy++; 
		} // for


		//xx(width-1,y) = 0;
		//yy(width-1,y) = 0;
		//xy(width-1,y) = 0;
		*pxx = 0; 
		*pyy = 0; 
		*pxy = 0; 
	} // for

	CornerBiFilter(xx, d_steps_correlation);
	CornerBiFilter(yy, d_steps_correlation);
	CornerBiFilter(xy, d_steps_correlation);



	Image<int32>  feat(size);
	//Image<bool> nonmax(size);

	int32 *fp = feat.getSamplePtr(0,0); 
	for (int x=0;x<width;x++) {
		*fp = 0; fp++;  
	} // for
	fp = feat.getSamplePtr(0,height-1); 
	for (int x=0;x<width;x++) {
		*fp = 0; fp++;  
	} // for


	for (int y=1;y<height-1;y++) {
		
		fp = feat.getSamplePtr(0,y);  
		*fp = 0; fp++;  

		int16* pxx = xx.getSamplePtr(1,y);  // start with index 1
		int16* pyy = yy.getSamplePtr(1,y);  // start with index 1
		int16* pxy = xy.getSamplePtr(1,y);  // start with index 1
		for (int x=1;x<width-1;x++) {

			//int det = xx(x,y) * yy(x,y) - xy(x,y) * xy(x,y);
			int32 det = (*pxx) * (*pyy) - (*pxy) * (*pxy);
			//int tr  = xx(x,y) + yy(x,y);
			int32 tr  = (*pxx) + (*pyy);
			
			//feat(x,y) = det - tr*tr/16; // approx. 0.06
			*fp = det - tr*tr/16; // approx. 0.06
			
			if (tr < d_flat_threshold) {
				//feat(x,y) = -1;
				*fp = -1;
			} // if

			fp++;  
			pxx++; 
			pyy++; 
			pxy++; 
		} // for

		*fp = 0; 
	} // for


#if 0
	// store frame
	ImageSize si(xx.getSize()[DIM_X], xx.getSize()[DIM_Y], 1);
	Image<uint8> disp (si);

  for (int y=0; y<xx.getSize()[DIM_Y]; y++)
    for (int x=0; x<xx.getSize()[DIM_X]; x++) {
      int32 v =  feat(x,y);
      if (v<0) v=0;
      if (v>255) v=255;
      disp(x,y) = (uint8)v;
    }

WriteImage("F:/corner-new.pgm", disp);
exit(1);
	// end store frame	
#endif

	

	corners.clear();
	if (useSignificantCorners) {
		const int32 MAXIMA_SUPRESSION = 4;
		for (int y=2;y<height-2;y++) {
			for (int x=2;x<width-2;x++) {
				int32 featXY = feat(x,y);

				if (featXY > 0) {
					bool isMax=true;
					for (int32 yy=-MAXIMA_SUPRESSION; yy<=MAXIMA_SUPRESSION; yy++) {
						int32 dy = yy+y;
						if ((dy>0) && (dy<height-1)) {
							for (int32 xx=-MAXIMA_SUPRESSION; xx<=MAXIMA_SUPRESSION; xx++) {
								int32 dx = xx+x;
								if ((dx>0) && (dx<width-1)) {
									if ((xx!=0) || (yy!=0)) {
										if (feat(dx,dy) >= featXY) {
											isMax=false;
										} // if
									} // if
								} // if
							} // for
						} // if
					} // for

					if (isMax) {
						double dx,dy;
						double a,b;
						a = (feat(x+1,y)+feat(x-1,y)-2*featXY)/2.0;
						b = (feat(x+1,y)-feat(x-1,y))/2.0;
						dx = -b/(2*a);
						
						a = (feat(x,y+1)+feat(x,y-1)-2*featXY)/2.0;
						b = (feat(x,y+1)-feat(x,y-1))/2.0;
						dy = -b/(2*a);
						
						CornerPosition p (x + dx, y + dy);
						corners.push_back(p);
					} // else
				} // if
			} // for
		} // for
	} else {
		for (int y=2;y<height-2;y++) {
			for (int x=2;x<width-2;x++) {
				int32 featXY = feat(x,y);
				if (featXY > 0) {
					if      (feat(x-1,y)   >= featXY) {}
					else if (feat(x+1,y)   >  featXY) {}
					else if (feat(x,y-1)   >= featXY) {}
					else if (feat(x,y+1)   >  featXY) {}
					else if (feat(x+1,y+1) >  featXY) {}
					else if (feat(x-1,y+1) >= featXY) {}
					else if (feat(x+1,y-1) >  featXY) {}
					else if (feat(x-1,y-1) >= featXY) {}
					else
					{
						double dx,dy;
						double a,b;
						a = (feat(x+1,y)+feat(x-1,y)-2*featXY)/2.0;
						b = (feat(x+1,y)-feat(x-1,y))/2.0;
						dx = -b/(2*a);
						
						a = (feat(x,y+1)+feat(x,y-1)-2*featXY)/2.0;
						b = (feat(x,y+1)-feat(x,y-1))/2.0;
						dy = -b/(2*a);
						
						CornerPosition p (x + dx, y + dy);
						corners.push_back(p);
					} // else
				} // if
			} // for
		} // for
	} // else
	// return corners.AskSize();
	return;
}


void 
CornerDetection::drawCorners (Image<uint8> &img, const vector<CornerPositionInt>& corners)
{
	DrawCross cross (6, Color::BLACK);
	//DrawCross cross (10, Color::WHITE);
	vector<CornerPositionInt>::const_iterator it; 
	for (it=corners.begin(); it!=corners.end(); it++) {
		cross.process(ImagePosition((*it).x, (*it).y), img);
	} // for

	return;
} // drawCorners

