/*
    CssDatabase.cpp     database for the curvature scale space approach (CSS)
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <CssDatabase.h>

CssDatabase::CssDatabase(CssDatabaseSchemaType schema)
{
    this->schema = schema;
    openDatabase();
}


void
CssDatabase::openDatabase ()
{
    cssData.clear();
    objectClassName.clear();
	
    CssDatabaseSchema dbSchema (schema);
    dbSchema.getCssStoredDatabaseData (cssData, objectClassName);
	
    return;
}

void
CssDatabase::createDatabase (const string &dbFilename, const string &cssSourcePath)
{
    // initilaize variables
    string fn = cssSourcePath;
    FileHandling::convertToValidUnixFilename(fn);
    
    databaseFilename = dbFilename;
    FileHandling::convertToValidUnixFilename(databaseFilename);
    
	
    // read CSS-filenames
    string fileExt = "*CSS.tag";
    vector<FileInfo> databaseList;
    FileList flData (fn, true);
    flData.process (fileExt, databaseList);
    cout << "("<< databaseList.size() << ")"<< endl;
	
	
    // open destination file
    ofstream ofs (databaseFilename.c_str(), std::ios::out | std::ios::binary);
    
	
	ofs << "/* -*- C++ -*-\n"
		<< "\n"
		<< "   Author: Stephan Kopf\n"
		<< "   Lehrstuhl fuer Praktische Informatik IV\n"
		<< "   Universitaet Mannheim, Germany\n"
		<< "   email: kopf@pi4.informatik.uni-mannheim.de\n"
		<< "\n"
		<< "*/\n"
		<< "\n";
	
	
    // get all object classes
    set<string> classNamesSet;
    vector<FileInfo>::iterator it;
    for (it = databaseList.begin(); it != databaseList.end(); it++) {
		// get filename
		string filename = (*it).name;
		
		// get object class name
		int p = filename.find_last_of('/') - string("result/").size();
		string objectClass = ((*it).name).substr(0,p);
		p = objectClass.find_last_of('/')+1;
		objectClass = objectClass.substr(p);
		
		// convert filename to upper case
		FileHandling::upper(objectClass);
		
		// add to set
		classNamesSet.insert(objectClass);
    } // for
    
    // create vector with object class and index
    vector<string> classNames;
    set<string>::iterator setIt;
    for (setIt = classNamesSet.begin(); setIt != classNamesSet.end(); setIt++) {
		classNames.push_back(*setIt);
    } // for
    
    // store DATABASE_OBJECT_CLASSES
    ofs << "const int32 OBJECT_CLASSES_SIZE = " << classNames.size() << "; \n";
    ofs << "static char* OBJECT_CLASSES [OBJECT_CLASSES_SIZE] = \n" 
		<< "{\n";
    for (int32 i=0; i<(int32)classNames.size(); i++) {
		ofs << "    \"" << classNames[i] << "\",\n";
    } // for
    ofs << "};\n\n";
    
    
    // store DATABASE_CSS
    ofs << "const int32 DB_SIZE = " << databaseList.size() << ";" << endl;
    ofs << "static CssDatabaseItem DB[DB_SIZE] = \n" 
		<< "{\n";
	
	// store css values
    for (it = databaseList.begin(); it != databaseList.end(); it++) {
		
		// store filename
		string filename = (*it).name;
		ofs << "    {\n";
		ofs << "        \"" << filename << "\"," << endl;
		
		// get object class name
		int p = filename.find_last_of('/') - string("result/").size();
		string objectClass = ((*it).name).substr(0,p);
		p = objectClass.find_last_of('/')+1;
		objectClass = objectClass.substr(p);
		FileHandling::upper(objectClass);
		
		// get object class index
		int32 objectClassIndex = 0;
		for (int32 i=0; i<(int32)classNames.size(); i++) {
			if (classNames[i] == objectClass) objectClassIndex=i;
		} // for
		if (classNames[objectClassIndex] != objectClass) {
			cerr << "CssDatabase::createDatabase: object class not found" << endl;
			exit(1);
		} // if
		
		// read CSS values
		CssMaximumData cm;
		ITagStream its ((*it).name);
		its >> cm;
		
		
		// store values
		ofs << "        " << objectClassIndex << ", "
			<< cm.xyRatio << ",\n";
		
		ofs << "        { ";
		vector<CssMaximumItem>::iterator cssDataIt = cm.data.begin();
		bool lastCssDataItFound = false;
		for (int32 i=0; i<10; i++) {
			CssDatabasePeak peak;
			peak.height=0;
			peak.isConcave=false;
			peak.width=0;
			peak.position=0;
			
			if (!lastCssDataItFound) {
				if (cssDataIt != cm.data.end()) {
					
					peak.height    =(*cssDataIt).height;
					peak.isConcave =(*cssDataIt).isConcave;
					peak.width     =(*cssDataIt).width;
					peak.position  =(*cssDataIt).position;
					cssDataIt++;
				} // if
				
				if (cssDataIt == cm.data.end()) {
					lastCssDataItFound=true;
				} // if
			} // if
			
			// store values
			ofs << "{" << peak.position << "," << peak.height << "," << peak.width << ",";
			if (peak.isConcave) {
				ofs << "1";
			} else {
				ofs << "0";
			} // else
			if (i==4) {
				ofs << "}, \n          " ;
			} else if (i==9) {
				ofs << "} ";
			} else {
				ofs << "}, ";
			} // if
		} // for
		ofs << "}\n";
		ofs << "    }, \n\n";
		
    } // for    
    
    ofs << "};\n" 
		<< "\n"
		<< "#endif\n" 
		<< "\n" 
		<< endl;
    ofs.close();
    
    
    return;
} // createDatabase


void 
CssDatabase::setCssMaximumData (const CssDatabaseItem &src, CssMaximumData &dest)
{
    dest.xyRatio = src.xyRatio;
    dest.data.clear();
    
    // set CSS peaks
    CssMaximumItem maximumItem;
    for (int32 j=0; j<10; j++) {
		if (src.peak[j].height > 5) {
			maximumItem.position   = src.peak[j].position;
			maximumItem.height     = src.peak[j].height;
			maximumItem.width      = src.peak[j].width;
			maximumItem.isConcave  = src.peak[j].isConcave;
			maximumItem.curveStart = 50;
			maximumItem.curveEnd   = 50;
			maximumItem.visited    = false;
			
			dest.data.push_back(maximumItem);
		} // if
    } // for
    
    return;
} // setCssMaximumData 

void 
CssDatabase::findBestMatch(const CssDatabaseItem &src, int32 &objectClassIndex, int32 &distance)
{
    objectClassIndex=-1;
    distance = -1;
    
    bool XYratio=true;
    int thCriticalHeight = 50; 
    int thPosition = 30; 
    int thHeight = 30; 
    int thWidth = 60;
    bool thCheckConcave = true; 
    bool rotateCss=true;
    
    CssMatching cssm(
		ContourParametrization::DEFAULT_CONTOUR_SIZE, 
		thPosition, thHeight, thWidth, thCriticalHeight, thCheckConcave, XYratio, rotateCss); 
    
	
    // create query item
    CssMaximumData queryItem;
    setCssMaximumData (src, queryItem);
    
    
    // create database items
    int32 tmpDistance = 0;
    distance = 10000000;
    CssMaximumData dbItems;
    for (int32 i=0; i<(int32)cssData.size(); i++) {								
		
		// get database item
		CssMaximumData dbItem = cssData[i].data;
		
		// calculate distance
		tmpDistance = 0;
		tmpDistance = cssm.process(dbItem, queryItem);
		if (distance > tmpDistance) {
			distance = tmpDistance;
			objectClassIndex = cssData[i].objectClassId;
		} // if
    } // for
    
    return;
}


