/*
   This file is part of the MoCA video library.
   Name: CutDetectionFeature.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef CUT_DETECTION_H
#define CUT_DETECTION_H

// C includes
#include <math.h>

// C++ includes
#include <iomanip>
#include <sstream>
#include <iostream>
using namespace std;


// MoCA includes
#include <MocaSignal.h>
#include <WriteMocaData.h>
#include <HistogramDifferenceFeature.h>
#include <VideoStatistic.h>
#include <GaussFilter.h>
#include <CannyFilter.h>
#include <BinarizeFilter.h>
#include <EcrFeature.h>
#include <FileHandling.h>


// windows includes
#ifdef COMPILER_MSVC
#include <process.h>
#endif


const static string CUT_DETECTION_VERSION  ="Version-0.9";
const static string CUT_DETECTION_FILENAME ="cuts.idx";


enum CutDetectionType {
	SHOT_CHANGE_UNKNOWN        = 0, 
	SHOT_CHANGE_NO_CUT         = 1, 

	SHOT_CHANGE_HARD_CUT       = 2, 
	SHOT_CHANGE_SOFT_CUT       = 3, 

	SHOT_CHANGE_DISSOLVE       = 4, 
	SHOT_CHANGE_DISSOLVE_FIRST = 5, 
	SHOT_CHANGE_DISSOLVE_LAST  = 6, 

	SHOT_CHANGE_WIPE           = 7,
	SHOT_CHANGE_WIPE_FIRST     = 8,
	SHOT_CHANGE_WIPE_LAST      = 9,

	SHOT_CHANGE_FADE_IN        = 10,
	SHOT_CHANGE_FADE_OUT       = 11,

	SHOT_CHANGE_OTHER          = 12

};

struct CutDetectionData
{
	CutDetectionData () 
	{
		version = CUT_DETECTION_VERSION;
		filename = "";
		numberOfHardCuts = 0;
		numberOfFadeIn   = 0;
		numberOfFadeOut  = 0;
		numberOfDissolves= 0;
		numberOfWipes= 0;
		numberOfUnknownSceneChanges = 0;
		threshold = 0.0;
		thHardCut = 0.0;
		numberOfFrames = 0;

	} 
    CutDetectionData& operator=(const CutDetectionData &other) {
		if (this != &other) {
			version           = other.version;
			filename          = other.filename;
			cuts              = other.cuts;
			numberOfHardCuts  = other.numberOfHardCuts;
			numberOfFadeIn    = other.numberOfFadeIn;
			numberOfFadeOut   = other.numberOfFadeOut;
			numberOfDissolves = other.numberOfDissolves;
			numberOfWipes     = other.numberOfWipes;
			numberOfUnknownSceneChanges = other.numberOfUnknownSceneChanges;
			threshold         = other.threshold;
			thHardCut         = other.thHardCut;
			numberOfFrames    = other.numberOfFrames;
		} // if
        return *this;
    } // operator=

	void updateCounter()
	{
		numberOfHardCuts = 0;
		numberOfFadeIn   = 0;
		numberOfFadeOut  = 0;
		numberOfDissolves= 0;
		numberOfWipes= 0;
		numberOfUnknownSceneChanges = 0;
		int32 nx=cuts.getSize()[DIM_X];
		for (int32 i=0; i<nx;i++) {
			if (cuts(i) == SHOT_CHANGE_HARD_CUT) numberOfHardCuts++;
			if (cuts(i) == SHOT_CHANGE_OTHER) numberOfUnknownSceneChanges++;
			if (cuts(i) == SHOT_CHANGE_FADE_IN) {
				numberOfFadeIn++;
				while ((i+1<nx) && (cuts(i+1) == SHOT_CHANGE_FADE_IN)) i++;
			} // if
			if (cuts(i) == SHOT_CHANGE_FADE_OUT) {
				numberOfFadeOut++;
				while ((i+1<nx) && (cuts(i+1) == SHOT_CHANGE_FADE_OUT)) i++;
			} // if
			if ((cuts(i) == SHOT_CHANGE_DISSOLVE) || 
				(cuts(i) == SHOT_CHANGE_DISSOLVE_FIRST) || 
				(cuts(i) == SHOT_CHANGE_DISSOLVE_LAST)) {
				numberOfDissolves++;
				while ((i+1<nx) && (cuts(i+1) == SHOT_CHANGE_DISSOLVE_FIRST)) i++;
				while ((i+1<nx) && (cuts(i+1) == SHOT_CHANGE_DISSOLVE)) i++;
				while ((i+1<nx) && (cuts(i+1) == SHOT_CHANGE_DISSOLVE_LAST)) i++;
			} // if
			if ((cuts(i) == SHOT_CHANGE_WIPE) || 
				(cuts(i) == SHOT_CHANGE_WIPE_FIRST) || 
				(cuts(i) == SHOT_CHANGE_WIPE_LAST)) {
				numberOfWipes++;
				while ((i+1<nx) && (cuts(i+1) == SHOT_CHANGE_WIPE_FIRST)) i++;
				while ((i+1<nx) && (cuts(i+1) == SHOT_CHANGE_WIPE)) i++;
				while ((i+1<nx) && (cuts(i+1) == SHOT_CHANGE_WIPE_LAST)) i++;
			} // if
		} // for
		return;
	}

	string version;
	string filename;
	uint16 numberOfHardCuts;
	uint16 numberOfFadeIn;
	uint16 numberOfFadeOut;
	uint16 numberOfDissolves;
	uint16 numberOfWipes;
	uint16 numberOfUnknownSceneChanges;

	Signal<CutDetectionType> cuts;
	ieee64 threshold;
	ieee64 thHardCut;
	int32  numberOfFrames;
};

std::ostream& operator<< (ostream &, const CutDetectionData&);
OTagStream& operator<< (OTagStream&, const CutDetectionData&);
ITagStream& operator>> (ITagStream&, CutDetectionData&);





/*
	Constructor: initializes variables
	
	process(): 
		if status file does not exist:
		   calculate the hard cuts from mpeg-file, save status file
		else:
		   read hard cuts from status file
*/
class CutDetection
{
public:

	CutDetection(const string &pathFilename);
	CutDetection(const string &pathFilename, bool detectFades, bool detectDissolves, bool storeImages, bool maxContrast=false);
	~CutDetection() {};

	void process (CutDetectionData&);
	void getShotFrameNumbers (CutDetectionData& data, Signal<int32>& shotFrameNumber);
	void getShotFrameNumbersContrast (CutDetectionData& data, Signal<int32>& shotFrameNumber);
	void saveJpegImages (CutDetectionData& data);
	void saveContrastImages (CutDetectionData& data);

private:
	
	ieee64 getMaxHistValue (HistogramDifferenceItem &, int32 p1, int32 p2);
	ieee64 getMaxHistValue (HistogramDifferenceItem &, int32 p1, int32 p2, int32 p3, int32 p4);
	ieee64 getMinHistValue (HistogramDifferenceItem &, int32 p1, int32 p2);
	ieee64 getMinHistValue (HistogramDifferenceItem &, int32 p1, int32 p2, int32 p3, int32 p4);
	ieee64 getAvgHistValue (HistogramDifferenceItem &, int32 p1, int32 p2);
	ieee64 getAvgHistValue (HistogramDifferenceItem &, int32 p1, int32 p2, int32 p3, int32 p4);

	void removeHardCut (HistogramDifferenceItem *item, int32 frame, int32 lastT);
	void removeSoftCut (HistogramDifferenceItem *item, int32 firstFrame, int32 lastFrame, int32 lastT);
	
	void readValues (CutDetectionData& scd);
	void initHistogram (int32 skip);
	void getDissolveArea (CutDetectionData& cutData, Signal<int32> &dissolvePosition);
	bool cutExist (CutDetectionData& data, int32 firstFrame, int32 lastFrame);


	void findHardCuts (CutDetectionData& data);
	void findFades (CutDetectionData& data);
	void removeFalseHardCuts (CutDetectionData& data);
	void findDissolves (CutDetectionData& data);
	void findShortDissolves (CutDetectionData& data);

	void findUnknownSceneChanges (CutDetectionData& data);

	HistogramDifferenceItem hist[32];
	bool histInitialized[32];
	string pathFilename;

	bool storeImages;
	bool maxContrast;
    bool detectFades;
    bool detectDissolves;
    
	Signal<int32> sigma; 
	SignalSize sigmaSize;
};

#endif
