/*
   This file is part of the MoCA video library.
   Name: FaceRecognition.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef FACE_RECOGNITION_H
#define FACE_RECOGNITION_H

// C++ includes
#include <float.h>
#include <iostream>
#include <string>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
using namespace std;


// MoCa includes
#include <ImageSize.h>
#include <FileList.h>
#include <GrayscaleFilter.h>
#include <Image.h>
#include <IImageStream.h>
#include <OImageStream.h>
#include <ScaleFilter.h>
#include <ImageSize.h>



// *****************************************
// Class: FaceNormalization
// *****************************************


/* FaceNormalizationclass: normalize a face based on:  
 * rotation, skaling, luminance, contast.  
 * The width of the face is defined in the constructor.
 * 
 * @author Stephan Kopf
 */
class FaceNormalization {
public:
	/* constructor
	 * @param width  int describes the width of the detected face
	 */
	FaceNormalization (int32 width = 80)	{
		this->width = width;
	};

	/* constructor
	 * @param imgage   the input image should be a good approximation of a face region
	 */
	void process (const Image<uint8> &src, Image<uint8> &dest);
	void process (Image<uint8> &img)
	{
		Image<uint8> src = img;
		process (src, img);
		return;
	}

private:
	int32 width;
};






// *****************************************
// Class: AverageFace
// *****************************************


/* struct for reading and writing 
 * the average image vector
 * 
 * @author Alexander Oertel
 */
struct AverageFace	{
	/* constructor for writing the average face to disk
	 * @param is ImageSize containing image size
	 * @param af gsl_vector containing the average face
	 */
	AverageFace(const ImageSize &is, gsl_vector *af)	{
		freeMemory = false;
		averageFace = NULL;
	
		imageSize = is;
		averageFace = af;
	};

	/* constructor for reading the average face from disk
	 */
	AverageFace()	{
		freeMemory  = true;
		averageFace = NULL;
	};

	/* destructor
	 */
	~AverageFace(){
		/* freeMemory is true only when memory 
		 * for averageFace is allocated in this class
		 * else averageFace contains a pointer (dont free)
		 */
		if(freeMemory && averageFace != NULL)
			gsl_vector_free(averageFace);
	};

	// holds average image
	gsl_vector *averageFace;

	// holds the image size of averageFace
	ImageSize imageSize;

	// the dimensions of imageSize
	int xDim;
	int yDim;

	// bool for destructor
	bool freeMemory;
};

/* overloading operator<<
 * @param os ostream
 * @param af AverageFace struct
 */
std::ostream& operator<<(ostream &os, const AverageFace &af);

/*overloading operator>>
 * @param is istream
 * @param af AverageFace struct
 */
std::istream& operator>>(istream &is, AverageFace &af);




// *****************************************
// Class: AvgFaceCalculation
// *****************************************


/* AvFaceCalculation class: class that calculates an average image from 
 * the images contained in the sd (sourceDirectory) and writes the calculated 
 * image into the 'outputFile' into the dd (destinationDirectory) if wanted
 * 
 * @author Alexander Oertel
 */
class AvgFaceCalculation	{
public:
	/* constructor
	 * @param sd string containing the directory of the source images
	 * @param dd string containing the directory for the output
	 * @param debug boolean containing if visual output wanted or not
	 */
	AvgFaceCalculation(const string &sd, const string &dd,const bool debug)	{
		averageFace = NULL;
		stop = 0;
		imageSize = ImageSize(0, 0);
	
		calculateAverageImage(sd);
		writeAverageImageVector(dd);

		// only for visual output	
		if(debug)
			writeAverageImage(dd);
	};

	/* destructor
	 */
	~AvgFaceCalculation()	{
		// free memory
		if(averageFace != NULL)
			gsl_vector_free(averageFace);
	};
	
	/* returns the average image as vector
	 * @return averageFace gsl_vector containing average face
	 */
	inline gsl_vector* getAverageImageVector()	{		
		return averageFace;
	}

private:
	/* calculates the average image
	 * @param srcDir string containing the directory of the source images
	 */
	void calculateAverageImage(const string &srcDir);
	
	/* invokes methods for converting all .jpg, .pmn, .png 
	 * images into .pgm gray scale images of rights size
	 * @param srcDir string containing directory of images
	 */
	void convertImages(const string &srcDir, bool getSize);
	
	/* write average image vector to destDir
	 * @param destDir string for output directory
	 */
	void writeAverageImageVector(string destDir);
	
	/* write average image to destDir
	 * @param destDir string for output directory
	 */
	void writeAverageImage(const string &destDir);
	
	// image size containing the size of average image
	ImageSize imageSize;

	// vector containing the average face
	gsl_vector *averageFace;
	
	// integer for exiting
	int stop;
};



// *****************************************
// Class: LightingCorrectionFilter
// *****************************************


/* class that averages the the intensity values
 * in the image
 * 
 * @author Alexander Oertel
 */
template<class T> 
class LightingCorrectionFilter{
public:
	// constructor
	LightingCorrectionFilter(){
		columnSum = NULL;
		rowSum = NULL;
		h = NULL;
		b = NULL;
		image = NULL;
	};
	
	//destructor
	~LightingCorrectionFilter()	{
		if(columnSum != NULL)
			delete[] columnSum;
			
		if(rowSum != NULL)
			delete[] rowSum;
			
		if(h != NULL)
			delete[] h;
			
		if(b != NULL)
			delete[] b;
			
		if(image != NULL)
			gsl_vector_free(image);
	};
	
	 /* if only src image as parameter
	  * @param srcDest image holding the src image
	  */
	 void process(Image<T> &srcDest);
	
	/*! the filter
	 * @param src image containing the src image
	 * @param dest image getting the processed image
	 */
    bool process(const Image<T> &src, Image<T> &dest);
    
private:
	ImageSize is;
	
	int *columnSum;
    int *rowSum;

    double *h;
    double *b;
    
    gsl_vector *image;
};


/*
* The filter itself, provide an input sample block, the data in the
* input sample block gets overwritten with the output data.
*
* This method calls filter(src, dest)
* 
* \sa LightingCorrectionFilter<T>::filter(SampleBlock<T>, SampleBlock<T>)
* 
*@author Alexander Oertel
*
*@param srcDest input and output sample block
*/
template<class T>
void LightingCorrectionFilter<T>::process(Image<T> &srcDest)	{
    Image<T> dest;
    process(srcDest, dest);
    srcDest = dest;
} // process



/* The filter istelf, the input data is read from src, the output data
 * written into dest. If dest does not have the same size as src, dest
 * is resized, but dest has only one band (grayscale image).
 *
 *	@param src source sample block
 *	@param dest destination sample block
 *
 */
template<class T>
bool LightingCorrectionFilter<T>::process(const Image<T> &src, Image<T> &dest)	{
	// get image size
	is = src.getSize();
	double s_x = is[DIM_X];
	double s_y = is[DIM_Y];
	
	// initialize double arrays
	h = new double[4];
	h[0]=h[1]=h[2]=h[3]=0;
	
	b = new double[3];
	b[0]=b[1]=b[2]=0;

	// calculate values for b[1]
	for(int j=0; j<s_x; j++)	{
		for(int i=0; i<s_y; i++)	{
			h[0] += j * src(j, i, 0);
			h[1] += j;
			h[2] += src(j, i, 0);
			h[3] += j*j;
		}
	}
	// calucate b[1]
	b[1] = ((s_x*s_y)*h[0]-h[1]*h[2])/(s_x*s_y*h[3]-h[1]*h[1]);
	
	// reset values for h
	h[0]=h[1]=h[2]=h[3]=0;
		
	// calulate values for b[2]
	for(int j=0; j<s_x; j++)	{
		for(int i=0; i<s_y; i++)	{
			h[0] += i * src(j, i, 0);
			h[1] += i;
			h[2] += src(j, i, 0);
			h[3] += i*i;
		}
	}
	//calculate b[2]
	b[2] = ((s_x*s_y)*h[0]-h[1]*h[2])/(s_x*s_y*h[3]-h[1]*h[1]);
	
	// reset values
	h[0]=h[1]=h[2]=h[3]=0;
	
	// calucate values for b[0]
	for(int j=0; j<s_x; j++)	{
		for(int i=0; i<s_y; i++)	{
			h[0] += src(j, i, 0);
			h[1] += j;
			h[2] += i;
		}
	}
	// calculate b[0]
	b[0] = (h[0] - b[1]*h[1] - b[2]*h[2])/(s_x*s_y);
	
	// allocate memory
	image = gsl_vector_alloc(is[DIM_X]*is[DIM_Y]);
	
	// calculate values of new image
	for(int y=0; y<is[DIM_Y]; y++)	{
		for(int x=0; x<is[DIM_X]; x++)	{
			 gsl_vector_set(image, y*is[DIM_X]+x, b[0] + b[1]*x + b[2]*y);
		}
	}
		
	dest.resize(is);
	double min = DBL_MAX;
	double max = 0;	
	
	// correct original image with values;	
	for(int y=0; y<is[DIM_Y]; y++)	{
		for(int x=0; x<is[DIM_X]; x++)	{
			double buf = src(x,y) - gsl_vector_get(image, y*is[DIM_X]+x) +128;
			
			// check for values that are to small or to big
			if(buf<0)
				buf = 0;
			if(buf>255)
				buf=255;
				
			// write values to image
			dest(x, y, 0) = (uint8) buf;
			 
			// needed for histogramm equalisation
			if(buf<min)
				min  = buf;
			if(buf>max)
				max =buf;
		}
	}
	
	   
	//contrast to small quit filter
	if (max-min < 30) {
	 	return false;
    } // if
    
    
    // histogram equalisation
    float c2 = 0;
    c2 = 255.0/(float)(max - min);  // TODO: max==min ???, e.g. c2=255;
    float c1 = (-(float)min)*c2;

	// rewrite corrected values
	for (int y = 0; y < is[DIM_Y];  y++) {
        for (int x = 0; x < is[DIM_X]; x++) {
        	double value = (int)(c2*dest(x, y) + c1);
            if (value < 0) 
            	value = 0;
            if (value > 255) 
            	value = 255;
            dest(x, y) = (uint8)value;
        } // for
    } // for
    
    return true;
} // lightingCorrection()




// *****************************************
// Class: SizeOccurrence
// *****************************************
/* struct for occurrences of a face 
 * and number of occurrences
 * 
 * @author Alexander Oertel
 */
struct  SizeOccurrence	{
	/*constructor
	 * @param n string containing the name of the face
	 */
	SizeOccurrence(const ImageSize &is)	{
		imageSize = is;
		amount = 1;
	};

	/* destructor
	 */
	~SizeOccurrence()	{}; 
	
	// holding the name of the face
	ImageSize imageSize;
	
	// holding the number of occurrences
	int amount;
};

/* overloading operator==
 * @param o1 first Occurrence to compare
 * @param o2 second Occurrence to compare
 */
bool operator== (const SizeOccurrence &so1, const SizeOccurrence &so2);



// *****************************************
// Class: DifferenceImagesMatrixCalculation
// *****************************************
/* DifferenceImagesMatrixCalculation class: class that 
 * calculates a matrix out of difference images calculated
 * from the training set and the average image
 * 
 * @author Alexander Oertel
 */
class DifferenceImagesMatrixCalculation	{
public:
	/* constructor
	 * @param ptoi string containing path to original images
	 * @param af gsl_vector containing the average face
	 */
	DifferenceImagesMatrixCalculation(const string &ptsi, const gsl_vector *af)	{
		allDifferenceImages = NULL;
		matrix = NULL;
	
		calculateMatrix(ptsi, af);
	};

	/* destructor
	 */
	~DifferenceImagesMatrixCalculation()	{
		// free memory
		if(allDifferenceImages != NULL)	{
			for(int i=0; i<differenceImagesMatrixSize; i++)	{
				gsl_vector_free(allDifferenceImages[i]);
			}
			delete[] allDifferenceImages;
		}
		
		if(matrix != NULL)
			gsl_matrix_free(matrix);
	};

	/* returns the M*M matrix calulated from the difference images
	 * @return matrix gsl_matrix
	 */
	inline gsl_matrix* getDifferenceImagesMatrix() {
		return matrix;
	}

	/* returns the size of the M*M matrix = amount of images
	 * @return int size of matrix containing difference images
	 */
	inline int getDifferenceImagesMatrixSize()	{
		return differenceImagesMatrixSize;
	}

	/* returns the gsl_vector** containing all difference images
	 * @return allDifferenceImages gsl_vector** containing all difference images
	 */
	inline gsl_vector** getAllDifferenceImages()	{
		return allDifferenceImages;
	}

	/* returns the length of an image stored in the gsl_vector **differenceImages
	 * @return imageSize size of average image
	 */
	inline ImageSize getImagesSize()	{
		return imageSize;
	}
private:
	/* calcualtes the M*M matrix out of the difference images, 
	 * the original images and the average face
	 * @param pathToSourceImages string containing path to training set
	 * @param averageFace gsl_vector containing the average face
	 */
	void calculateMatrix(const string &pathToSourceImages, const gsl_vector *averageFace);
	
	/* calculating the single values for the M*M matrix
	 * @param y integer for y position in matrix
	 * @param x integer for x position in matrix
	 * @return double value for M*M matrix
	 */
	double calculateValues(const int &y, const int &x);

	// holding the image size
	ImageSize imageSize;

	// the M*M matrix
	gsl_matrix *matrix;
	
	// vector** for all difference images
	gsl_vector **allDifferenceImages;

	// size of the M*M matrix
	int differenceImagesMatrixSize;
};


// *****************************************
// Class: EigenVectorCalculation
// *****************************************

/* EigenVectorCalculation class; 
 * class that calculates the eigenfaces from the
 * eigenvectors of a given matrix m
 * 
 * @author Alexander Oertel
 */
class EigenVectorCalculation	{
public:
	/* constructor
	 * @param m gsl_matrix containing the M*M matrix 
	 * @param ms integer holding the size of the gsl_matrix m
	 * @param adi gsl_vector** holding all difference images
	 * @param is imageSize holding the size of an image
	 * @param dd string containing the destination directory for the eigenvectors
	 * @param debug boolean for visual output
	 */
	EigenVectorCalculation(gsl_matrix *m, const int &ms, gsl_vector **adi, const  ImageSize is, const string &dd, bool debug)	{
		eigenVectors = NULL;
		result = NULL;
	
		matrixSize = ms;
		allDifferenceImages = adi;
		lengthOfImageVector = is[DIM_X]*is[DIM_Y];

		calculateEigenValuesAndVectors(m);
		multiplyWithDifferenceImages();

		// write visual output of faces before normalisation of the vectors
		if(debug)
			writeFaces(is, dd);

		normalizeEigenVectors();
		writeEigenVectors(dd); 
	};

	/*destructor
	*/
	~EigenVectorCalculation()	{
		// free memory
		if(eigenVectors!=NULL)
			gsl_matrix_free(eigenVectors);
		if(result!=NULL)
			gsl_matrix_free(result);
	};

private:
	/* calculates the eigenvectors and eigenvalues of the gsl_matrix *matrix
	 * @param matrix the matrix from which to calculate the eigenvectors
	 */
	void calculateEigenValuesAndVectors(gsl_matrix *matrix);
	
	/*calculates the eigenvectors with the differenceImages
	*/
	void multiplyWithDifferenceImages();
	
	/* normalizes the eigenvectors to length 1
	 * */
	void normalizeEigenVectors();
	
	/* writes the images to destDir
	 * @param imageSize the size of each eigenface
	 * @param destDir string containing the directory to write to
	 */
	void writeFaces(ImageSize imageSize, const string &destDir);
	
	/* write all eigenvectors to one file
	 * @param output string containing the output directory
	 */
	void writeEigenVectors(string output);

	/* calculates the values of the eigenvectors;
	 * calculates the new value at position (i, j)
	 * @param yCoordinate position in matrix
	 * @param xCoordinate position in matrix
	 * @return value returns the value for the position
	 */
	double calculateValue(const int yCoordinate, const int xCoordinate);

	// holds the eigenvectors
	gsl_matrix *eigenVectors;
	
	//holds the normalized eigenvectors
	gsl_matrix *result;
	
	// holds all difference images
	gsl_vector **allDifferenceImages;
	
	// holds the size of the matrix
	int matrixSize;
	
	// holds the length of an image vector
	int lengthOfImageVector;
};



// *****************************************
// Class: EigenVectors	
// *****************************************

/* struct for reading for reading 
 * or writing the eigenvectors
 * 
 * @author Alexander Oertel
 */
struct EigenVectors	{
	/* constructor for writing the eigenvectors
	 * @param noev integer containing the number of all eigenvectors
	 * @param loev integer holding the length of an eigenvectors
	 * @param ev gsl_matrix holding the eigenvectors
	 */
	EigenVectors(const int &noev, const int &loev, gsl_matrix *ev)	{
		freeMemory = false;
		eigenVectors = NULL;
	
		numberOfEigenVectors = noev;
		lengthOfEigenVectors = loev;
		eigenVectors = ev;
	};

	/* constructor for reading eigenvectors from disk
	 */
	EigenVectors()	{
		freeMemory = true;
		eigenVectors = NULL;
	};

	/* destructor
	 * */
	~EigenVectors()	{
		// free memory
		if(freeMemory && eigenVectors != NULL)
			gsl_matrix_free(eigenVectors);
	};

	// holds the eigenvectors
	gsl_matrix *eigenVectors;

	// holds the number of all eigenvectors
	int numberOfEigenVectors;
	
	// holds the length of an eigenvector
	int lengthOfEigenVectors;

	// boolean for destructor
	bool freeMemory;
};

/* overloading operator<<
 * @param os ostream
 * @param ev EigenVectors struct
 */
std::ostream& operator<<(ostream &os, const EigenVectors &ev);

/*overloading operator>>
 * @param is istream
 * @param ev EigenVectors struct
 */
std::istream& operator>>(istream &is, EigenVectors &ev);



// *****************************************
// Class: FaceWeightsCalculation
// *****************************************

/* FaceWeightsCalculation class; class that 
 * calculates the weight for each eigenface 
 * for the contribution to the new face
 * 
 * @author Alexander Oertel
 */
class FaceWeightsCalculation	{
public:
	/* constructor for face class calculation also writes 
	 * calculated files to disk
	 * @param nf string containing path to new image
	 * @param ptw string containing path to weights
	 * @param loiv integer containing length of an image vector
	 * @param noev integer containing numbers of eigenvectors
	 * @param af gsl_vector containing the average face
	 * @param ev gsl_matrix containing the eigenvectors
	 */
	FaceWeightsCalculation(const string &nf, const string &ptw, const int &loiv, const int &noev, const gsl_vector *af, const gsl_matrix *ev)	{
		difference = NULL;
		weights = NULL;

		/* initialize here that 'calculateDifference(af)'
		 * works for the second constructor too
		 */
		IImageStream<uint8> iis(nf);
		iis.read(src);
	
		// correct lighting here because image is initialized here
		LightingCorrectionFilter<uint8> lcf;
		lcf.process(src);

		// calculate
		calculateDifference(af);
		calculateWeights(loiv, noev, ev);

		// write all weights into files
		writeWeights(nf, ptw, noev);
	};
	
	/* constructor for calculation of distance to face
	 *  space and face class with all eigenvectors
	 * @param s new image
	 * @param loiv integer containing length of an image vector
	 * @param noev integer containing numbers of eigenvectors
	 * @param af gsl_vector containing the average face
	 * @param ev gsl_matrix containing the eigenvectors
	 */
	FaceWeightsCalculation(Image<uint8> s, const int &loiv, const int &noev, const gsl_vector *af, const gsl_matrix *ev)	{
		difference = NULL;
		weights = NULL;
	
		src = s;

		calculateDifference(af);
		calculateWeights(loiv, noev, ev);
	};

	/* destructor
	 */
	~FaceWeightsCalculation()	{
		// free memory
		if(difference != NULL)
			gsl_vector_free(difference);
		if(weights != NULL)
			delete []weights;
	};

	/* returns a pointer to an array containing the calculated weights
	 * @return weights double array containing the weights
	 */
	double* getWeights()	{
		return weights;
	} // getWeights

private:
	/* calculates the difference image of input image and average face
	 * @param averageFace gsl_vector holding the average face
	 */
	void calculateDifference(const gsl_vector *averageFace);
	
	/* calculates the weights with all eigenvectors
	 * @param lengthOfImageVector integer containing the length of an image vector
	 * @param numberOfEigenVectors integer containing the number of all eigenvectors
	 * @eigenVectors gsl_matrix containing the eigenvectors
	 */
	void calculateWeights(const int &lengthOfImageVector, const int &numberOfEigenVectors, const gsl_matrix *eigenVectors);
	
	/* writes the weights of the face classes
	 * @param fileName string containing the actual name of the image
	 * @param numberOfEigenVectors integer containing the number weights to write to a file
	 */
	void writeWeights(string fileName, string path, const int &numberOfEigenVectors);

	// holding the actual image
	Image<uint8> src;

	// holding the difference image
	gsl_vector *difference;

	// holding the weights of the input image
	double *weights;
};



// *****************************************
// Class: Weights
// *****************************************

/* struct for the weight array
 * 
 * @author Alexander Oertel
 */
struct Weights	{
	/* constructor for writing weights to disk
	 * @param fn  string containing file name for weight
	 * @param now integer containing number of weights
	 * @param w double array containing the weights
	 */
	Weights(const string &fn, const int &now, double *w)	{
		fileName = fn;
		numberOfWeights = now;
		weights = w;
	};

	/* constructor for reading weights
	 */
	Weights(){};
	
	/*destructor
	 * */
	~Weights(){};

	// holding the file name of the weight
	string fileName;
	
	// holding number of weights to be written
	int numberOfWeights;
	
	// holding the weights aary
	double *weights;
};

/* overloading the operator<<
 * @param os ostream
 * @param w Weights object
 */
std::ostream& operator<<(ostream &os, const Weights &w);

/*overloading the operator>>
 * @param is istream
 * @param w Weights object
 */
std::istream& operator>>(istream &is, Weights &w);


// *****************************************
// Class: Threshold
// *****************************************

/* struct for the weight array
 * 
 * @author Alexander Oertel
 */
struct Threshold	{
	/* constructor for writing the threshold
	 * @param th double containing the threshold
	 */
	Threshold(const double &th)	{
		threshold	= th;
	};

	/* constructor for reading threshold
	 */
	Threshold(){};

	 /* destructor
	  */
	~Threshold(){};

	// holding the threshold
	double threshold;
};

/* overloading the operator<<
 * @param os ostream
 * @param t Threshold object
 */
std::ostream& operator<<(ostream &os, const Threshold &t);

/* overloading the operator>>
 * @param is istream
 * @param t Threshold object
 */
std::istream& operator>>(istream &is, Threshold &t);



// *****************************************
// Class: ThresholdCalculation
// *****************************************

/* ThresholdCalculation class; class
 *  that calculates threshold between 
 * two faces from the training set
 */
class ThresholdCalculation	{
public:
	/* constructor
	 * @param sd string containing the path to the weights
	 * @param dd string containing the output path
	 */
	ThresholdCalculation(const string &sd)	{
		calculateThreshold(sd);
		writeThreshold(sd);
	};

	 /* destructor
	  */
	~ThresholdCalculation(){};

	/* returns the threshold
	 * @return threshold double containing the threshold
	 */
	double getThreshold()	{
		return threshold;
	} // calculateDistance

private:
	/* calculates the threshold
	 * @param srcDir string containing the path to the weights
	 */
	void calculateThreshold(const string &srcDir);
	
	/* write threshold
	 * @param destDir write threshold to destDir
	 */
	void writeThreshold(const string &destDir);
	
	/* calculate the values for the threshold
	 * @return value double containing the actual threshold
	 */
	double calculate();

	// holds weight one for comparison
	Weights wgt;
	
	// holds second weight of the comparison
	Weights wgt2;

	// holds the threshold
	double threshold;
};



// *****************************************
// Class: FaceClassesCalculation
// *****************************************

/* FaceClasscalculation class; class 
 * that calculates all the face classes 
 * using one face for each class
 * 
 * @author Alexander Oertel
 */
class FaceClassesCalculation
{
public:
	/* constructor
	 * @param sd string containing the source directory
	 * @param dd string containing the path to the data
	 * @param ptw string containing the path for weights
	 */
	FaceClassesCalculation(const string &sd, const string &dd)	{
		averageFace = NULL;
		eigenVectors = NULL;
		dataPath = dd;
	
		// read needed values
		readAverageFace();
		readEigenVectors();

		// calculate the face classes	
		calculateClasses(sd, dd);
	};

	/*destructor
	*/
	~FaceClassesCalculation()	{
		// free memory
		if(averageFace != NULL)
			gsl_vector_free(averageFace);
	
		if(eigenVectors!=NULL)
			gsl_matrix_free(eigenVectors);
	};

private:
	/* reading the eigenvectors from disk
	 */
	void readEigenVectors();
	
	/* reading the average face from disk
	 */
	void readAverageFace();
	
	/* calculates the face classes
	 * @param srcDir string containing the source images
	 * @param pathToWeights string containing the output path for the weight files
	 */
	void calculateClasses(const string &srcDir, const string &pathToWeights);

	// holds the eigenvectors
	gsl_matrix *eigenVectors;
	
	// holds the average face
	gsl_vector *averageFace;

	// holds the length of an image vecotr
	int lengthOfImageVector;
	
	// holds the number of the eigenvectors
	int numberOfEigenVectors;
	
	// holding the dimensions of an image
	int xDim, yDim;
	
	// holds the path where the data lies
	string dataPath;
};

// *****************************************
// Class: 
// *****************************************

// *****************************************
// Class: 
// *****************************************


#endif
