/*
   This file is part of the MoCA video library.
   Name: HistogramDifferenceFeature.h
   Copyright (C) 2004 Markus Beier, Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: beier@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  \author    Markus Beier, Stephan Kopf
  
   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: beier@pi4.informatik.uni-mannheim.de
   
	(c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	
*/

#ifndef HISTOGRAM_DIFFERENCE_FEATURE_H
#define HISTOGRAM_DIFFERENCE_FEATURE_H


//C++ includes
#include <iomanip>
#include <sstream>
#include <vector>

// MoCA includes
#include <Video.h>
#include <VideoDecoder.h>
#include <HistogramFeature.h>
#include <Image.h>
#include <VideoStatistic.h>


//! Class to calculate the differences of the histograms of consecutive frames in a video
/*!
This class will calculate the histograms of the frames of a video. The difference of consecutive 
histograms will be calculated and stored in a vector provided as parameter.
*/

const static string HISTOGRAM_DIFFERENCE_VERSION    ="Version-0.39";
const static string HISTOGRAM_DIFFERENCE_FILENAME   ="histogram.idx";


// a maximum of 60 MB should be allocated for the histograms
// in grayscale videos (bits=3):  32 bytes/frame (~2000000 frames, approx.  20h of video)
// in grayscale videos (bits=5): 128 bytes/frame (~ 500000 frames, approx.   5h of video)
// in color videos (bits=2)    :  64 bytes/frame (~1000000 frames, approx.  10h of video)
// in color videos (bits=3)    :2048 bytes/frame (~  30000 frames, approx. 0,5h of video)
// if you increase the bits, you can keep less histograms in memory (much slower with option "skip") 
const static int32  HISTOGRAM_DIFFERENCE_MAX_MEMORY = 60*1024*1024;



enum HistogramDifferenceMetric {
  HDM_L1,  // L1 norm
  HDM_L2,  // L2 norm
  HDM_EMD  // Earth Movers distance
};



struct HistogramDifferenceItem
{
	HistogramDifferenceItem () 
	{
		msBits             = 3;
		skipedFrames       = 0;
		grayscale          = false;
		optimize           = false;
		normalize          = false;
		frames             = 0;
		minHistValue       = 1e10;
		maxHistValue       = -1 ;
		avgHistValue       = 0;
		histFilename       = "";
		areas			   = 1;
		metric             = HDM_L1;
		values.clear();
	} 
    HistogramDifferenceItem& operator=(const HistogramDifferenceItem &other) {
		if (this != &other) {
			msBits             = other.msBits;
			skipedFrames       = other.skipedFrames;
			grayscale          = other.grayscale;
			optimize           = other.optimize;
			normalize          = other.normalize;
			areas              = other.areas;
			frames             = other.frames;
			minHistValue       = other.minHistValue;
			maxHistValue       = other.maxHistValue;
			avgHistValue       = other.avgHistValue;
			histFilename       = other.histFilename;
			values             = other.values;
			metric             = other.metric;
		} // if
        return *this;
    } // operator=
	  
	uint8 msBits;
	uint8 skipedFrames;
	bool grayscale;
	bool optimize;
	bool normalize;
	int16 areas;
	uint32 frames;
	ieee64 minHistValue;
	ieee64 maxHistValue;
	ieee64 avgHistValue;
	string histFilename;
	vector<ieee64> values;
	HistogramDifferenceMetric metric;
};


struct HistogramDifferenceData
{
	HistogramDifferenceData () 
	{
		version = HISTOGRAM_DIFFERENCE_VERSION;
		filename="";
		numberOfElements = 0;
		hdi.clear();
	} 
    HistogramDifferenceData& operator=(const HistogramDifferenceData &other) {
		if (this != &other) {
			version = other.version;
			filename= other.filename;
			numberOfElements = other.numberOfElements;
			hdi = other.hdi;
		} // if
        return *this;
    } // operator=
	
	string version;
	string filename;
	uint32 numberOfElements;
	vector<HistogramDifferenceItem> hdi;
};

std::ostream& operator<< (ostream&,    const HistogramDifferenceItem&);
OTagStream&   operator<< (OTagStream&, const HistogramDifferenceItem&);
ITagStream&   operator>> (ITagStream&, HistogramDifferenceItem&);

std::ostream& operator<< (ostream&,    const HistogramDifferenceData&);
OTagStream&   operator<< (OTagStream&, const HistogramDifferenceData&);
ITagStream&   operator>> (ITagStream&, HistogramDifferenceData&);



class HistogramDifferenceFeature
{
public:
	static const string DEFAULT_HISTOGRAM_DIFFERENCE_FILE_NAME;
	static const int MAX_SKIP_FRAME_SIZE;

	
    //! Contructor initializing the class with a filename of the MPEG file 
    /*!
	\param fullname of Type string& provides the filename of the mpeg file to be processed
	\param msBits of Type uint8 determines the number of most significant bits to contemplate while calculating the histograms
	\param skipFrames of Type uint8 determines the frames which will be compared, e. g. 
	       skipFrames == 0: compare frames (0,1), (1,2), ...
	       skipFrames == 1: compare frames (0,2), (1,3), ...
	\param grayscale of Type bool determines, if grayscale images should be compared
	\param optimize of Type bool determines, if an optimized histogram calculation method is used
	\sa process(HistogramDifferenceData& hdd)
    */
    HistogramDifferenceFeature (
		string &pathFilename, 
		uint8 msBits = 3, 
		uint8 skipFrames = 0, 
		bool grayscale = false,
		bool optimize=false,
		bool normalizeHistograms=false,
		int16 areas=1,
		HistogramDifferenceMetric histogramDifferenceMetric=HDM_L1);
	
    //! Default destructor
    ~HistogramDifferenceFeature() 
	{
		if (hist!=NULL) delete[] hist;
		hist=NULL;
		if (histFrame!=NULL) delete[] histFrame;
		histFrame = NULL;
		if (reader!=NULL) delete reader;
		reader = NULL;
		if (histCalc!=NULL) delete histCalc;
		histCalc = NULL;		 
	};
	
	
    //! Calculating the difference in the histograms of consecutive (skipFrames==0) frames in the given 
	//! mpeg video and store the differences in a file
    /*!
	\param hdd of Type HistogramDifferenceData& is the structure wherein the resulting data will be stored. 
    */
	void process (HistogramDifferenceItem& hdd);

	
private:
	void readValues (HistogramDifferenceData& data);
	void calculateHistogramDifference (HistogramDifferenceItem &item, int32 skip); 
	void normalize (Signal<uint32> &hist1, Signal<uint32> &hist2);
	void getHistogram (int32 frame, int32 &framePosition);


	string pathFilename;
	string histogramDiffFilename;
	
	uint8 bits;
	uint8 skipedFrames;
	bool grayscale;
	bool optimize;
	bool normalizeHistograms;
	int16 areas;
	HistogramDifferenceMetric histogramDifferenceMetric;

	VideoDecoder *reader;
	ieee64 numberOfPixelFactor;
	int32 numberOfFrames;
	int32 histSize;
	uint32 numberOfPixel;
	Signal<uint32> *hist; // buffer with histogram values
	int32 *histFrame;     // frame number of histogram at index i
	Image<uint8> img;
	int32 nextImage;
	int32 nextPosition;
	int32 bufferSize;
	HistogramFeature *histCalc;

}; //class HistogramDifferenceFeature

#endif


