/*
    HistogramFeature.cpp     calculate histograms
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <HistogramFeature.h>




HistogramFeature::
~HistogramFeature()
{
	if (factor!=NULL) {
		delete[] factor;
	} // if
	factor=NULL;
}


void
HistogramFeature::
processColor (Image<uint8>& img, Signal<uint32>& histogram)
{

    // getting the size info
    indexType height, width, band;
    ImageSize size(img.getSize());
    height = size[DIM_Y];
    width = size[DIM_X];
    band = size[DIM_BAND];

    // calculating the shifting value of the colour values to contemplate only the x most significant bits
    uint8 shiftVal = 8-msbits;

    // calculating the size of the histogram needed to store the histogram
	ieee64 tmpBits = msbits;
	ieee64 tmpBand = band;
	ieee64 tmpSize = 1;
	tmpSize = pow(2.0, tmpBits);
	tmpSize = pow(tmpSize, tmpBand);

    uint32 histsize = (uint32) tmpSize;
	if (histsize<=0) {
		cerr << "HistogramFeature: size of histogram is too large!" << endl;
		cerr << "                  reduce number of significant bits!" << endl << endl;
		exit(1);
	} // if

	histogram.resize(SignalSize(histsize,areas));
    for (int32 a= 0; a < areas; a++) {
		for (uint32 i= 0; i < histsize; i++) {
			histogram (i,a) = 0;
		} // for
	} // for

    //calculating the values
/*	if (areas == 1) {
		if (band == 1) {
			for (int y= 0; y < height; y++) {
				for (int x= 0; x < width; x++) {
					histogram ((img(x,y)) >> shiftVal)++;
				} // width for
			} // height for
		} else {
			// I assume here, that all media treated with this class which have more than one band, 
			// have three bands (usually a colour picture)
			for (int y= 0; y < height; y++) {
				for (int x= 0; x < width; x++) {
					ieee64 value = (img(x,y,0) >> shiftVal) +
						(img(x,y,1) >> shiftVal) * pow(2,msbits) +
						(img(x,y,2) >> shiftVal) * pow(2,2*msbits);
					histogram ((int) value)++;
				} // width for
			} // height for
		} // if - else
	} else {
*/
	int steps=1;
	if (areas>=4)  steps=2;
	if (areas>=9)  steps=3;
	if (areas>=16) steps=4;
	if (areas>=25) steps=5;
	for (int32 px=0; px<steps; px++) {
		for (int32 py=0; py<steps; py++) {
			
			// set borders of small windows
			uint32 left=px*width/steps;
			uint32 right=(1+px)*width/steps;
			uint32 top=py*height/steps;
			uint32 bottom=(1+py)*height/steps;
			if (px==steps-1) right=width;
			if (py==steps-1) bottom=height;
			uint32 pos = px*steps+py;
			
			if (band == 1) {
				for (int32 y= top; y < (int32)bottom; y++) {
					for (int32 x= left; x < (int32)right; x++) {
						histogram ((img(x,y)) >> shiftVal,pos)++;
					} // width for
				} // height for
			} else {
				// I assume here, that all media treated with this class which have more than one band, 
				// have three bands (usually a colour picture)
				for (int32 y= top; y < (int32)bottom; y++) {
					for (int32 x= left; x < (int32)right; x++) {
						ieee64 value = (img(x,y,0) >> shiftVal) +
							(img(x,y,1) >> shiftVal) * pow(2,(ieee64)msbits) +
							(img(x,y,2) >> shiftVal) * pow(2,(ieee64)2*msbits);
						histogram ((uint32) value, pos)++;
					} // width for
				} // height for
			} // if - else
		} // for
	} // for
	return;
} // process



void
HistogramFeature::
processGrayscale (Image<uint8>& img, Signal<uint32>& histogram)
{

    // helper variables to store the size of the image to work on
    indexType height, width;
    ImageSize size(img.getSize());
    height = size[DIM_Y];
    width = size[DIM_X];

    // calculating the size of the histogram needed to store the histogram
    uint16 histSize = 1;
    histSize = (uint16) pow(2,(ieee64)msbits);
	SignalSize sigSize (histSize, areas);
	histogram.resize(sigSize);

	// calculate size of blocks
	uint16 blockSize = 256/histSize;


	int steps=1;
	if (areas>=4)  steps=2;
	if (areas>=9)  steps=3;
	if (areas>=16) steps=4;
	if (areas>=25) steps=5;
	for (int32 px=0; px<steps; px++) {
		for (int32 py=0; py<steps; py++) {
			
			// set borders of small windows
			uint32 left=px*width/steps;
			uint32 right=(1+px)*width/steps;
			uint32 top=py*height/steps;
			uint32 bottom=(1+py)*height/steps;
			if (px==steps-1) right=width;
			if (py==steps-1) bottom=height;
			uint32 pos = px*steps+py;
    
			//calculate histogram values
			int32 tmp [256];
			for (int i= 0; i < 256; i++) {
				tmp[i] = 0;
			} // for
			for (int32 y= top; y < (int32)bottom; y++) {
				for (int32 x= left; x < (int32)right; x++) {
					tmp [img(x,y,0)]++;
				} // for
			} // for

			if (!smoothing) {
				for (int i=0; i<histSize; i++) {
					int32 v=0;
					for (int j=0; j<blockSize; j++) {
						v+=tmp[i*blockSize+j];
					} // for
					histogram(i,pos) = v;
				} // for
				
			} else {
				// example: merge blocks
				// | 0   1   2   3 | 4   5   6   7 | 8   9   10   11  | 
				// |       0       |       1       |       2          | 
				// factors for block 1:
				//          1/8 3/8 5/8 7/8 7/8 5/8 3/8 1/8
				
				for (int i= 0; i < histSize; i++) {
					histogram (i,pos) = 0;
				} // for
				if (factor==NULL) {
					factor = new uint32 [2*blockSize];
					for (int i=0; i<blockSize; i++) {
						factor[i] = 2*i + 1;
						factor[2*blockSize-i-1] = 2*i + 1;
					} // for
				} // if
				int32 sumFactor = 0;
				for (int i=0; i<2*blockSize; i++) {
					sumFactor += factor[i];
				} // for
				sumFactor = sumFactor / blockSize;
				
				
				int32 total=0;
				for (int i=0; i<histSize; i++) {
					
					// summarize values
					int32 sum=0;
					for (int j=0; j<2*blockSize; j++) {
						int32 pHist = i*blockSize+j-blockSize/2;
						if (pHist<0) {
							pHist=-pHist-1;
						} // if
						if (pHist>255) {
							pHist=511-pHist; // = 255 + (256-pHist)
						} // if
						sum += tmp[pHist] * factor[j];
					} // for
					histogram(i,pos) = (int32)(((ieee64)sum) / sumFactor + 0.5);
					total+=histogram(i,pos);
				} // for
				
				// be sure, that number of histogram entries is exact
				int32 p=0;
				while (total>((int32)right-(int32)left)*((int32)bottom-(int32)top)) {
					if (histogram(p,pos)>0) {
						histogram(p,pos)--;
						total--;
					} // if
					p=(p+histSize/2-1)%histSize;
				} // while
				p=0;
				while (total<((int32)right-(int32)left)*((int32)bottom-(int32)top)) {
					histogram(p,pos)++;
					total++;
					p=(p+histSize/2-1)%histSize;
				} // while
								
			} // if
		} // for
	} // for

	return;
} // process


/*
template<class T>
void
HistogramFeature<T>::
process (string &filename, Signal<int32>& histogram, bool saveData)
{
	// read data
	HistogramData data;
	ReadMocaData<HistogramData> reader (filename, HISTOGRAM_FILENAME);
	bool dataValid = reader.process (HISTOGRAM_VERSION, data); 
	if (dataValid) {
		if ((data.msBits != msbits)       || (data.grayscale != grayscale) || 
			(data.smoothing != smoothing) || (data.areas != areas)) {
			dataValid=false;
		} // if
	} // if

	int32 numberOfHistValues=0;
	int32 numberOfFrames=0;
	if (dataValid) {
		histogram = data.data;

	} else {
		// calculate histogram

		// open video
removed
//VideoCodec codec = CODEC_DVDVIEW_YUV;
		if (grayscale) {
			codec = CODEC_DVDVIEW_Y;
		} // if
		VideoDecoder reader (filename, codec, 3, -1);
		
		// get histogram of first image 
		numberOfFrames=reader.getSize()[DIM_Z];
		Image<uint8> img = reader.getImage(0);
		
		Signal<int32> data;
		process(img, data);
		numberOfHistValues=data.getSize()[DIM_X];
		histogram.resize(SignalSize(numberOfFrames, areas*numberOfHistValues));

		// copy values from first imgage
		for (int32 a=0; a<areas; a++) {
			for (int32 k=0; k<numberOfHistValues; k++) {
				histogram (0,a*numberOfHistValues+k) = data(k,a);
			} // for
		} // for
		
		// process video
		//cout << "frame 29" << endl;
		for (int32 i=1; i<numberOfFrames; i++) {
			img = reader.getImage(i);
			cerr << "\rframe: " << i << ends;
			process(img, data);

			// copy values from first imgage
			for (int32 a=0; a<areas; a++) {
				for (int k=0; k<numberOfHistValues; k++) {
					histogram (i,a*numberOfHistValues+k) = data(k,a);
				} // for
			} // for

		} // for
	} // if
	//cout << "end" << endl;


	if ((saveData) && (!dataValid)) {
		data.version   = HISTOGRAM_VERSION;
		data.msBits    = msbits;
		data.grayscale = grayscale;
		data.smoothing = smoothing;
		data.areas     = areas;
		data.histSize  = numberOfHistValues;
		data.frames    = numberOfFrames;
		data.data      = histogram;

		WriteMocaData<HistogramData> writer (filename, HISTOGRAM_FILENAME);
		writer.process(data);
	} // if

	return;
} // process
*/
