/*
   This file is part of the MoCA video library.
   Name: ObjectMaskFeature.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef OBJECT_MASK_FEATURE_H
#define OBJECT_MASK_FEATURE_H

// C++ includes
#include <string>
#include <iostream>
using namespace std;

// MoCA includes
#include <VideoDecoder.h>
#include <CutDetectionFeature.h>
#include <CameraModel.h>
#include <IYuvVideo.h>
#include <OYuvVideo.h>
#include <RegionFeature.h>
#include <ContourParametrization.h>
#include <CameraBackgroundFeature.h>
#include <DifferenceFilter.h>
#include <IImageStream.h>
#include <ContourEdgeList.h>
#include <DrawDot.h>
#include <LogoDetection.h>
#include <TrackingFeature.h>



const int16 DEFAULT_NUMBER_OF_CONTOUR_PIXEL = ContourParametrization::DEFAULT_CONTOUR_SIZE;

// defines the object mask of one frame
struct ObjectMaskItem
{

	ObjectMaskItem (int32 frame, int32 shot)
	{
		this->frame = frame;
		this->shot  = shot;

		borderMinX = 0;
		borderMinY = 0;
		borderMaxX = 0;
		borderMaxY = 0;
		centerX    = 0;
		centerY    = 0;
		length     = 0;

		reliability = 0;
		
		SignalSize size (ContourParametrization::DEFAULT_CONTOUR_SIZE,2);
		contour.realloc (size, SignalSize(8), BORDER_PERIODIC);
		for (int32 i=0; i<ContourParametrization::DEFAULT_CONTOUR_SIZE; i++) {
			contour (i,0) = 0;
			contour (i,1) = 0;
		} // for
	}

    ObjectMaskItem& operator=(const ObjectMaskItem &other) {
		if (this != &other) {

			frame      = other.frame;
			shot       = other.shot;
			contour    = other.contour;

			borderMinX = other.borderMinX;
			borderMinY = other.borderMinY;
			borderMaxX = other.borderMaxX;
			borderMaxY = other.borderMaxY;
			centerX    = other.centerX;
			centerY    = other.centerY;
			length     = other.length;

			reliability = other.reliability;
		} // if
        return *this;
    } // operator=

	int32 frame;
	int32 shot;

	int32 borderMinX;
	int32 borderMinY;
	int32 borderMaxX;
	int32 borderMaxY;
	int32 centerX;
	int32 centerY;
	int32 length;     // length of the contour

	int32 reliability;  // value ranges between 0..1000 (0=not reliable, 1000=very reliable)

	Signal<int16> contour;
};



static const string OBJECT_MASK_VERSION  = "Version-1.00";
static const string OBJECT_MASK_FILENAME = "objectmask.idx";

struct ObjectMaskData
{
	ObjectMaskData ()
	{
		filename = "";
		version = OBJECT_MASK_VERSION;

		cropLeft   = 0;
		cropRight  = 0;
		cropTop    = 0;
		cropBottom = 0;

		numberOfItems = 0;
	}
	string version;
	string filename;

	int16 cropLeft;
	int16 cropRight;
	int16 cropTop;
	int16 cropBottom;

	int32 numberOfItems;
	list<ObjectMaskItem> frameData;
};


std::ostream& operator<< (ostream&,    const ObjectMaskItem&);
OTagStream&   operator<< (OTagStream&, const ObjectMaskItem&);
ITagStream&   operator>> (ITagStream&, ObjectMaskItem&);

std::ostream& operator<< (ostream&,    const ObjectMaskData&);
OTagStream&   operator<< (OTagStream&, const ObjectMaskData&);
ITagStream&   operator>> (ITagStream&, ObjectMaskData&);



class ObjectMask
{
public:

	// analyse AVI-/MPEG video
	ObjectMask (string &srcFilename, int minimumShotLength=12, bool preciseCalculation=false);

	// analyse one shot
	ObjectMask (VideoSize &vidSize, bool preciseCalculation, DefaultObjectPosition objPos=POSITION_UNKNOWN);


	~ObjectMask()
	{
		//TODO: ???
//		deleteMetadataFiles();
	}

	// process AVI-/MPEG video
	void process (ObjectMaskData &data);

	// process one shot
	void process (const Video<uint8> &vid, ObjectMaskData &data);

	void storeFrames (ObjectMaskData &videoObjects);
	void validateResults (ObjectMaskData &videoObjects);

private:

	void readValuesAvi (ObjectMaskData &videoObjects);
	void readValuesShot (const Video<uint8> & vid, ObjectMaskData &videoObjects);

	void selectFrames (int32 firstFrame, int32 lastFrame, const CameraModelData &cameraParameters,
		               int32 & bgSizeX, int32 &bgSizeY, int32 &bgShiftX, int32 &bgShiftY, Signal<int32> &frameNumberList);

	void initializeFrameList(CameraModelData &cameraParameters, Signal<bool> &frameSelected);
	void analyseMpegVideo(ObjectMaskData &videoObjects, CameraModelData &cameraParameters, 
		                  Signal<bool> &frameSelected);
	void getContour (const Signal<int32> &frameNumberList, const Video<uint8> &maskVid, 
					 const int32 shotNumber, ObjectMaskData &videoObjects);
	void getObjectSize (Image<uint8> &img, int32 &minX, int32 &maxX, int32 &minY, int32 &maxY);
	bool calculateObjects (const Video<uint8> &vid, 
						   const int32 bgSizeX, const int32 bgSizeY, const int32 bgShiftX, const int32 bgShiftY,
						   const int32 shotNumber, Image<uint8> &background, Video<uint8> &maskVid);
	void calcBackgroundSize(int32 firstFrame, int32 lastFrame, const CameraModelData &cameraParameters, 
							int32 &minX, int32 &maxX, int32 &minY, int32 &maxY, int32 &numberOfValidFrames);

	bool readYuvVideo(Signal<int32> &frameNumberList, Video<uint8> &vid, VideoDecoder &mpegReader);
	bool readYuvShot (Signal<int32> &frameNumberList, Video<uint8> &vid, const Video<uint8> &shot);




	// ------------------------------------------------

	void selectValidFrames(CameraModelData &cameraParameters, ObjectMaskData &videoShots);



	bool getPgmBackgroundImage(ObjectMaskItem &item, Image<uint8> &img);
	bool calculateExactBackground (ObjectMaskItem &item);


	bool analyseShot (ObjectMaskItem &item);
	void checkEmptyBorders (Video<uint8> &vid, int32 frame, int32 &minX, int32 &maxX, int32 &minY, int32 &maxY);


public:

	VideoSize srcVidSize;
	VideoSize analyseVidSize;
	string pathFilename;
	string destPath;
	int16 thMinimumShotLength;

	int16 cropLeft;
	int16 cropRight;
	int16 cropTop;
	int16 cropBottom;

	int32 thMaxBackgroundWidth;
	int32 thMaxBackgroundHeight;
	int32 thMaxFramesPerShot;
	bool preciseCalculation;
	DefaultObjectPosition defaultObjPos;
};

#endif
