/*
   This file is part of the MoCA video library.
   Name: SceneDetection.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef SCENE_DETECTION_H
#define SCENE_DETECTION_H

// C++ includes
#include <iomanip>
#include <sstream>
#include <iostream>
using namespace std;


// MoCA includes
#include <MocaSignal.h>
#include <HistogramDifferenceFeature.h>
#include <CutDetectionFeature.h>
#include <OMpegVideo.h>
#include <VideoStatistic.h>


// windows includes
#ifdef COMPILER_MSVC
#include <process.h>
#endif


const static string SCENE_DETECTION_VERSION  ="Version-0.08";
const static string SCENE_DETECTION_FILENAME ="scenes.idx";

const static int SCENE_DETECTION_HIST_AREAS = 9;
const static int SCENE_DETECTION_HIST_MSBITS_COLOR = 3;
const static int SCENE_DETECTION_HIST_MSBITS_GRAY = 5;
const static int SCENE_DETECTION_HIST_BINS_COLOR = 512; // 2^3^3 = 512
const static int SCENE_DETECTION_HIST_BINS_GRAY = 32;  // 2^5   =  32

#include <SceneDetectionDefaultScenes.h>

struct SceneDetectionItem
{
	SceneDetectionItem () 
	{
		sceneID         = -1;
		firstFrame      = 0;
		lastFrame       = 0;
		referenceFrame  = 0;
		irrelevantFrame = false;
	} 
	SceneDetectionItem (int32 id, int32 first, int32 reference, int32 last, bool irrelevant) 
	{
		sceneID         = id;
		firstFrame      = first;
		referenceFrame  = reference;
		lastFrame       = last;
		irrelevantFrame = irrelevant;
	} 
    SceneDetectionItem& operator=(const SceneDetectionItem &other) {
		if (this != &other) {
			sceneID         = other.sceneID;
			firstFrame      = other.firstFrame;
			lastFrame       = other.lastFrame;
			hist            = other.hist;
			referenceFrame  = other.referenceFrame;
			irrelevantFrame = other.irrelevantFrame;
		} // if
        return *this;
    } // operator=

    inline bool operator< (const SceneDetectionItem& other) const {
		if (sceneID == other.sceneID) {
			return firstFrame< other.firstFrame;
		} // if
		return (sceneID< other.sceneID);
	}

	int32 sceneID;
	int32 firstFrame;
	int32 lastFrame;
	int32 referenceFrame;
	bool irrelevantFrame;
	Signal<uint32> hist;

	int32 clusterDistance;              // required for the calculation of the clusters, temporary
};

struct SceneDetectionData
{
	SceneDetectionData () 
	{
		version = SCENE_DETECTION_VERSION;
		this->filename = "";
		numberSceneItems = 0;
		numberOfScenesIDs = 0;
		scenes.clear();
		usePredefinedClusters = true;
	} 
    SceneDetectionData& operator=(const SceneDetectionData &other) {
		if (this != &other) {
			version               = other.version;
			filename              = other.filename;
			numberSceneItems      = other.numberSceneItems;
			numberOfScenesIDs     = other.numberOfScenesIDs;
			scenes                = other.scenes;
			usePredefinedClusters = other.usePredefinedClusters;
		} // if
        return *this;
    } // operator=

	string version;
	string filename;
	int32 numberSceneItems;
	int32 numberOfScenesIDs;             // with different sceneID
	bool usePredefinedClusters;

	list<SceneDetectionItem> scenes;
};

std::ostream& operator<< (ostream &, const SceneDetectionData&);
OTagStream& operator<< (OTagStream&, const SceneDetectionData&);
ITagStream& operator>> (ITagStream&, const SceneDetectionData&);





/*
	Constructor: initializes variables
	
	process(): 
		if status file does not exist:
		   calculate border of scenes from mpeg-file, save status file
		else:
		   read borders of scenes from status file
*/
class SceneDetection
{
public:

	// usePredefinedClusters: defines white, black, or gray frames (irrelevant clusters)
	SceneDetection(const string &pathFilename, int numberOfClusters, bool usePredefinedClusters);
	~SceneDetection() {};

	void process (SceneDetectionData&);

private:
	
	void readValues (SceneDetectionData& scd);
//	void initializeScenes (SceneDetectionData& data, CutDetectionData &cuts, Signal<int32> &hist);
//	void mergeScenes (SceneDetectionData& data, Signal<int32> &hist, int32 &histThreshold);
	//int32 histDiff (int32 firstFrame, int32 secondFrame, Signal<int32> &hist);
	int32 histDiff (Signal<uint32> &hist1, Signal<uint32> &hist2);
	int32 getHistDist (int32 firstBorder, int32 lastBorder, Signal<uint32> &hist);
	void initializeHistograms (SceneDetectionData& data);
	void clusterScenes (SceneDetectionData& data);
	void initializePredefinedClusters (list<SceneDetectionItem> &clusters);


    bool usePredefinedClusters;
	int numberOfClusters;
	string pathFilename;
	int32 numberOfPixel;      // number of pixel in image (frame)
	int32 videoSize;          // size of the video
	SignalSize histSize;      // size of the histogram
	bool grayscale;
};

#endif


