/*
    TextDetection.cpp     detect text
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <TextDetection.h>


std::ostream& operator<< (ostream &os, const TextDetectionBox &data) 
{
	os << "minX : " << data.minX << endl;
	os << "maxX : " << data.maxX << endl;
	os << "minY : " << data.minY << endl;
	os << "maxY : " << data.maxY << endl;

	return os;
} // operator <<

std::ostream& operator<< (ostream &os, const TextDetectionData &data) 
{
	os << "version  : " << data.version << endl;
	os << "filename : " << data.filename << endl;
	for (int32 i=0; i<data.box.size(); i++) {
		os << (data.box[i]) << endl;
	} // for

	return os;
} // operator <<

OTagStream& operator<< (OTagStream& os, const TextDetectionData &data) 
{
    TagFormat::writeStartTag (os, "TextArea");
	os << endl;

    TagFormat::writeNodeValue(os, "version", data.version, 1);
    TagFormat::writeNodeValue(os, "filename", data.filename, 1);
    TagFormat::writeNodeValue(os, "numberOfBoxes", (int32)data.box.size(), 1);
	os << endl;

	for (int32 i=0; i<data.box.size(); i++) {
	    TagFormat::writeNodeValue(os, "firstFrame",data.box[i].firstFrame, 1);
	    TagFormat::writeNodeValue(os, "lastFrame",data.box[i].lastFrame, 1);
	    TagFormat::writeNodeValue(os, "minX", data.box[i].minX,  1);
	    TagFormat::writeNodeValue(os, "maxX", data.box[i].maxX,  1);
	    TagFormat::writeNodeValue(os, "minY", data.box[i].minY,  1);
	    TagFormat::writeNodeValue(os, "maxY", data.box[i].maxY,  1);
	    TagFormat::writeNodeValue(os, "rel",  data.box[i].rel,   1);
		os << endl;
	} // for

    TagFormat::writeEndTag (os, "TextArea");
	os << endl;
	
    return os;
} // operator<<


ITagStream& operator>> (ITagStream& is, TextDetectionData& data)
{
	
    char c;
    bool failFlag = true;
	
	if (is >> c) {
		failFlag = false;
		is.putback(c);

		TagFormat::readStartTag (is, "TextArea");
		
		TagFormat::readNodeValue(is, "version", data.version);
		TagFormat::readNodeValue(is, "filename", data.filename);
		TagFormat::readNodeValue(is, "numberOfBoxes", data.numberOfBoxes);

		for (int32 i=0; i<data.numberOfBoxes; i++) {
			TextDetectionBox box;
			TagFormat::readNodeValue(is, "firstFrame",box.firstFrame);
			TagFormat::readNodeValue(is, "lastFrame",box.lastFrame);
			TagFormat::readNodeValue(is, "minX", box.minX);
			TagFormat::readNodeValue(is, "maxX", box.maxX);
			TagFormat::readNodeValue(is, "minY", box.minY);
			TagFormat::readNodeValue(is, "maxY", box.maxY);
			TagFormat::readNodeValue(is, "rel",  box.rel);
			data.box.push_back(box);
		} // for

		TagFormat::readEndTag (is, "TextArea");
	} // if

	if (is.rdstate() & ios::failbit) {
		throw MocaException ("TextDetection: Illegal index file format!");
	} else {
        is.clear();
    } // else
	
    return is;
} // operator>> 


void getStatistics (Signal<uint8> &values, ieee64 &variance)
{
	int32 size = values.getSize()[DIM_X];

	// get avg 
	ieee64 avg=0;
	for (int32 i=0; i<size; i++) {
		avg+=values(i);
	} // for
	avg /= size;
	
	// get variance 
	variance = 0;
	for (int32 i=0; i<size; i++) {
		ieee64 d=(values(i)-avg);
		variance += d*d;
	} // for
	variance /= size;

	return;
} // getStatistics 


void 
TextDetection::filterText (const Image<uint8> &img, Image<uint8> &region)
{
	ImageSize size = img.getSize();
	if (size[DIM_BAND] != 1) {
		cerr << "Error in TextDetection::filterText: only grayscale images allowed!" << endl;
		exit(1);
	} // if

	region.resize(size);
	region.clear(0);

	// feature 1: variance of histograms in radius 3
	SignalSize histSize(256/16);
	Signal<uint8> hist (histSize);

	for (int32 y=0; y<size[DIM_Y]; y++) {
		cout << y << endl;
		for (int32 x=0; x<size[DIM_X]; x++) {

			// get local histograms
			hist.clear(0);
			for (int32 yy=-3; yy<=3; yy++) {
				for (int32 xx=-3; xx<=3; xx++) {
					hist(img(x+xx,y+yy)/16)++;
				} // for
			} // for

			// calc variance
			ieee64 variance=0;
			getStatistics (hist, variance);

#if 0 
			ieee64 v=255;
			ieee64 thMin = 20;
			ieee64 thMax = 55;
			if (variance<thMin) {
				v = 255.0 - 255.0*(thMin - variance)/ thMin;
			} // if
			if (variance>thMax) {
				v = 255.0 - 255.0*(variance-thMax)/ (255.0-thMax);
			} // if
#else
			ieee64 v = 1.8*variance;
#endif

			if (v<0) v=0.0;
			if (v>255) v=255.0;
			region(x,y) =(uint8)v;
		} // for
	} // for

	cout << endl;

	return;
} // void



void 
TextDetection::process(TextDetectionData& data) 
{
	// read data
	ReadMocaData<TextDetectionData> reader (pathFilename, TEXT_DETECTION_FILENAME);
	bool dataValid = reader.process (TEXT_DETECTION_VERSION, data); 
	
	if (!dataValid) {
		
		// create new index
		cerr << "\rTextDetection: Creating index file ...          " << endl;
		readValues(data);
		
		// save index file
		WriteMocaData<TextDetectionData> writer (pathFilename, TEXT_DETECTION_FILENAME);
		writer.process(data);
		cerr << "\rTextDetection: Creating index file ... done          " << endl;
	} // if

	if (store) {
		storeFrame(data);
	} // if
	return;
} // process

void 
TextDetection::storeFrame(TextDetectionData& data)
{

	// open reader
	VideoCodec codec = CODEC_RGB;
	VideoDecoder reader (pathFilename, 3, -1, codec);
	VideoSize vidSize = reader.getSize();
	CreateMocaPath mocaPath (pathFilename);
	string path = mocaPath.destinationPath;

	
	// calculate text areas
	Image<uint8> img;
	DrawRectangle draw(Color::RED);

	for (int32 frame=0; frame<vidSize[DIM_Z]; frame++) {

		bool textFound = false;
		for (int32 i=0; i<data.box.size(); i++) {
			if ((frame >=data.box[i].firstFrame) && (frame <=data.box[i].lastFrame)) {
				textFound = true;
				break;
			} // if
		} // for

		if (textFound) {
			img = reader.getImage(frame);
			for (int32 i=0; i<data.box.size(); i++) {
				if ((frame >=data.box[i].firstFrame) && (frame <=data.box[i].lastFrame)) {

					int32 minX = data.box[i].minX;
					int32 maxX = data.box[i].maxX;
					int32 minY = data.box[i].minY;
					int32 maxY = data.box[i].maxY;
					
					ImagePosition p0 (minX-1, minY-1);
					ImagePosition p1 (maxX+1, maxY+1);
					draw.process(p0,p1, img);
				} // if
			} // for

			// store jpeg image
			string fn = path + FileHandling::getNumberedFilename ("Text", frame, ".jpg");
			WriteImage (img, fn);
		} // if
		cerr << "\rstore frame: " << frame << "  " << ends;
	} // for
	cerr << endl;

	return;
}


void summarizeDifference (const Image<uint8> &img, const Image<uint8> &dx, const Image<uint8> &dy, Signal<int32> &sumX, Signal<int32> &sumY)
{
	// dx and dy must have same resolution
	ImageSize size = dx.getSize();
	if (size != dy.getSize()) {
		cerr << "error in summarizeDifference: the size of dx and dy is different!" << endl;
		exit(1);
	} // if
	size[DIM_BAND]=1;

	// initialize sumX, sumY
	SignalSize sizeX(size[DIM_X]);
	SignalSize sizeY(size[DIM_Y]);
	Signal<int32> xProjection(sizeX);
	Signal<int32> yProjection(sizeY);
	xProjection.clear(0);
	yProjection.clear(0);
	
	// calculate projection
	for (int32 y=0; y<size[DIM_Y]; y++) {
		for (int32 x=0; x<size[DIM_X]; x++) {
			xProjection(x) += dx(x,y);
			yProjection(y) += dy(x,y);
		} // for
	} // for


	// get average and remove all values by this average value
	int32 avgX=0;
	int32 avgY=0;
	for (int32 i=0; i<sizeX[DIM_X]; i++) {
		avgX += xProjection(i);
	} // for
	for (int32 i=0; i<sizeY[DIM_X]; i++) {
		avgY += yProjection(i);
	} // for
	avgX /= sizeX[DIM_X];
	avgY /= sizeY[DIM_X];
	for (int32 i=0; i<sizeX[DIM_X]; i++) {
		if (xProjection(i) < avgX) {
			xProjection(i) = 0;
		} else {
			xProjection(i) -= avgX;
		} // else
	} // for
	for (int32 i=0; i<sizeY[DIM_X]; i++) {
		if (yProjection(i) < avgY) {
			yProjection(i) = 0;
		} else {
			yProjection(i) -= avgY;
		} // else
	} // for


	// non-maxima supression: 
	for (int32 i=0; i<sizeX[DIM_X]; i++) {
		int32 v = xProjection(i);
		if (xProjection(i-5) < v) xProjection(i-5) = 0;
		if (xProjection(i-4) < v) xProjection(i-4) = 0;
		if (xProjection(i-3) < v) xProjection(i-3) = 0;
		if (xProjection(i-2) < v) xProjection(i-2) = 0;
		if (xProjection(i-1) < v) xProjection(i-1) = 0;
		if (xProjection(i+1) < v) xProjection(i+1) = 0;
		if (xProjection(i+2) < v) xProjection(i+2) = 0;
		if (xProjection(i+3) < v) xProjection(i+3) = 0;
		if (xProjection(i+4) < v) xProjection(i+4) = 0;
		if (xProjection(i+5) < v) xProjection(i+5) = 0;
	} // for
	for (int32 i=0; i<sizeY[DIM_X]; i++) {
		int32 v = yProjection(i);
		if (yProjection(i-5) < v) yProjection(i-5) = 0;
		if (yProjection(i-4) < v) yProjection(i-4) = 0;
		if (yProjection(i-3) < v) yProjection(i-3) = 0;
		if (yProjection(i-2) < v) yProjection(i-2) = 0;
		if (yProjection(i-1) < v) yProjection(i-1) = 0;
		if (yProjection(i+1) < v) yProjection(i+1) = 0;
		if (yProjection(i+2) < v) yProjection(i+2) = 0;
		if (yProjection(i+3) < v) yProjection(i+3) = 0;
		if (yProjection(i+4) < v) yProjection(i+4) = 0;
		if (yProjection(i+5) < v) yProjection(i+5) = 0;
	} // for




	// get maximum
	ieee64 fx = -1;
	for (int32 i=0; i<sizeX[DIM_X]; i++) {
		if (fx < xProjection(i)) {
			fx = xProjection(i);
		} // if
	} // for
	ieee64 fy = -1;
	for (int32 i=0; i<sizeY[DIM_X]; i++) {
		if (fy < yProjection(i)) {
			fy = yProjection(i);
		} // if
	} // for
	if (fx <= 0.0) fx = 1.0;
	if (fy <= 0.0) fy = 1.0;
	fx = 255.0 / fx;
	fy = 255.0 / fy;


	// copy values to sumX and sumY (normalized to 0..255)
	sumX.resize(sizeX);
	sumY.resize(sizeY);
	for (int32 i=0; i<sizeX[DIM_X]; i++) {
		ieee64 v = fx*xProjection(i);
		if (v<0) v=0;
		if (v>255) v=255;
		sumX(i) = (uint8)(v);
	} // for
	for (int32 i=0; i<sizeY[DIM_X]; i++) {
		ieee64 v = fy*yProjection(i);
		if (v<0) v=0;
		if (v>255) v=255;
		sumY(i) = (uint8)(v);
	} // for


#if 0
	// show results
	{
		ImageSize tmpSize(size[DIM_X]+255+1, size[DIM_Y]+255+1,1);
		Image<uint8> tmp (tmpSize);
		tmp.clear(0);
		tmp.copy(img,ImagePosition(0,256));

		DrawLine line (Color::GRAY);
		for (int32 x=0; x<size[DIM_X]; x++) {
			ImagePosition p1(x,256);
			ImagePosition p2(x,256-sumX(x));
			line.process(p1,p2,tmp);
		} // for
		for (int32 y=0; y<size[DIM_Y]; y++) {
			ImagePosition p1(size[DIM_X],256+1+y);
			ImagePosition p2(size[DIM_X]+sumY(y),256+1+y);
			line.process(p1,p2,tmp);
		} // for

		string fn=WRITE_DEBUG_PATH;
		fn += "lines.png";
		WriteImage (tmp, fn);
	}
#endif

	return;
} // summarizeDifference 



void 
analyseTextLine (const Image<uint8> &img, Image<uint8> &area) 
{

#if 0
	// debug: show results
	{
		string fn="C:/temp/";
		fn += "small.png";
		WriteImage (img, fn);
	}
#endif

	// region growing
	uint8 regionTh = 3;
	Image<uint32> regionNumbers;
	RegionFeature<uint8> regionGrowing (regionTh);
	regionGrowing.process(img, regionNumbers);
	ImageSize size = img.getSize();
	size[DIM_BAND]=1;

	// ignore small regions (less than 10 pixels)
	vector<RegionType> regionVector;
	regionGrowing.getRegionVector(regionVector);
	uint32 regionIgnore = regionVector.size()+5;
	for (int32 y=0; y<size[DIM_Y]; y++) {
		for (int32 x=0; x<size[DIM_X]; x++) {
			uint32 r = regionNumbers (x,y);
			uint32 np = regionVector[r].numPixels;
			if (regionVector[r].numPixels < 10) {
				regionNumbers (x,y) = regionIgnore;
			} // if
		} // for
	} // for


#if 0
	// debug: show results
	{
		Image<uint8> tmp (size);
		tmp.clear();
		for (int32 y=0; y<size[DIM_Y]; y++) {
			for (int32 x=0; x<size[DIM_X]; x++) {
				uint32 v = regionNumbers(x,y);
				v = v % 256;
				tmp(x,y) = (uint8) v;
			} // for
		} // for


		string fn="C:/temp/";
		fn += "regions.png";
		WriteImage (tmp, fn);
	}
#endif


	// select all regions at the top or bottom of the image
	Signal<int32> selectedRegion(regionIgnore);
	selectedRegion.clear();
	for (int32 x=0; x<size[DIM_X]; x++) {
		uint32 r = regionNumbers (x,0);
		if (r<regionIgnore) selectedRegion(r)++;
		r = regionNumbers (x,size[DIM_Y]-1);
		if (r<regionIgnore) selectedRegion(r)++;
	} // for

	area.resize(size);
	area.clear(255);
	for (int32 y=0; y<size[DIM_Y]; y++) {
		for (int32 x=0; x<size[DIM_X]; x++) {
			uint32 r = regionNumbers (x,y);
			if (selectedRegion(r) > 0) {
				area(x,y) = 0;
			} // if
		} // for
	} // for

#if 0
	// debug: show results
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "region-final.png";
		WriteImage (area, fn);
	}
#endif

	return;
} // analyseArea



void 
analysePeak (const int32 maxPosition, const int32 thV, const Signal<int32> &sumY, const Image<uint8> &img, const Image<uint8> &dx, 
			 int32 &startPosition, int32 &endPosition, Image<uint8> &region) 
{
	startPosition = 0;
	endPosition = 0;
	SignalSize size = sumY.getSize();

	// analyse lower text area (a possible text area)
	int32 startLower = maxPosition+1;
	int32 endLower=startLower+1;
	while ((endLower<size[DIM_X]) && (sumY(endLower) < thV) && (sumY(endLower) >=0)) endLower++;

	// other text line found?
	bool skipLower=false;
	if (sumY(endLower) <0) {
		skipLower=true;
	} // if
	// charcters too small (textheight)
	if (endLower-startLower<4) {
		skipLower=true;
	} // if

	// init variables
	Image<uint8> small;
	Image<uint8> lowerArea;
	Image<uint8> upperArea;

	int32 pixelLower=0;
	if (!skipLower) {
		if (endLower-startLower>8) {
			startLower++;
			endLower--;
		} // if
		{
			CropFilter<uint8> crop(0,0,startLower,img.getSize()[DIM_Y] - endLower);
			crop.process(img, small);
		}
		
		analyseTextLine(small, lowerArea);

		// count horizontal text pixel
		uint32 yPosLower = lowerArea.getSize()[DIM_Y]/2;
		for (int32 i=0; i<size[DIM_X]; i++) {
			if (lowerArea(i,yPosLower) == 0) {
				pixelLower++;
			} // if
		} // for

	} // if


	// analyse upper text area (a possible text area)
	int32 startUpper = maxPosition-1;
	int32 endUpper=startUpper-1;
	while ((endUpper>0) && (sumY(endUpper) < thV) && (sumY(endUpper) >= 0)) endUpper--;

	// other text line found?
	bool skipUpper=false;
	if (sumY(endUpper) <0) {
		skipUpper=true;
	} // if
	// charcters too small (textheight)
	if (startUpper-endUpper<4) {
		skipUpper=true;
	} // if


	int32 pixelUpper=0;
	if (!skipUpper) {
		if (startUpper-endUpper>8) {
			startUpper--;
			endUpper++;
		} // if
		{
			CropFilter<uint8> crop(0,0,endUpper,img.getSize()[DIM_Y] - startUpper);
			crop.process(img, small);
		}
		
		analyseTextLine(small, upperArea);

		// count horizontal text pixel
		uint32 yPosUpper = upperArea.getSize()[DIM_Y]/2;
		for (int32 i=0; i<size[DIM_X]; i++) {
			if (upperArea(i,yPosUpper) == 0) {
				pixelUpper++;
			} // if
		} // for
	} // if


	uint8 select = 0;
	// at least 5 text pixels should be available
	if (pixelLower > 5) select +=1;
	if (pixelUpper > 5) select +=2;

	if (select == 3) {

		// both areas are possible, get average of dx of both areas
		ieee64 avgLower = 0.0;
		int32 n=0;
		ImageSize tmpSize = lowerArea.getSize();
		for (int32 y=0; y<tmpSize[DIM_Y]; y++) {
			for (int32 x=0; x<tmpSize[DIM_X]; x++) {
				int32 d = dx(x,y+startLower);
				if (d>16) avgLower += d;
				n++;
			} // for
		} // for
		if (n==0) n=1;
		avgLower/=n;

		ieee64 avgUpper = 0.0;
		n=0;
		tmpSize = upperArea.getSize();
		for (int32 y=0; y<tmpSize[DIM_Y]; y++) {
			for (int32 x=0; x<tmpSize[DIM_X]; x++) {
				int32 d = dx(x,y+endUpper);
				if (d>16) avgUpper += d;
				n++;
			} // for
		} // for
		if (n==0) n=1;
		avgUpper/=n;

		select = 1;
		if (avgLower < avgUpper) {
			select = 2;
		} // if

	} // if

	if (select == 1) {
		// copy lowerArea to region
		ImageSize tmpSize = lowerArea.getSize();
		for (int32 y=0; y<tmpSize[DIM_Y]; y++) {
			for (int32 x=0; x<tmpSize[DIM_X]; x++) {
				region(x,y+startLower) = lowerArea(x,y);
			} // for
		} // for
		
		startPosition = startLower;
		endPosition = endLower;
	} // if

	if (select == 2) {
		// copy upperArea to region
		ImageSize tmpSize = upperArea.getSize();
		for (int32 y=0; y<tmpSize[DIM_Y]; y++) {
			for (int32 x=0; x<tmpSize[DIM_X]; x++) {
				region(x,y+endUpper) = upperArea(x,y);
			} // for
		} // for
		
		startPosition = endUpper;
		endPosition = startUpper;
	} // if

#if 0
	// debug: show results
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "region-large.png";
		WriteImage (region, fn);
	}
#endif

	return;
}

void 
TextDetection::analyze (const Image<uint8> &img, const Image<uint8> &dx, const Image<uint8> &dy, Image<uint8> &region) 
{
	// calculate sums
	ImageSize imgSize = img.getSize();
	imgSize[DIM_BAND] = 1;
	region.resize(imgSize);
	region.clear(255);

	Signal<int32> sumX;
	Signal<int32> sumY;
	summarizeDifference (img, dx, dy, sumX, sumY);

	bool maximumFound=true;
	while (maximumFound) {

		// locate maximum horizontal sum
		SignalSize size = sumY.getSize();
		int32 maxPosition=0;
		for (int32 i=0; i<size[DIM_X]; i++) {
			if (sumY(maxPosition) < sumY(i)) {
				maxPosition = i;
			} // if
		} // for
		int32 thV = sumY(maxPosition) / 3 + 1;
		if (sumY(maxPosition)<1) {
			maximumFound=false;
			break;
		} // if
		
		// detect next text line
		int32 startPosition;
		int32 endPosition;
		analysePeak(maxPosition, thV, sumY, img, dx, startPosition, endPosition, region);

		// mark text line (-1:text line, -2:border near text line)
		sumY(maxPosition)=0;
		if (endPosition > startPosition) {
			for (int32 i=startPosition; i<=endPosition; i++) {
				sumY(i) = -1;
			} // for
			if (startPosition-1>=0) sumY(startPosition-1) = -2;
			if (startPosition-2>=0) sumY(startPosition-2) = -2;
			if (startPosition-3>=0) sumY(startPosition-3) = -2;
			if (endPosition+1<size[DIM_X]) sumY(endPosition+1) = -2;
			if (endPosition+2<size[DIM_X]) sumY(endPosition+2) = -2;
			if (endPosition+3<size[DIM_X]) sumY(endPosition+3) = -2;
		} // if

	} // while
		

#if 0
	// debug: show results
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "region-largefinal.png";
		WriteImage (region, fn);
	}
#endif

	// calculate exact character height
	uint8 regionTh = 3;
	Image<uint32> regionNumbers;
	RegionFeature<uint8> regionGrowing (regionTh);
	regionGrowing.process(img, regionNumbers);

	vector<RegionType> regionVector;
	regionGrowing.getRegionVector(regionVector);
	uint32 selectedRegionSize = regionVector.size()+5;
	Signal<uint8> selectedRegions(selectedRegionSize);
	selectedRegions.clear(0);

	for (int32 y=0; y<imgSize[DIM_Y]; y++) {
		for (int32 x=0; x<imgSize[DIM_X]; x++) {
			if (region(x,y) == 0) {
				uint32 r = regionNumbers (x,y);
				selectedRegions(r)=1;
			} // if
		} // for
	} // for


	Image<uint8> newRegion(imgSize);
	newRegion.clear(255);
	for (int32 y=0; y<imgSize[DIM_Y]; y++) {
		for (int32 x=0; x<imgSize[DIM_X]; x++) {
			uint32 r = regionNumbers (x,y);
			if (selectedRegions(r)==1) {
				newRegion(x,y) = 0;
			} // if
		} // for
	} // for

#if 0
	// debug: show results
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "newregion.png";
		WriteImage (newRegion, fn);
	}
#endif

	
	SignalSize histSize(256);
	Signal<ieee64> hist(histSize);
	Image<uint8> colorRegion(imgSize);
	colorRegion.clear(0);
	int32 maxP[3];
	for (int32 band=0; band<img.getSize()[DIM_BAND]; band++) {

		hist.clear();
		for (int32 y=0; y<imgSize[DIM_Y]; y++) {
			for (int32 x=0; x<imgSize[DIM_X]; x++) {
				if (newRegion(x,y) == 0) {
					hist(img(x,y,band))++;
				} // if
			} // for
		} // for
		
		// get maximum
		maxP[band]=0;
		int32 maxV= (int32)hist(maxP[band]);
		for (int32 y=1; y<256; y++) {
			if (maxV < hist(y)) {
				maxP[band] = y;
				maxV = (int32)hist(maxP[band]);
			} // if
		} // for

		// copy values to color image
		for (int32 y=0; y<imgSize[DIM_Y]; y++) {
			for (int32 x=0; x<imgSize[DIM_X]; x++) {
				if (newRegion(x,y) != 0) {
					int32 diff = abs(maxP[band]-img(x,y,band));
					if (diff>0) {
						colorRegion(x,y) = 255;
					} // if
				} // if
			} // for
		} // for
	} // for


#if 0
	// debug: show results
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "colorregion.png";
		WriteImage (colorRegion, fn);
	}
#endif


	// increase character width 
	Image<uint8> colorRegionBold;
	ErosionFilter<uint8> ef(3);
	ef.process(colorRegion, colorRegionBold);


#if 0
	// debug: show results
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "colorregionBold.png";
		WriteImage (colorRegionBold, fn);
	}
#endif

	// increase character width 
	Image<uint8> colorRegionOpen;
	OpeningFilter<uint8> of(3);
	of.process(colorRegion, colorRegionOpen);


#if 0
	// debug: show results
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "colorregionOpen.png";
		WriteImage (colorRegionOpen, fn);
	}
#endif

	return;
} // 





void 
TextDetection::getTextRegion (const Image<uint8> &input, Image<uint8> &textRegion, TextDetectionData& data) 
{

	// initialize textData
	data.box.clear();
	data.filename="";
	data.numberOfBoxes=0;


	// crop src
	const int32 INIT_CROP_SIZE = 4;
	Image<uint8> src;
	src=input;
#if 0
	{
		CropFilter<uint8> crop (INIT_CROP_SIZE,INIT_CROP_SIZE,INIT_CROP_SIZE,INIT_CROP_SIZE);
		crop.process(input, src);
	}
#endif


	// get derivation in x and y direction
	ImageSize size = src.getSize();
	size[DIM_BAND]=1;
	Image<uint8> dx (size);
	Image<uint8> dy (size);
	uint8 maxDx=0;
	uint8 maxDy=0;
	Image<uint8> gray;
	if (src.getSize()[DIM_BAND]==1) {
		gray = src;
		for (int32 y=0; y<size[DIM_Y]; y++) {
			for (int32 x=0; x<size[DIM_X]; x++) {
				uint8 v = abs(src(x,y)-src(x+1,y));
				dx(x,y) = v;
				if (maxDx<v) maxDx=v;
				
				v = abs(src(x,y)-src(x,y+1));
				dy(x,y) = v;
				if (maxDy<v) maxDy=v;
			} // for
		} // for
	} else {
		GrayscaleFilter<uint8> filter;
		filter.process(src,gray);

		for (int32 y=0; y<size[DIM_Y]; y++) {
			for (int32 x=0; x<size[DIM_X]; x++) {

				// get maximum of difference 
				uint8 v0 = abs(gray(x,y)-gray(x+1,y));
				uint8 v1 = abs(src(x,y,0)-src(x+1,y,0));
				uint8 v2 = abs(src(x,y,1)-src(x+1,y,1));
				uint8 v3 = abs(src(x,y,2)-src(x+1,y,2));
				uint8 v4 = max(v0,v1);
				uint8 v5 = max(v2,v3);
				uint8 v = max(v4,v5);
				dx(x,y) = v;
				if (maxDx<v) maxDx=v;

				// get maximum of difference 
				v0 = abs(gray(x,y)-gray(x,y+1));
				v1 = abs(src(x,y,0)-src(x,y+1,0));
				v2 = abs(src(x,y,1)-src(x,y+1,1));
				v3 = abs(src(x,y,2)-src(x,y+1,2));
				v4 = max(v0,v1);
				v5 = max(v2,v3);
				v  = max(v4,v5);
				dy(x,y) = v;
				if (maxDy<v) maxDy=v;
			} // for
		} // for
	} // else

	// normalize derivation to 0..255
	if (maxDx > 0) {
		ieee64 f = 255.0 / maxDx;
		for (int32 y=0; y<size[DIM_Y]; y++) {
			for (int32 x=0; x<size[DIM_X]; x++) {
				dx(x,y) = (uint8)(f * dx(x,y));
			} // for
		} // for
	} // if
	if (maxDy > 0) {
		ieee64 f = 255.0 / maxDy;
		for (int32 y=0; y<size[DIM_Y]; y++) {
			for (int32 x=0; x<size[DIM_X]; x++) {
				dy(x,y) = (uint8)(f * dy(x,y));
			} // for
		} // for
	} // if





	// initlialize textRegion
	textRegion.resize(size);
	textRegion.clear(0);

#if 0
	// debug: show results
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "dx.png";
		WriteImage (dx, fn);
	}
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "dy.png";
		WriteImage (dy, fn);
	}
#endif


	// find horizontal lines without text 
	analyze (src, dx, dy, textRegion);
#if 0
	// debug: show results
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "region_1.png";
		WriteImage (textRegion, fn);
	}
#endif


	// remove non-edge pixel
	uint8 thRemove=60;
	for (int32 y=0; y<size[DIM_Y]; y++) {
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (dx(x,y)<thRemove) {
				dx(x,y) = 0;
			} // if
			if (dy(x,y)<thRemove) {
				dy(x,y) = 0;
			} // if
		} // for
	} // for

#if 0
	// debug: show results
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "dx1.png";
		WriteImage (dx, fn);
	}
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "dy1.png";
		WriteImage (dy, fn);
	}
#endif

	// analyse color for edge pixels
	uint32 colorHist[256];
	for (int32 i=0; i<256; i++) colorHist[i]=0;
	uint8 thSimilarTextcolor=32;
	for (int32 y=1; y<size[DIM_Y]-1; y++) {
		for (int32 x=1; x<size[DIM_X]-1; x++) {
			if (dx(x,y)>0) {
				uint8 left0 = gray(x,y);
				uint8 left1 = gray(x-1,y);
				uint8 left2 = gray(x-2,y);

				uint8 leftMin=min(min(left0,left1),left2);
				uint8 leftMax=max(max(left0,left1),left2);
				int32 leftDiff = leftMax-leftMin;


				uint8 right0 = gray(x+1,y);
				uint8 right1 = gray(x+2,y);
				uint8 right2 = gray(x+3,y);

				uint8 rightMin=min(min(right0,right1),right2);
				uint8 rightMax=max(max(right0,right1),right2);
				int32 rightDiff = rightMax-rightMin;

				// this pixel is not the border of a text region
				if ((leftDiff > thSimilarTextcolor) && (rightDiff > thSimilarTextcolor)) {
					dx(x,y) = 0;
				} // if


				if (leftDiff < thSimilarTextcolor) {
					// get median pixel
					uint8 color = max(min(left0,left1),min(left0,left2));
					colorHist[color]++;
				} // if
				if (rightDiff < thSimilarTextcolor) {
					// get median pixel
					uint8 color = max(min(right0,right1),min(right0,right2));
					colorHist[color]++;
				} // if
			} // if
		} // for
	} // for

	for (int32 i=0; i<256; i++) {
		cout << colorHist[i] << endl;
	} // for

#if 0
	// debug: show results
	{
		string fn=WRITE_DEBUG_PATH;
		fn += "dx2.png";
		WriteImage (dx, fn);
	}
	{
		exit(1);
	}
#endif


	// analyse image based on blocks counting the edges in the image
	Image<uint8> edge;
	Image<ieee64> edgeF;
	{
		CannyFilter<uint8> canny;
		canny.process(src, edge);

		ConvertFilter<uint8, ieee64> convert(CONVERT_SCALE, 0, 255, 0, 255);
		convert.process(edge, edgeF);
	}
//	analyseEdgeImage (edge, src, data);




#if 0
	// debug: show results
	{
		string filename = "C:/dev/vortrag/img_out/image-1.png";
		storeImage(edge, filename, data);
	}
#endif
	

	// find exact text boundaries
	for (int32 m=0; m<data.box.size(); m++) {
		TextDetectionBox box1 = data.box[m];

		// crop edge image
		int32 cropLeft   = box1.minX;
		int32 cropTop    = box1.minY;
		int32 cropRight  = size[DIM_X] - box1.maxX -1;
		int32 cropBottom = size[DIM_Y] - box1.maxY -1;

		Image<uint8> cropSrc;
		Image<uint8> cropDx;
		Image<uint8> cropDy;
		{
			CropFilter<uint8> crop(size, cropLeft, cropRight, cropTop, cropBottom);
			crop.process(src, cropSrc);
			crop.process(dx,  cropDx);
			crop.process(dy,  cropDy);
		}


#if 0
	// debug: show results
		{
			string fn1 = "C:/dev/vortrag/img_out/" + FileHandling::getNumberedFilename ("cropSrc_", m, ".png");
			WriteImage (cropSrc, fn1);
			string fn2 = "C:/dev/vortrag/img_out/" + FileHandling::getNumberedFilename ("dx_", m, ".png");
			WriteImage (cropDx, fn2);
			string fn3 = "C:/dev/vortrag/img_out/" + FileHandling::getNumberedFilename ("dy_", m, ".png");
			WriteImage (cropDy, fn3);
		}
#endif

	} // for
	data.numberOfBoxes = data.box.size();


	return;
}



void 
TextDetection::readValues (TextDetectionData& data) 
{
	// set variables
	data.version = TEXT_DETECTION_VERSION;

	// get path of filename, do not create new directory
	CreateMocaPath mocaPath (pathFilename, false);
	data.filename = mocaPath.file;
	data.box.clear();

	// open reader
	VideoCodec codec = CODEC_Y;
	VideoDecoder reader (pathFilename, 3, -1, codec);
	VideoSize vidSize = reader.getSize();
	ImageSize imgSize = reader.getImageSize();

	
	// calculate text areas
	Image<uint8> img;
	for (int32 i=0; i<vidSize[DIM_Z]; i++) {

		//img = reader.getImage(i);
		// do something
		// code temporary removed (add code)

		// debug
		if (i==1) {
			TextDetectionBox box;
			box.firstFrame = 0;
			box.lastFrame = 106;
			box.minX = 115;
			box.maxX = 327;
			box.minY = 38;
			box.maxY = 112;
			box.rel  = 0.9; 
			data.box.push_back (box);
		} // if
		if (i==2) {
			TextDetectionBox box;
			box.firstFrame = 0;
			box.lastFrame = 106;
			box.minX = 60;
			box.maxX = 324;
			box.minY = 196;
			box.maxY = 207;
			box.rel  = 0.9; 
			data.box.push_back (box);
		} // if

		cerr << "\rframe: " << i << "  " << ends;

	} // for
	cerr << endl;


	return;	
} // readValues


