/*
   This file is part of the MoCA video library.
   Name: VideoAbstractFeature.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef VIDEO_ABSTRACT_FEATURE_H
#define VIDEO_ABSTRACT_FEATURE_H

// C++ includes
#include <string>
#include <iostream>
using namespace std;

// MoCA includes
#include <OMpegVideo.h>
#include <IWaveAudio.h>
#include <OWaveAudio.h>
#include <AudioDecoder.h>
#include <OMpegAudio.h>
#include <OMpegMplex.h>
#include <VideoAbstractFeature.h>
#include <CutDetectionFeature.h>
#include <FaceDetection.h>
#include <ActionDetection.h>
#include <SceneDetection.h>
#include <AudioEnergy.h>
//#include <AudioConvert.h>
#include <AudioCut.h>



enum SceneChangeMode {
	SCENE_CHANGE_HARD_CUT = 0, 
	SCENE_CHANGE_FADE     = 1, 
	SCENE_CHANGE_DISSOLVE = 2, 
	SCENE_CHANGE_WIPE     = 3
};

enum WipeMode {
	WIPE_RIGHT_LEFT = 0, 
	WIPE_LEFT_RIGHT = 1, 
	WIPE_TOP_DOWN   = 2, 
	WIPE_DOWN_TOP   = 3 
};

struct SceneChange
{
	// constructor for hard cuts
	SceneChange (int32 frameNo) 
	{
		frame = frameNo;
		mode = SCENE_CHANGE_HARD_CUT;
		sceneChangeLength = 0;
		fadeColor = 0;
		wipeDirection = WIPE_RIGHT_LEFT;
	} // SceneChange

	// general constructor 
	SceneChange (int32 frameNo, SceneChangeMode sceneMode, int16 length, uint8 fadingColor, WipeMode wipeDir=WIPE_RIGHT_LEFT)
	{
		frame = frameNo;
		mode = sceneMode;
		sceneChangeLength = length;
		fadeColor = fadingColor;
		wipeDirection = wipeDir;
	} // SceneChange


    SceneChange& operator=(const SceneChange &other) {
		if (this != &other) {
			frame = other.frame;
			mode = other.mode;
			sceneChangeLength = other.sceneChangeLength;
			fadeColor = other.fadeColor;
			wipeDirection = other.wipeDirection;
		} // if
        return *this;
    } // operator=

	int32 frame;
	SceneChangeMode mode;
	int16 sceneChangeLength;
	uint8 fadeColor;        // required for fades	
	WipeMode wipeDirection; // required for wipes
};



enum SceneSelection {SELECTION_UNKNOWN, SELECTED, IGNORED};

struct SceneDescription
{
	SceneDescription (int32 firstFrame, int32 lastFrame, ieee64 framerate, int32 sceneID) 
	{
		this->firstFrame=firstFrame;
		this->lastFrame=lastFrame;
		this->sceneID=sceneID;
		numberOfFrames=lastFrame-firstFrame+1;

		lengthSeconds        = numberOfFrames /framerate;
		histogramDifferences = -1.0;

		priorityLength   =0;
		priorityAction   =0;
		priorityContrast =0;
		priorityFaces    =0;
		priorityCluster  =0;
		priorityPosition =0;
		summarizedPriority = 0;

		selected=SELECTION_UNKNOWN;
	} 

    SceneDescription& operator=(const SceneDescription &other) {
		if (this != &other) {
			firstFrame     =other.firstFrame;
			lastFrame      =other.lastFrame;
			sceneID        =other.sceneID;
			numberOfFrames =other.numberOfFrames;

			lengthSeconds        =other.lengthSeconds;
			histogramDifferences =other.histogramDifferences;

			priorityLength  =other.priorityLength;
			priorityAction  =other.priorityAction;
			priorityContrast=other.priorityContrast;
			priorityFaces   =other.priorityFaces;
			priorityCluster =other.priorityCluster;
			priorityPosition=other.priorityPosition;

			summarizedPriority=other.summarizedPriority;

			selected =other.selected;
		} // if
        return *this;
    } // operator=

	void updateSummarizedPriority ()
	{
		summarizedPriority = 
			priorityAction +
			priorityCluster +
			priorityContrast +
			priorityFaces +
			priorityLength +
			priorityPosition;
		if (selected == IGNORED) summarizedPriority-=1000;
	} // updateSummarizedPriority

	SceneSelection getSelected() 
	{
		return selected;
	} // getSelected
	SceneSelection getSelected() const
	{
		return selected;
	} // getSelected


	void setSelected(SceneSelection s) 
	{
		selected=s;
		return; 
	} // setSelected

	// values
	int32 firstFrame;
	int32 lastFrame;
	int32 numberOfFrames;


	// derivated values
	ieee64 lengthSeconds;           // e. g. 10.5 seconds
	ieee64 histogramDifferences;    // 0.5 in scenes with much action

	/*

	ieee64 largeFaces;              // large face >= 0.5*min [videoheight, videowidth] 
	                                // largeFaces = SUM (number of large faces in each frame) / numberOfFrames
	ieee64 mediumFaces;             // medium face >= 0.25*min [videoheight, videowidth] 
	ieee64 smallFaces;              // small face  <  0.25*min [videoheight, videowidth] 

	ieee64 cameraMotion;    // if there is one significant camera motion (pan, tilt), the value is larger than zero
	                        // the value is defined by:
	                        //   (number of pixel the image moves) / (imagesize in x- or y-direction) [in percent]
	                        // the largest camera motion is stored 
	                        // example: image moves 3 pixel/frame to the left, hotizontal size=300 => cameraMotion=100*3/300=1.0 percent 
	                        

	ieee64 cameraZoom;      // if there is one significant zooming operation, the value is larger than zero
	                        // the value is defined by:
							// largestMotionVector=search largest motion vector in the image 
	                        // imageZoom=largestMotionVector/
	                        // 
	                        //   (max (length of motionvectors of block) / (imagesize in x- or y-direction)
	                        // the largest camera motion is stored 

*/
	// other categories: position of faces


	// priorities
	uint8 priorityLength;
	uint8 priorityAction;
	uint8 priorityContrast;
	uint8 priorityFaces;
	uint8 priorityCluster;
	int32 priorityPosition;
	int32 summarizedPriority;

	int32 sceneID;

private:
	SceneSelection selected;  
};



const static string VIDEO_DESCRIPTION_VERSION  ="Version-0.01";

struct VideoDescription
{
	VideoDescription (ieee64 frameRate=25.0) 
	{
		scenes.clear();

		version = VIDEO_DESCRIPTION_VERSION;
		
		numberOfScenes  = 0;
		numberOfFrames  = 0;
		numberOfSeconds = 0;  

		numberOfSelectedScenes  = 0;
		numberOfSelectedFrames  = 0;
		numberOfSelectedSeconds = 0;  

		numberOfIgnoredScenes  = 0;
		numberOfIgnoredFrames  = 0;
		numberOfIgnoredSeconds = 0;  

		this->frameRate = frameRate;
		if (frameRate<1.0) {
			frameRate=1.0;
		} // if

		audioSelectedFrames = 0;
		audioSelectedSeconds = 0;

	} 
	~VideoDescription () 
	{
		scenes.clear();
	} 

    VideoDescription& operator=(const VideoDescription &other) {
		if (this != &other) {
			scenes = other.scenes;
			version          = other.version;
			numberOfScenes   = other.numberOfScenes;
			numberOfFrames   = other.numberOfFrames;
			numberOfSeconds  = other.numberOfSeconds;

			numberOfSelectedScenes   = other.numberOfSelectedScenes;
			numberOfSelectedFrames   = other.numberOfSelectedFrames;
			numberOfSelectedSeconds  = other.numberOfSelectedSeconds;

			numberOfIgnoredScenes   = other.numberOfIgnoredScenes;
			numberOfIgnoredFrames   = other.numberOfIgnoredFrames;
			numberOfIgnoredSeconds  = other.numberOfIgnoredSeconds;

			//numberOfFramesMin   = other.numberOfFramesMin;
			//numberOfSecondsMin  = other.numberOfSecondsMin;

			frameRate        = other.frameRate;

			audioSelectedFrames     = other.audioSelectedFrames;
			audioSelectedSeconds    = other.audioSelectedSeconds;
			audioSelectedFramesList = other.audioSelectedFramesList;
		} // if
        return *this;
    } // operator=

	void setFrameRate(ieee64 rate) {
		frameRate=rate;
	} // setFrameRate

	void addScene(SceneDescription& scene) {
		scenes.push_back(scene);
		numberOfScenes++;
		numberOfFrames+=scene.numberOfFrames;
		numberOfSeconds=numberOfFrames/frameRate;  

		//numberOfFramesMin+=scene.numberOfFramesMin;
		//numberOfSecondsMin=numberOfFramesMin/frameRate;
		return;
	} // addScene

	void selectScene(SceneDescription& scene) {
		unignoreScene(scene);
		if (scene.getSelected()!=SELECTED) {
			scene.setSelected(SELECTED);
			numberOfSelectedScenes++;
			numberOfSelectedFrames+=scene.numberOfFrames;
			numberOfSelectedSeconds=numberOfSelectedFrames/frameRate;  
		} // if
		return;
	} // selectScene

	void unselectScene(SceneDescription& scene) {
		if (scene.getSelected()==SELECTED) {
			scene.setSelected(SELECTION_UNKNOWN);
			numberOfSelectedScenes--;
			numberOfSelectedFrames-=scene.numberOfFrames;
			numberOfSelectedSeconds=numberOfSelectedFrames/frameRate;  
		} // if
		return;
	} // unselectScene

	void ignoreScene(SceneDescription& scene) {
		unselectScene(scene);
		if (scene.getSelected()!=IGNORED) {
			scene.setSelected(IGNORED);
			numberOfIgnoredScenes++;
			numberOfIgnoredFrames+=scene.numberOfFrames;
			numberOfIgnoredSeconds=numberOfSelectedFrames/frameRate;  
		} // if
		return;
	} // selectScene

	void unignoreScene(SceneDescription& scene) {
		if (scene.getSelected()==IGNORED) {
			scene.setSelected(SELECTION_UNKNOWN);
			numberOfIgnoredScenes--;
			numberOfIgnoredFrames-=scene.numberOfFrames;
			numberOfIgnoredSeconds=numberOfSelectedFrames/frameRate;  
		} // if
		return;
	} // unselectScene


	void getTimeCode (int32 frameNo, int32 &ms)
	{
		ms=(int32)(1000.0*frameNo/frameRate);
		return;
	} // getTimeCode
	
	void getTimeCode (int32 frameNo, string &ms)
	{
		int32 h=(int32)(frameNo/(frameRate*3600));
		int32 m=(int32)((frameNo%((int32)frameRate*3600))/(frameRate*60));
		int32 s=(int32)((frameNo%((int32)frameRate*60))/frameRate);
		int32 f=(frameNo%(int32)frameRate);
		stringstream str;
		if (h<10) str<<"0";
		str << h << ":";
		if (m<10) str<<"0";
		str << m << ":";
		if (s<10) str<<"0";
		str << s << ":";
		if (f<10) str<<"0";
		str << f;
		ms=str.str();
		return;
	} // getTimeCode


	void updateAudioSelectedSeconds (AudioCutData &audioCuts);
	
	void printSelected () {
		list<SceneDescription>::const_iterator it;
		cout << endl << "selected shots:" << endl;
		string timecode1, timecode2;
		for (it=scenes.begin(); it!=scenes.end(); it++) {
			if ((*it).getSelected() == SELECTED) {
				getTimeCode ((*it).firstFrame, timecode1);
				getTimeCode ((*it).lastFrame,  timecode2);
				cout << timecode1 << "-" << timecode2 << "   "
					<< (*it).firstFrame << ".." << (*it).lastFrame << endl;
			} // if
		} // for
		return;
		cout << endl;
	} // printSelected

	void printAudioSelected () {

		int32 numberOfFrames = audioSelectedFramesList.getSize()[DIM_X];
		int32 i=0;
		cout << endl << "selected shots (audio cuts included):" << endl;
		while (i<numberOfFrames) {

			while ((i<numberOfFrames) && (audioSelectedFramesList(i)!=1)) i++;
			int32 firstFrame = i;
			while ((i<numberOfFrames) && (audioSelectedFramesList(i)==1)) i++;

			if (firstFrame < i-1) {
				string timecode1, timecode2;
				getTimeCode (firstFrame, timecode1);
				getTimeCode (i-1,  timecode2);
				cout << timecode1 << "-" << timecode2 << "   "
					<< firstFrame << ".." << (i-1) << endl;
			} // if
		} // while
		cout << endl;

		return;
	} // printSelected

	/*
	void removeScene(SceneDescription& scene) {
		if (!scene.remove) {
			numberOfScenes--;
			numberOfFrames-=scene.numberOfFrames;
			numberOfSeconds=numberOfFrames/frameRate;  
			scene.remove=true;	
		} // if
		return;
	} // addScene
	*/

	string version;

	list<SceneDescription> scenes;
	ieee64 frameRate;  

	int32 numberOfScenes;
	int32 numberOfFrames;
	ieee64 numberOfSeconds;  

	int32 numberOfSelectedScenes;
	int32 numberOfSelectedFrames;
	ieee64 numberOfSelectedSeconds;  

	int32 numberOfIgnoredScenes;
	int32 numberOfIgnoredFrames;
	ieee64 numberOfIgnoredSeconds;  


	//int32 numberOfFramesMin;
	//ieee64 numberOfSecondsMin;  

	// data for selected frames, after the analysis of the audio track
	int32 audioSelectedFrames;
	ieee64 audioSelectedSeconds;
	Signal<uint8> audioSelectedFramesList; // 0=unknown, 1=selected, 2=not selected
};

std::ostream& operator<< (ostream &os, const SceneDescription &data);
std::ostream& operator<< (ostream &, const VideoDescription&);
OTagStream& operator<< (OTagStream&, const VideoDescription&);
ITagStream& operator>> (ITagStream&, const VideoDescription&);


/*
	Constructor: initializes variables
	
	process():  do abstracting
*/

class VideoAbstract;
std::ostream& operator<< (ostream&, const VideoAbstract&);


struct VideoAbstractParameters
{
	VideoAbstractParameters () 
	{
		recurse             = false;
		faces               = true;
		deleteMetadata      = false;
		deleteTemporaryData = true;
		storeExtensionAbs   = false;
		createAbstract      = true;

		lengthMinutes = -1;
		lengthPercent = -1;
		srcFilename   = "";
		destFilename  = "";

		AVDocumentFilename = "";
		videoLevelFilename = "";
		VDXFilename = "";
	}
	bool recurse;
	bool faces;
	bool deleteMetadata;
	bool deleteTemporaryData;
    bool storeExtensionAbs;
    bool createAbstract;

	ieee64 lengthMinutes;
	ieee64 lengthPercent;
	string srcFilename;
	string destFilename;

	string AVDocumentFilename;
	string videoLevelFilename;
	string VDXFilename;
};

class VideoAbstract
{
public:

	VideoAbstract(const string &srcFilename, const string &destFilename, VideoAbstractParameters &param);
	~VideoAbstract();

	void process ();

	void createFadeIn   (VideoDecoder &reader, int32 firstFrame, int16 numberOfFrames, Video<uint8> &vid, uint8 color);
	void createFadeOut  (VideoDecoder &reader, int32 lastFrame,  int16 numberOfFrames, Video<uint8> &vid, uint8 color);
	void createDissolve (VideoDecoder &reader, int32 lastFrameFirstScene, int32 firstFrameSecondScene, int16 numberOfFrames, Video<uint8> &vid);
	void createDissolve (Image<uint8> &img1, Image<uint8> &img2, int16 numberOfFrames, Video<uint8> &vid);

private:
	//void selectScene(int32 firstFrame, int32 lastFrame);
	//void analyseAudio (Signal<int32> &silence);
	void selectFrames();
	void mergeImages(Image<uint8> &src1, Image<uint8> &src2, uint8 percentImage1, Image<uint8> &dest);

	void setPriority(ActionData &hist, VideoStatisticData &statistic, FaceData &faces);
	void setPriorityLength();

	void createMpegVideo();
	void createMpegAudio();
	void multiplexAudioVideo();
	void updateClusterInformation(Signal<uint8> &selectedFrames);
	void updatePositionInformation(Signal<uint8> &selectedFrames);
	void scalePriorityValues();
	void deleteMetadataFiles();
	void deleteTemporaryFiles();


public:

	string srcFilename;
	string destFilename;

	string destMv1File;
	string destWavFile;
	string destMp2File;
	string srcAudioWavFile;

	VideoSize  srcSize;
	ImageSize  imgSize;
	int32 numberOfFrames;
	//SignalSize audioSize;
	//WaveHeader audioHeader;

	int32 videoFrames;
	int32 lastFrameToDecode;
	ieee64 frameRate;
	VideoDescription videoScenes;
	ieee64 srcSeconds;
	ieee64 srcMinutes;
	ieee64 destSeconds;
	ieee64 destPercent;
	int16 numberOfCluster;

	bool noAudio;
	bool analyseFaces;
	bool deleteMetadata;
	bool deleteTemporaryData;
	bool createGrayscale;
	bool createAbstract;

	// ECHO metadata files (metadata information of the ECHO archive)
	string AVDocumentFilename;
	string videoLevelFilename;
	string VDXFilename;

};

#endif
