/*
   This file is part of the MoCA video library.
   Name: ColorSubsamplingFilter.cpp
   Copyright (C) 2005 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/




#include<ColorSubsamplingFilter.h>


/*!
  \param src source image
  \param dest destination image with reduced color depth
 */
void
ColorSubsamplingFilter::
process(const Image<uint8> &src, Image<uint8> &dest) 
{
	// reduce color depth to grayscale
	GrayscaleFilter<uint8> gf;
	gf.process (src, dest);

	switch (mode) {
	case ADAPTIVE_COLOR_SUBSAMPLING:
		subsample_adaptive (dest);
		break;
	case LINEAR_COLOR_SUBSAMPLING:
		subsample_linear (dest);
		break;
	}
	return;
}



void
ColorSubsamplingFilter::
subsample_adaptive (Image<uint8> &dest) 
{
	// reduce color depth 
	ImageSize s = dest.getSize();
	int32 reducedColors = numberOfColors;
	if (numberOfColors==2) reducedColors=16;

	// analyse histogram
	int8 msbits = 8;
	bool grayscale=true;
	bool smoothing=false;
	uint8 areas=1;
	Signal<uint32> hist;
	HistogramFeature hf(msbits, grayscale, smoothing, areas);
	hf.process(dest,hist);


// debug
reducedColors=14;

#if 0
	// define intervals
	SignalSize intervalSize (reducedColors);
	Signal<ieee64> interval (intervalSize);
	interval.clear(0);
	for (int32 i=0; i<intervalSize[DIM_X]; i++) {
		interval(i) = (1+i) * s[DIM_X]*s[DIM_Y] / ((ieee64)reducedColors);
	} // for
#endif

#if 0
	const ieee64 avgLuminance=0.5;
	ieee64 intervalWidth  = s[DIM_X]*s[DIM_Y] / ((ieee64)reducedColors);
	ieee64 intervalOffset = (avgLuminance - 1.0) * intervalWidth;
	ieee64 intervalStep   = intervalOffset / reducedColors;
	for (int32 i=0; i<intervalSize[DIM_X]; i++) {
		ieee64 tmp = interval(i);
		interval(i) += intervalOffset;
		intervalOffset -= intervalStep;
		tmp = interval(i);
	} // for
#endif
	
	// define mapping for colors 
	SignalSize mapRegionSize (hist.getSize()[DIM_X]);
	Signal<int32> mapRegions (mapRegionSize);
	mapRegions.clear(0);
	ieee64 th = s[DIM_X]*s[DIM_Y] / ((ieee64)reducedColors);
	int32 cnt=0;

	ieee64 offset=1.0;
	ieee64 offsetStep = 0.0;
#if 0
	// create an image with fewer dark areas
	offset=0.8;
	offsetStep = 2.0 * (1.0-offset) / mapRegionSize[DIM_X];
#endif
	int32 intervalPos=0;
	for (int32 i=0; i<mapRegionSize[DIM_X]; i++) {
		//if (interval(intervalPos) < cnt) intervalPos++;
		//if (intervalPos>=reducedColors) intervalPos=reducedColors-1;
		//mapRegions(i) = intervalPos;

		int v = (int)(offset * cnt / th + 0.01);
		offset +=offsetStep;
		if (v<0) v=0;
		if (v>=reducedColors) v=reducedColors-1;
		mapRegions(i) = v;

		cnt+=hist(i);
	} // for

// debug
for (int32 i=0; i<mapRegionSize[DIM_X]; i++) {
	mapRegions(i) +=2;
} // for
reducedColors=16;
CannyFilter<uint8> cf (3, 15, 50);
Image<uint8> canny;
cf.process (dest, canny);
WriteImage (canny, "I:/temp/b.png");
//exit(1);
#if 0
	// debug: modify average luminance
	for (int32 i=0; i<mapRegionSize[DIM_X]; i++) {
		if (mapRegions(i) ==1) mapRegions(i)--;
		if (mapRegions(i) ==2) mapRegions(i)++;
		if (mapRegions(i) ==5) mapRegions(i)--;
		//if (mapRegions(i) ==7) mapRegions(i)++;
		if (mapRegions(i) ==8) mapRegions(i)++;
		//if  (mapRegions(i) ==5) mapRegions(i)++;
		//if  (mapRegions(i) ==3) mapRegions(i)++;
	} // for
#endif

	// map colors
	int32 intervallSize = 256 / reducedColors;
	for (int32 y=0; y<s[DIM_Y]; y++) {
		for (int32 x=0; x<s[DIM_X]; x++) {
			ieee64 v = 256.0 / reducedColors * mapRegions(dest(x,y,0)) + intervallSize/2;
// debug
if (canny(x,y) == 255) v-=4*intervallSize;

			if (v<0) v=0;
			if (v>255) v=255;
			dest(x,y,0) = (uint8)v;
		} // for
	} // for


	// create a texture image for binary images
	if (numberOfColors==2) {
		createBinary(dest);
	} // if

	return;
} // process

void
ColorSubsamplingFilter::
subsample_linear (Image<uint8> &dest) 
{
	int32 intervallSize = 256 / numberOfColors;
	ImageSize  s = dest.getSize();
	for (int32 y=0; y<s[DIM_Y]; y++) {
		for (int32 x=0; x<s[DIM_X]; x++) {
			int32 v = (dest(x,y) / intervallSize) * intervallSize + intervallSize/2;

			if (v<0) v=0;
			if (v>255) v=255;
			dest (x,y) = (uint8) v;
		} // for
	} // for

	return;
} // process




/*!
  \param img  source and destination image
 */
void
ColorSubsamplingFilter::
createBinary(Image<uint8> &img)
{
#if 1
	// debug
	ImageSize tempSize (16*35+4,35+4,1);
	Image<uint8> temp (tempSize);
	temp.clear(128);
#endif

	// define 16 texture images
	const int textureModuloSize=31;
	const int textureSize=512;
	ImageSize tSize (textureSize,textureSize,1);
	Image<uint8> textures [16];
	for (int32 i=0; i<16; i++) {
		
		int32 skip1=15-i;
		int32 skip2=i;
		switch (i) {
		case 0: 
			skip1=1;
			skip2=0;
			break;
		case 1: 
			skip1=39;
			skip2=1;
			break;
		case 2: 
			skip1=23;
			skip2=1;
			break;
		case 3: 
			skip1=12;
			skip2=1;
			break;
		case 4: 
			skip1=8;
			skip2=1;
			break;
		case 5: 
			skip1=3;
			skip2=1;
			break;
		case 6: 
			skip1=2;
			skip2=1;
			break;
		case 7: 
			skip1=1;
			skip2=1;
			break;
			
		case 15: 
			skip1=0;
			skip2=1;
			break;
		case 14: 
			skip1=1;
			skip2=39;
			break;
		case 13: 
			skip1=1;
			skip2=23;
			break;
		case 12: 
			skip1=1;
			skip2=12;
			break;
		case 11: 
			skip1=1;
			skip2=8;
			break;
		case 10: 
			skip1=1;
			skip2=6;
			break;
		case 9: 
			skip1=1;
			skip2=3;
			break;
		case 8: 
			skip1=1;
			skip2=2;
			break;
		} // switch
		

		// fill textiue image
		int32 cnt=0;
		int32 lastCnt=cnt;
		bool cnt1Active=true;
		textures[i].resize(tSize);
		textures[i].clear(255);
		for (int32 y=0; y<textureSize; y++) {
			cnt=lastCnt;
			for (int32 x=0; x<textureSize; x++) {
				cnt++;
				if (x<textureModuloSize) lastCnt=cnt;
				if (cnt % (skip1+skip2) <skip1) {
					textures[i](x,y) = 0;
				} // if
			} // for
		} // for
		
#if 1
		// debug
		temp.copy (textures[i], SampleBlockPosition(4+i*33, 4, 0, 0));
#endif
	} // for

#if 1
	// debug
	WriteImage (temp, "D:/temp/out3.png");
#endif
	
	
	// replace textures in original image
	ImageSize s = img.getSize();
	for (int32 y=0; y<s[DIM_Y]; y++) {
		for (int32 x=0; x<s[DIM_X]; x++) {
			int32 v = (img(x,y,0) / 16);
			img(x,y,0) = textures[v](x%textureSize,y%textureSize);
		} // for
	} // for
	
	
#if 1
	// debug
	WriteImage (img, "D:/temp/out4.png");
#endif

	return;
}



