/*
    MarkovFilter.cpp     apply Markov filter
    Copyright (C) 2004 Dirk Farin

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <MarkovFilter.h>



void
MarkovFilter::process(const Image<uint8> &lumi, const Image<uint8> &diff, Image<uint8> &dest)
{

	ImageSize size = lumi.getSize();
	dest.resize(size);
	dest.clear(0);
  
	int32 w = size[DIM_X];
	int32 h = size[DIM_Y];
	
	Image<uint8> markov (size);
	Image<uint8> busy (size);
	busy.clear(0);

	// read busy
	for (int32 y=1;y<h-1;y++) {

		uint8* busyPtr = busy.getSamplePtr (1, y);
		for (int32 x=1;x<w-1;x++) {
			*busyPtr = max( max(abs(lumi(x,y)-lumi(x,y-1)), abs(lumi(x,y)-lumi(x,y+1))),
						    max(abs(lumi(x,y)-lumi(x-1,y)), abs(lumi(x,y)-lumi(x+1,y))));
			busyPtr++;
//			busy(x,y) = max( max(abs(lumi(x,y)-lumi(x,y-1)), abs(lumi(x,y)-lumi(x,y+1))),
//							 max(abs(lumi(x,y)-lumi(x-1,y)), abs(lumi(x,y)-lumi(x+1,y))));
		} // for
	} // for


	// read prob
	Image<ieee64> prob (size);
	for (int32 y=0;y<h;y++) {

		ieee64* probPtr = prob.getSamplePtr (0, y);
		const uint8* diffPtr = diff.getSamplePtr (0, y);
		uint8* busyPtr = busy.getSamplePtr (0, y);
		for (int32 x=0;x<w;x++) {
//			prob(x,y) = 1.0 - ProbBack(diff(x,y), busy(x,y));
			*probPtr = 1.0 - ProbBack(*diffPtr, *busyPtr);
			probPtr++;
			diffPtr++;
			busyPtr++;
		} // for
	} // for


#if 0
	// debug	
	Image<uint8> disp(size);	
	string debugPath = "V:/object_recognition/";
	string s = debugPath + "probobj.pgm";
	for (int32 y=0;y<h;y++) {
		for (int32 x=0;x<w;x++) {
			disp(x,y) = (uint8)(100*prob(x,y));
		} // for
	} // for
	WriteImage (disp, s);


	cerr << "markov: debug" << endl;
	s = debugPath + "busy.pgm";	  
	WriteImage (busy, s);
#endif

//	Image<uint8> disp = markov;


	// init markov
	for (int32 y=0;y<h;y++) {

		ieee64* probPtr = prob.getSamplePtr (0, y);
		uint8*  markovPtr = markov.getSamplePtr (0, y);
		for (int32 x=0;x<w;x++) {
			if (*probPtr>0.5) {
//			if (prob(x,y)>0.5) {
//				markov(x,y) = 0;
				*markovPtr = 0;
			} else {
//				markov(x,y) = 1;
				*markovPtr = 1;
			} // else
			probPtr++;
			markovPtr++;
			//markov(x,y) = (ProbBack(diff(x,y), busy(x,y)) > 0.5) ? 1 : 0;
		} // for
	} // for

	OpeningFilter<uint8> of(3);
	of.process(markov);

#if 0
	// debug	
	s = debugPath + "markov-init.pgm";
	for (int32 y=0;y<h;y++) {
		for (int32 x=0;x<w;x++) {
			disp(x,y) = (uint8)(255-markov(x,y)*255);
		} // for
	} // for
	WriteImage (disp, s);


	s = debugPath + "absdiff.pgm";
	for (int32 y=0;y<h;y++) {
		for (int32 x=0;x<w;x++) {
			int d = 3*abs(diff(x,y)-128);
			if (d<0) d=0;
			if (d>255) d=255;
			disp(x,y) = (uint8)(d);
		} // for
	} // for
	WriteImage (disp, s);
#endif
	

	//double b1 = 1.0; 
	//double b2 = 0.5; 
	double b1 = 0.6; 
	double b2 = 0.4; 
	//double bkg_b1 = 0.5; 
	//double bkg_b2 = 0.5; 
	//double g1 = 1.0;
	//double g1 = 2.0;
	//double f1 = 0.2;
	

	for (int iter=0;iter<20;iter++) {
		Image<uint8> tmp = markov;

		for (int y=1;y<h-1;y++) {

			ieee64* probPtr = prob.getSamplePtr (1, y);
			uint8* markovPtr = markov.getSamplePtr (1, y);
			for (int x=1;x<w-1;x++) {

				//double pbobj = prob (x,y);

				// calc prob of object
				double u=0;
				if(tmp(x-1,y) == 0) u+=b1; else u-=b1;
				if(tmp(x+1,y) == 0) u+=b1; else u-=b1;
				if(tmp(x,y-1) == 0) u+=b1; else u-=b1;
				if(tmp(x,y+1) == 0) u+=b1; else u-=b1;

				if(tmp(x-1,y-1) == 0) u+=b2; else u-=b2;
				if(tmp(x+1,y+1) == 0) u+=b2; else u-=b2;
				if(tmp(x+1,y-1) == 0) u+=b2; else u-=b2;
				if(tmp(x-1,y+1) == 0) u+=b2; else u-=b2;

				//u += 1.0 * pbobj;
				u += 1.0 * (*probPtr);

				bool isObj = true;
				if (u<0) {
					isObj=false;
					u=-u;
				} // if
				u = 1.0 - exp(-u-1.0)/2.0;

				double r = ((double) (rand()%287237))/287237.0;
				
				if (r<u) {
					//if (isObj) markov(x,y) = 0; else markov(x,y) = 1;
					if (isObj) *markovPtr = 0; else *markovPtr = 1;
				} else {
					//if (isObj) markov(x,y) = 1; else markov(x,y) = 0;
					if (isObj) *markovPtr = 1; else *markovPtr = 0;
				} // else

				probPtr++;
				markovPtr++;
			} // for
		} // for

		// smooth every 3rd frame
		if (iter%3==0) {
			of.process(markov);
		} // if

	} // for

	for (int y=0;y<h;y++) {
		uint8* destPtr = dest.getSamplePtr (0, y);
		uint8* markovPtr = markov.getSamplePtr (0, y);
		for (int x=0;x<w;x++) {
			//dest(x,y) = 255-markov(x,y)*255;
			*destPtr = 255- (*markovPtr)*255;
			destPtr++;
			markovPtr++;
		} // for
	} // for


	return;
} // process


