/*
    IWaveAudio.cpp     read WAV files
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <IWaveAudio.h>

// constructor, parameter: WAV-file
IWaveAudio::IWaveAudio (string &pathFilename) 
{
	this->pathFilename = pathFilename;
	fh = NULL;
	bytesRead = -1;
	open ();

} // constructor


void 
IWaveAudio::close ()
{
	// open file
	if (fh != NULL) {
		fclose (fh);
	} // if
	fh = NULL;

	bytesRead = -1;
	return;
} // close



void 
IWaveAudio::open ()
{
	// open file
	if (fh != NULL) {
		fclose (fh);
	} // if
	fh = NULL;
	OpenFile::process(&fh, pathFilename, F_READ);

	// initialize buffer
	uint8 *buf = NULL;
	buf = new uint8 [WAV_HEADER_SIZE];
	if (buf == NULL) {
		cerr << endl << "IWaveAudio: cannot allocate memory for buffer." << endl << endl;
		exit(1);
	} // if

	// read header
	int counter = 0;
	counter = fread(buf,1, WAV_HEADER_SIZE, fh);		
	if (counter != WAV_HEADER_SIZE) {
		cerr << endl << "IWaveAudio: could not read header information from file " << pathFilename << endl << endl;
		exit(1);
	} // if

	// set header
	if (!header.setHeader (buf)) {
		cerr << endl << "IWaveAudio: illegal header information in file " << pathFilename << endl << endl;
		exit(1);		
	} // if

	// valid header parameters
	if ((header.getChannels() != 1) && (header.getChannels() != 2)) {
		cerr << endl << "IWaveAudio: illegal header data, only 1 or two channels supported." << endl << endl;
		exit(1);				
	} // if
	if (header.getBitsPerSample() != 16) {
		cerr << endl << "IWaveAudio: illegal header data, only 16 bits per sample." << endl << endl;
		exit(1);				
	} // if

	delete[] buf;
	bytesRead = 0;
	return;
} // initFile

void 
IWaveAudio::read (Signal<int16> &sig, int32 samples, int32 offsetSamples)
{
	if (bytesRead == -1) {
		cerr << endl << "IWaveAudio: cannot read data, probably file not open." << endl << endl;
		exit(1);				
	} // if
	if (offsetSamples != -1) {

		// reset
		int32 p = getOffset();
		if (offsetSamples < p) {
			reset();
		} // if

		// seek position
		p = header.getBlockAlign() * (offsetSamples - getOffset());
		const int maxBuf = 2*3*1024;
		uint8 bufTmp[maxBuf];
		int32 n;
		while (p>0) {
			n = p;
			if (n>maxBuf) {
				n = maxBuf;
			} // if
			int32 counter = fread(bufTmp,1, n, fh);		
			bytesRead += counter;
			p -= counter;
		} // if
	} // if

	// resize signal
	SignalSize size = getSize();
	int32 samplesLeft = size[DIM_X] - bytesRead / header.getBlockAlign();
	if (samples > 0) {
		size[DIM_X] = samples;
	} // if
	if (size[DIM_X] > samplesLeft) {
		//cerr << endl << "IWaveAudio: read " << samplesLeft << " samples instead of " << size[DIM_X] << endl;
		size[DIM_X] = samplesLeft;
	} // if
	if (size[DIM_X] <= 0) {
		cerr << endl << "IWaveAudio: could not read data." << endl;
		if (samplesLeft == 0) {
			cerr << "            EOF reached, reset stream before reading." << endl;
		} // if
		cerr << endl;
		return;
	} // if

	sig.resize (size);

	// read raw data
	uint8 buf [4];
	int16 *sigPtr0 = sig.getSamplePtr(0,0);
	int16 *sigPtr1 = sig.getSamplePtr(0,1);
	int32 lastBytesRead = bytesRead;
	int32 counter = 0;

	// mono
	if (header.getChannels() == 1) {
		for (int32 i=0; i<size[DIM_X]; i++) {
			counter = 0;		
			counter = fread(buf,1, 2, fh);		
			bytesRead += counter;
			
			*sigPtr0++ = buf[0] + (buf[1] << 8);
		} // for
	} // if

	// stereo
	if (header.getChannels() == 2) {
		// save data
		for (int32 i=0; i<size[DIM_X]; i++) {
			counter = 0;		
			counter = fread(buf,1, 4, fh);		
			bytesRead += counter;

			*sigPtr0++ = buf[0] + (buf[1] << 8);
			*sigPtr1++ = buf[2] + (buf[3] << 8);
		} // for
	} // if

	lastBytesRead += size[DIM_X] * header.getBlockAlign();
	if (lastBytesRead != bytesRead) {
		cerr << endl << "IWaveAudio: cannot read all data to signal." << endl << endl;
		exit(1);
	} // if

	return;
} // write



FILE*
IWaveAudio::getFileHandle (int32 offsetSamples)
{
	close();
	open ();

	// seek offset
	if (offsetSamples != -1) {
		
		// seek position
		int32 p = header.getBlockAlign() * (offsetSamples - getOffset());
		if (p > 0) {
			const int maxBuf = 2*3*1024;
			uint8 bufTmp[maxBuf];
			int32 n;
			while (p>0) {
				n = p;
				if (n>maxBuf) {
					n = maxBuf;
				} // if
				int32 counter = fread(bufTmp,1, n, fh);		
				bytesRead += counter;
				p -= counter;
			} // if
		} // if
	} // if

	return fh;
} // write

