/*
    OMpegVideo.cpp     store MPEG videos (new version: VideoEncoder.cpp)
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#include <OMpegVideo.h>




// constuctor setting filename
OMpegVideo::OMpegVideo (const string &pathFilename, ImageSize &size, ieee64 framerate, ieee64 bitrate, bool createGrayscale, bool cropTo16)
{

#ifdef COMPILER_MSVC
	//  Switch Win32 standard output to binary mode, so that piping the
	//  raw image data works. 
	{
		int result = _setmode( _fileno(stdout), _O_BINARY );
		if (result == -1)
		{
			cerr << "Cannot set standard output to binary mode.\n";
			exit(1);
		}
	}
	//  Switch Win32 standard input to binary mode, so that piping the
	//  raw image data works. 
	{
		int result = _setmode( _fileno(stdin), _O_BINARY );
		if (result == -1)
		{
			cerr << "Cannot set standard input to binary mode.\n";
			exit(1);
		}
	}
#endif


	this->pathFilename = pathFilename;
	this->framerate= framerate;
	this->bitrate = bitrate;
	this->createGrayscale= createGrayscale;
	this->cropTo16= cropTo16;
	cropLeft = 0;
	cropRight = 0;
	cropTop = 0;
	cropBottom = 0;

	fh_out = NULL;
	buffer = NULL;
	bufSize = -1;

	this->size = size;
	error = 0;
#ifdef _WIN32
	cmd = WINDOWS_SAMPEG_EXECUTABLE;
#else
	cmd = LINUX_SAMPEG_EXECUTABLE;
#endif

	open();
} // OMpegVideo::OMpegVideo 


// destructor
OMpegVideo::~OMpegVideo ()
{
	close();
} // OMpegVideo::~OMpegVideo 


void
OMpegVideo::writeData(const unsigned char *buf, int size)
{
	// stream data to sampeg
	int bytesOut = 0;
	int n;
	int out;
	error = 0;
	while (bytesOut < size) {
		
		n = 128;
		if (size - bytesOut < 128) {
			n = size - bytesOut;
		} // if
		
		out = 0;
		out = fwrite(&buf[bytesOut],1,n, fh_out);
		if (out == n) {
			error=0;
		} else {

			if (out == 0) {
				error++;
				if (error == 5) {
					cerr << "OMpegVideo: warning, could not all bytes in buffer, trying again..." << endl;
				} // if
				if (error > 50) {
					cerr << endl << endl;
					cerr << "OMpegVideo::writeData: could not write data." << endl << endl;
					cerr << "executable " << cmd << " probably not available." << endl << endl;
					exit(1);
				} // else
			} // else
		} // if
		bytesOut += out;	
	} // while

	// flush buffer
	out = fflush (fh_out);

	return;
} // OMpegVideo::writeData

	
// write frame (the frame will be croped)
void 
OMpegVideo::write(const Image<uint8> &srcCroped)
{

	// convert from srcCroped to buffer
	int pY = 0;
	int pU = fullResolution;
	int pV = fullResolution + quarterResolution;
	for (int y=0; y<cropedImageSize[DIM_Y]; y++) {
		if (y%2 == 0) {
			for (int x=0; x<cropedImageSize[DIM_X]; x++) {
				
				int r = srcCroped(x+cropLeft,y+cropTop,0);
				int g = srcCroped(x+cropLeft,y+cropTop,1);
				int b = srcCroped(x+cropLeft,y+cropTop,2);
				
				// write YUV
				buffer[pY++] = Clip((( 65*r +  129*g +  24*b)>>8)+16);
				if (x%2 == 0) {
					buffer[pU++] = Clip(((-37*r +  -74*g + 112*b)>>8)+128);
					buffer[pV++] = Clip(((112*r +  -93*g + -18*b)>>8)+128);
				} // if
			} // for			
		} else {
			
			for (int x=0; x<cropedImageSize[DIM_X]; x++) {
				
				int r = srcCroped(x+cropLeft,y+cropTop,0);
				int g = srcCroped(x+cropLeft,y+cropTop,1);
				int b = srcCroped(x+cropLeft,y+cropTop,2);
				
				// write Y
				buffer[pY++] = Clip((( 65*r +  129*g +  24*b)>>8)+16);
			} // for
		} // else
	} // for
		
	
	writeData (buffer, bufSize);

	return;
} // OMpegVideo::write


void OMpegVideo::open()
{
	close();


	/*
sampeg 0.0.1
Usage: sampeg 
	   --format=yuv:SIZEXxSIZEY 
	   --input=- 
	   --output=filename 
	   --bitrate=7K  1.5M
	   --framerate=framerate
	   [--colorkiller] 
*/


	//char buf[250];
	int sizeX = size[DIM_X];
	int sizeY = size[DIM_Y];
	//string sCrop=" ";

	cropLeft   = 0;
	cropRight  = 0;
	cropTop    = 0;
	cropBottom = 0;
	if (cropTo16) {
		int tmp=(sizeX/16);
		tmp*=16;
		cropLeft=(sizeX-tmp)/2;
		cropRight=sizeX-tmp-cropLeft;

		tmp=(sizeY/16);
		tmp*=16;
		cropTop=(sizeY-tmp)/2;
		cropBottom=sizeY-tmp-cropTop;

		tmp = cropLeft + cropRight + cropTop + cropBottom;
		if (tmp==0) {
			cropTo16=false;
		} // if
	} // if 
	cropedImageSize = size;
	if (cropTo16) {
		cropedImageSize[DIM_X] = sizeX-cropLeft-cropRight;
		cropedImageSize[DIM_Y] = sizeY-cropTop-cropBottom;
	} // if
	int32 nPixel = cropedImageSize[DIM_X]*cropedImageSize[DIM_Y];

	// define bitrate
	string bitrateStr = "1.5M";
	if (bitrate < 1000.0) {
		bitrateStr = "1.5M";
		if (nPixel >  90000) bitrateStr="1.5M";
		if (nPixel > 120000) bitrateStr="2.0M";
		if (nPixel > 160000) bitrateStr="3.0M";
		if (nPixel > 200000) bitrateStr="4.0M";
		if (nPixel > 270000) bitrateStr="5.0M";
		if (nPixel > 350000) bitrateStr="6.0M";
	} else {

		int32 origNPixel = size[DIM_X]*size[DIM_Y];
		ieee64 bitrateReduction = nPixel * 1.0 / origNPixel;
		if (bitrateReduction>1.0) bitrateReduction=1.0;
		if (bitrateReduction<0.8) bitrateReduction=0.8;

		int32  bitRateCalcInt = (int32)(100.0 * (bitrateReduction * bitrate /1024.0 / 1024.0));
		ieee64 bitRateCalc    = bitRateCalcInt / 100.0;

		if ((bitRateCalc>0.3) && (bitRateCalc < 50.0)) {
			stringstream oss;
			oss << bitRateCalc << "M";
			bitrateStr = oss.str();
		} // if
	} // else

	// modify filename, if spaces in filename/path exist
	string tmpFilename=pathFilename;
	CreateMocaPath pathCheck (tmpFilename, false);
	pathCheck.validatePath (tmpFilename);


	string sGrayscale = "";
	if (createGrayscale)  sGrayscale=" --colorkiller ";

	stringstream ossCmd;
	ossCmd 
		<< cmd 
		<< " --format=yuv --input=-:" 
		<< cropedImageSize[DIM_X] << "x"<< cropedImageSize[DIM_Y] 
		<< " --output=" << tmpFilename
		<< " --bitrate="<< bitrateStr 
		<< " --framerate=" << framerate 
		//			<< " "<< sCrop 
		<< sGrayscale;
	string sCmd = ossCmd.str();

	 //cerr << "executing command: " << endl << sCmd << endl;
	fh_out = NULL;
#if 1
	OpenProcess::open(&fh_out, sCmd, P_WRITE);
#else
	// store result as file
	cerr << "output to file" << endl;
	string ffn = "V:/test.yuv";
	OpenFile::process(&fh_out, ffn, F_WRITE);
#endif


	// initialize buffer
	fullResolution    = cropedImageSize[DIM_X] * cropedImageSize[DIM_Y];
	quarterResolution = ((1+cropedImageSize[DIM_X])/2) * ((1+cropedImageSize[DIM_Y])/2);
	
	// full resolution (Y), 2 * quarter resolution (U,V)
	int32 tmpBufSize = fullResolution + 2*quarterResolution;
	if (tmpBufSize != bufSize) {
		if (buffer!=NULL) {
			delete[] buffer;
		} // if
		buffer=NULL;
		bufSize=tmpBufSize;
		buffer = new unsigned char [bufSize];
	} // if

	return;	
} // OMpegVideo::open


void 
OMpegVideo::close()
{
	int error = fflush (fh_out);
	if (error != 0) {
		cerr << endl << "OMpegVideo: could not flush buffer" << endl << endl;
	} // if

	if (fh_out != NULL) {
#if 1
		pclose(fh_out);
#else
		fclose(fh_out);
#endif
	} // if
	fh_out = NULL;

	if (buffer!=NULL) {
		delete[] buffer;
	} // if
	buffer=NULL;
	bufSize=-1;

	return;
} // OMpegVideo::close


