/*
    OWaveAudio.cpp     store WAV audio files
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <OWaveAudio.h>

// constructor
OWaveAudio::OWaveAudio (string &pathFilename, uint16 channels, uint32 samplesPerSec, uint16 bitsPerSample, uint32 samples) 
{
	if ((channels != 1) && (channels != 2)) {
		cerr << endl << "OWaveAudio: only one or two channels 1 supported." << endl << endl;
		exit(1);
	} // if
	if (bitsPerSample != 16) {
		cerr << endl << "OWaveAudio: only 16 bit per sample supported." << endl << endl;
		exit(1);
	} // if

	this->pathFilename = pathFilename;
	uint32 dataLength = 2 * samples * channels; // 16bit
	header.setHeader (channels, samplesPerSec, bitsPerSample, dataLength);
} // constructor


void 
OWaveAudio::write (Signal<int16> sig)
{
	ofstream ofs (pathFilename.c_str(), std::ios::out | std::ios::binary); 
	
	// write header
	ofs.write ((const char*)header.getRiffBuffer(), WAV_HEADER_SIZE);
	
	// write data
	uint8 buf [4];
	int32 v;

	// resize
	SignalSize size = sig.getSize();
	int32 samples = header.getDataLength() / (header.getChannels() * 2); // 16bit
	if (size[DIM_X] > samples) {
		size[DIM_X] = samples;
		cerr << "OWaveAudio: size of signal and wave header are different." << endl;
		cerr << "            reducing size of signal to " << size[DIM_X] << " samples." << endl;
	} // if
	int32 fill = 0;
	if (samples > size[DIM_X]) {
		fill = samples - size[DIM_X];
		cerr << "OWaveAudio: size of signal and wave header are different." << endl;
		cerr << "            copy last sample of signal " << fill << " times." << endl;
	} // if
	if (size[DIM_X] <= 0) {
		cerr << endl << "OWaveAudio: size of signal is zero." << endl;
		cerr << "            no data is written." << endl << endl;
		ofs.close();
		return;
	} // if

	int16 *sigPtr0 = sig.getSamplePtr(0,0);
	int16 *sigPtr1 = sig.getSamplePtr(0,1);
	if (header.getChannels() == 1) {
		// save data
		for (int32 i=0; i<size[DIM_X]; i++) {
			v = *sigPtr0++;
			buf[0] = v;
			buf[1] = (v >> 8);
			ofs.write ((const char *)buf, 2);
		} // for

		// fill
		for (int32 i=0; i<fill; i++) {
			ofs.write ((const char *)buf, 2);
		} // for
	} // if
	if (header.getChannels() == 2) {
		// save data
		for (int32 i=0; i<size[DIM_X]; i++) {
			v = *sigPtr0++;
			buf[0] = v;
			buf[1] = (v >> 8);
			v = *sigPtr1++;
			buf[2] = v;
			buf[3] = (v >> 8);
			ofs.write ((const char *)buf, 4);
		} // for

		// fill
		for (int32 i=0; i<fill; i++) {
			ofs.write ((const char *)buf, 4);
		} // for

	} // if
	ofs.close();
	return;
} // write



void 
OWaveAudio::write (FILE *fh_in)
{
	FILE *fh_out=NULL;
	OpenFile::process(&fh_out, pathFilename, F_WRITE, false);
	int counter_out = 0;
	int counter_in = 0;

	// write header
	counter_out = 0;		
	counter_out = fwrite ((char*)header.getRiffBuffer(), 1, WAV_HEADER_SIZE, fh_out);
	if (counter_out != WAV_HEADER_SIZE) {
		cerr << endl << "OWaveAudio: could not write header to file " << pathFilename << endl << endl;
		exit(1);
	} // if
	
	uint8 buf [1024];
	uint32 totalBytes = 0;
	while (!feof(fh_in)) {

		// read data
		counter_in = 0;		
		counter_in = fread(&buf,1, 1024, fh_in);		
		totalBytes += counter_in;

		// if we read more bytes than the length of the data, we reduce the number of input bytes (totalBytes and counter_in)
		if (totalBytes > header.getDataLength()) {
			int32 diff = totalBytes - header.getDataLength();
			counter_in -= diff;
			totalBytes -= diff;
		} // if

		// write bytes
		if (counter_in > 0) {
			counter_out = 0;		
			counter_out = fwrite ((char*)(&buf), 1, counter_in, fh_out);
			if (counter_out != counter_in) {
				cerr << endl << "OwaveAudio: could not write data to file " << pathFilename << endl << endl;
				exit(1);
			} // if
		} // if
		
		if ((counter_in != 1024) || (totalBytes >= header.getDataLength())) {
			break;
		} // if
	} // while

	// fill if not all bytes are written
	int32 n = header.getBlockAlign();
	int32 missingBytes = ((int32)header.getDataLength() - totalBytes) /n;
	counter_in -= n; 
	if (missingBytes > 0) {
		cerr << "OWaveAudio: size of signal and wave header are different." << endl;
		cerr << "            copy last sample of signal " << missingBytes << " times." << endl;		
		for (int32 i=0; i<missingBytes; i++) {
			counter_out = 0;		
			counter_out = fwrite ((char*)(&buf[counter_in]), 1, n, fh_out);
			if (counter_out != n) {
				cerr << endl << "OwaveAudio: could not fill data to file " << pathFilename << endl << endl;
				exit(1);
			} // if
		} // for
	} // if

	//delete[] buf;
	fflush (fh_out);
	fclose (fh_out);
	return;
} // write

