/*
    PpmVideoDecoder.cpp     abstract class to decode videos
    Copyright (C) 2005 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <PpmVideoDecoder.h>


PpmVideoDecoder::PpmVideoDecoder (const string &pathFile)
: VideoDecoderBase (pathFile)
{
	buf=NULL;
	init();
}


PpmVideoDecoder::~PpmVideoDecoder ()
{
	if (buf!=NULL) {
		delete[] buf;
	} // if
	buf=NULL;
}



// read next image from current file position (this should be the correct position) 
bool 
PpmVideoDecoder::getImageData (Image<uint8> &img)
{
	bool error = false;
	img.resize(imgSize);

	
	// read data
	if (binary) {
		
		try {
			// handle raw case
			is->read (buf, bufSize);
		} catch (...) {
			error = true;
		}
		if (is->gcount() != bufSize) {
			error = true;
		} // if

		// copy data
		char* dataP = buf;
		for (int y = 0; y < size[DIM_Y]; y++) {
			for (int x = 0; x < size[DIM_X]; x++) {
				for (int band = 0; band < 3; band++) {
					img (x,y,band) = *dataP;
					dataP++;
				} // for
			} // for
		} // for
		
	} else {
			
		try {
			// handle ASCII case
			for (int y = 0; y < size[DIM_Y]; y++) {
				for (int x = 0; x < size[DIM_X]; x++) {
					for (int band = 0; band < 3; band++) {
						int value;
						*is >> value;
						img (x,y,band) = (uint8)value;
					} // for
				} // for
			} // for
		} catch (...) {
			error = true;
		}
	} // else 

	return error;
}


// read next image 
void 
PpmVideoDecoder::getImage (int32 frame, Image<uint8> &img)
{
	if (frame<0) frame=0;
	if (frame>=size[DIM_Z]) frame=size[DIM_Z]-1;

	// seek image position
	std::ios::pos_type newPos = filePos[frame];
	is->seekg (newPos);
	if (is->fail()) {
		cerr << "Error in PpmVideoDecoder, getImage: Cannot access data in file!" << endl << endl;
		exit(1);
	} // if

	// read image data
	if (getImageData (img)) {
		cerr << "Error in PpmVideoDecoder, getImage: Cannot access data in file!" << endl << endl;
		exit(1);
	} // if
	return;
}


// read video
void
PpmVideoDecoder::getVideo (Video<uint8> &vid)
{
	vid.resize(size);
	Image<uint8> img;
	for (int32 i=0; i<size[DIM_Z]; i++) {
		cerr << "\r" << i << "  " << ends;
		getImage(i,img);	
		vid.setImage(i, img);
	}
	cerr << endl;
	return;
}

// read video
void
PpmVideoDecoder::getVideo (ImageVector<uint8> &vid)
{
	vid.resize(size[DIM_Z]);
	Image<uint8> img;
	for (int32 i=0; i<size[DIM_Z]; i++) {
		cerr << "\r" << i << "  " << ends;
		getImage(i,img);	
		vid.setImage(i, img);
	}
	cerr << endl;
	return;
}



bool
PpmVideoDecoder::getHeader (bool &isBinary, int &width, int &height)
{
	bool error = true;
	try {
		
		// get magic number
		char magicNumber[2];
		magicNumber[0] = getChar (is);
		rmWhiteSpace (is);
		magicNumber[1] = getChar (is);
		rmWhiteSpace (is);
		
		// validate magic number
		if (magicNumber[0]!='P') {
			cerr << "Error in PpmVideoDecoder, getHeader: Wrong Magic Number, no PPM file!" << endl << endl;
			exit(1);
		} // if
		if (magicNumber[1]!='6' && magicNumber[1]!='3') {
			cerr << "Error in PpmVideoDecoder, getHeader: Wrong Magic Number, no PPM file!" << endl << endl;
			exit(1);
		} // if
		isBinary=true;
		if (magicNumber[1]=='3') isBinary=false;
		
		
		// ignore comments
		while (is->peek() == '#') {
			is->ignore(INT_MAX, '\n');
		} // while
		
		// read width and height of image
		width = getInt (is);
		rmWhiteSpace (is);
		height = getInt (is);
		rmWhiteSpace (is);
		
		// ignore comments
		while (is->peek() == '#') {
			is->ignore(INT_MAX, '\n');
		} // while
		
		// read max grey values
		int maxGrey = getInt (is);
		
		
		// read one (!) single whitespace as delimiter!
		char tmp = getChar (is);

		error = false;

	} catch(...) {
		error = true;
	} // catch
	
	return error;
}


// init
void
PpmVideoDecoder::init ()
{
	// open input stream
	open();

	// init variables
	filePos.clear();
	imgSize[DIM_X]=0;
	imgSize[DIM_Y]=0;
	size[DIM_X]=0;
	size[DIM_Y]=0;


	// get header information
	bool isBinary;
	int width;
	int height;
	bool error = getHeader (isBinary, width, height);
	if (error) {
		cerr << "Error in PpmVideoDecoder, init: cannot read header information of PPM file!" << endl << endl;
		exit(1);
	} // if
	imgSize[DIM_X]=width;
	imgSize[DIM_Y]=height;
	imgSize[DIM_BAND]=3;
	size[DIM_X]=width;
	size[DIM_Y]=height;
	size[DIM_Z]=0;
	size[DIM_BAND]=3;
	binary = isBinary;


	// read images
	Image<uint8> img;
	bufSize = width*height*3;
	try {
		while (!error) {
			// store current file position
			std::ios::pos_type currPos = is->tellg();
			filePos.push_back(currPos);

			if (binary) {
				// seek next header
				is->seekg(bufSize, std::ios::cur);
			} else {
				error = getImageData (img);
			} // else
			
			// read next header
			error = getHeader (isBinary, width, height);
			
		} // while
	} catch (...) {}
	size[DIM_Z]=filePos.size();


    // read data in buffer for performance reasons and for Gerald
	if (binary) {
		if (buf==NULL) {
			buf = new char[bufSize];
		} // if
	} // if

	// reopen file
	open();

	return;
}

