/*
    VideoDecoder.cpp     decode a video
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#include <VideoDecoder.h>


// constuctor defines filename of mpeg audio/video file
VideoDecoder::VideoDecoder (const string &pathFile, VideoCodec cMode)
{
	colorMode = cMode;
	fh_in  = NULL;
	images = NULL;
	skipBytes = 0;
	
	init(pathFile, -1);
	initializeCache(DEFAULT_CACHED_VIDEOFRAMES);
} // constuctor 

// constuctor defines filename of mpeg audio/video file, the number of frames 
// to be cached, and the last frame to decode
VideoDecoder::VideoDecoder (const string &pathFile, int cacheSize, int32 lastFrameToDecode, VideoCodec cMode)
{
	colorMode = cMode;
	fh_in  = NULL;
	images = NULL;
	skipBytes = 0;
	
	init(pathFile, lastFrameToDecode);
	initializeCache(cacheSize);
} // constuctor 

// destructor
VideoDecoder::~VideoDecoder ()
{
	closeDecoder();
	closeCache();
}

// --------------------------------------------------

void 
VideoDecoder::init(const string &filename, int32 lastFrame)
{
#ifdef COMPILER_MSVC
	//  Switch Win32 standard input to binary mode, so that piping the
	//  raw image data works. 
	int32 result = _setmode( _fileno(stdin), _O_BINARY );
	if (result == -1) {
		cerr << "Cannot set standard input to binary mode.\n";
		exit(1);
	}
#endif

	// modify filename, if spaces in filename/path exist
	pathFilename = filename;
	CreateMocaPath pathCheck (pathFilename, false);
	pathCheck.validatePath (pathFilename);

	VideoIndex reader (pathFilename, false, true);
	reader.process(index);
	sx = index.videoSize[DIM_X];
	sy = index.videoSize[DIM_Y];

	videodecodingExecutable = reader.getVideoDecoderCmd();
	lastFrameToDecode = lastFrame;
	reachedEof = false;

	lastDecodedFrame = -1;

	return;
}

void
VideoDecoder::initializeCache(int32 cSize)
{
	cacheSize = cSize; 
	oldestCachePosition = 0;

	// close cache
	closeCache();

	// initialize images
	images=NULL;
	images = new Image<uint8> [cacheSize];
	if (images==NULL) {
		cerr << "VideoDecoder: Cannot allocate memory for " << cacheSize << " frames." << endl;
		exit(1);
	} // if 

	// initialize skiped bytes
	skipBytes = 2*((sx+1)/2)*((sy+1)/2);

	// initialize number of bands (grayscale)
	index.videoSize[DIM_BAND] = 3;
	if (colorMode == CODEC_Y) {
		index.videoSize[DIM_BAND] = 1;
	} // if
	if (colorMode == CODEC_YUV8) {
		index.videoSize[DIM_BAND] = 1;
	} // if

	// create images
	ImageSize imgSize (index.videoSize[DIM_X], index.videoSize[DIM_Y], index.videoSize[DIM_BAND]);
	for (int i=0; i<cacheSize; i++) {
		images[i].resize(imgSize);
	} // for

	SignalSize sigSize (cacheSize);
	cache.resize(sigSize);
	for (int i=0; i<cacheSize; i++) {
		cache(i) = -1;
	} // for

	return;
} // initializeCache()

void
VideoDecoder::initializeDecoder()
{
	// close decoder
	closeDecoder();
	oldestCachePosition = 0;
	lastDecodedFrame = -1;
	reachedEof = false;

	// open decoder
	string cmd = videodecodingExecutable + " -nosound -noaspect  -vo stdout  -speed 100 " + pathFilename;
/*	if (lastFrameToDecode != -1) {
		char buf[35];
		sprintf (buf, "%li",lastFrameToDecode);
		//ltoa (lastFrameToDecode, (char*)buf, 10);
		string sFrame = buf;
		cmd = videodecodingExecutable + " -last " + sFrame + " " + pathFilename;
	} // if
*/
	// open process
	bool silent = true;
	OpenProcess::open(&fh_in, cmd, P_READ, silent);

	// seek start position of data
	bool error = true;
	try {
		char buf[2];
		if (fh_in != NULL) {
			
			// read next character
			while (!feof(fh_in)) {
				
				// search "MPLAYER-START-DECODING"
				fread (buf, 1, 1, fh_in); if (buf[0]=='M') {

							fread (buf, 1, 1, fh_in); if (buf[0]=='P') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='L') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='A') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='Y') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='E') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='R') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='-') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='S') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='T') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='A') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='R') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='T') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='-') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='D') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='E') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='C') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='O') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='D') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='I') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='N') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='G') {
								error=false;	
								break;
							}}}}}}}}}}}}}}}}}}}}}
				} // if
			} // while
			
		} // if
	}  catch (...) {
		error=true;
	} // catch

	if (error) {
		cerr << "VideoDecoder::initializeDecoder: unable to initialize video decoder (start code not found)!" << endl;
		exit(1);
	} // if

	return;
} // initializeDecoder

// --------------------------------------------------

void
VideoDecoder::closeCache()
{
	if (images!=NULL) delete[] images;
	images = NULL;

	return;
} // closeCache


void
VideoDecoder::closeDecoder()
{
	try {
		if (fh_in != NULL) {
			int error = pclose(fh_in);
			if (error == -1) {
				cerr << "error while closing process: "<< videodecodingExecutable << endl;
			} // if
		} // if
	} catch (...) {
		cerr << "error while closing process: "<< videodecodingExecutable << endl;
	} // catch
	fh_in = NULL;
	
	return;
} // closeDecoder


// --------------------------------------------------

ImageSize 
VideoDecoder::getImageSize()
{ 
	ImageSize size (index.videoSize[DIM_X], index.videoSize[DIM_Y], index.videoSize[DIM_BAND]);
	return size; 
} // getImageSize

VideoSize 
VideoDecoder::getSize() 
{ 
	VideoSize size = index.videoSize;
	return size; 
} // getSize

ieee64 
VideoDecoder::getFrameRate() 
{ 
	return index.fps; 
} // getFrameRate

int32 
VideoDecoder::getBitRate() 
{ 
	return index.bitrate; 
} // getBitRate


bool 
VideoDecoder::isRgb() {
	return (colorMode == CODEC_RGB);
} // isRGB

bool 
VideoDecoder::isYuv() {
	return (colorMode == CODEC_YUV444);
} // isYUV

bool 
VideoDecoder::isY() {
	return (colorMode == CODEC_Y);
} // isGrayscale

VideoIndexData 
VideoDecoder::getIndex() {
	return index;
} // getIndex

// --------------------------------------------------

// decode next image
void 
VideoDecoder::nextImage (Image<uint8> &img) 
{
	lastDecodedFrame++;

	switch (colorMode){
	case CODEC_YUV444: 
		nextImageYuv (img);
		break;
	case CODEC_Y: 
		nextImageY (img);
		break;

    // convert from YUV444
	case CODEC_RGB: 
		nextImageYuv (img);
		ConvertColorSpace::process(CODEC_YUV444, CODEC_RGB, img);
		break;
	case CODEC_YUV422: 
		nextImageYuv (img);
		ConvertColorSpace::process(CODEC_YUV444, CODEC_YUV422, img);
		break;
	case CODEC_YUV420: 
		nextImageYuv (img);
		ConvertColorSpace::process(CODEC_YUV444, CODEC_YUV420, img);
		break;
	case CODEC_YUV8: 
		nextImageYuv (img);
		ConvertColorSpace::process(CODEC_YUV444, CODEC_YUV8, img);
		break;
	default:
		cerr << "VideoDecoder::nextImage: codec not implemented! " << endl;
		cerr << endl;
		exit(1);
	} // switch

	return;
} // nextImage


void 
VideoDecoder::nextImageY (Image<uint8> &img)
{
	// get Y plane
	uint8 buf;
	ImageSize imgSize = img.getSize();
	for (int32 y=0; y<imgSize[DIM_Y]; y++) {
		for (int32 x=0; x<imgSize[DIM_X]; x++) {
			fread(&buf,1, 1, fh_in);
			img(x,y,0) = buf;
		} // for
	} // for

	// get U and V plane
	for (int32 i=0; i<skipBytes; i++) {
		fread(&buf,1, 1, fh_in);
		// do nothing
	} // for

	return;
} // nextImageY

void 
VideoDecoder::nextImageYuv (Image<uint8> &img)
{
	// get Y plane
	uint8 buf;
	ImageSize imgSize = img.getSize();
	for (int32 y=0; y<imgSize[DIM_Y]; y++) {
		for (int32 x=0; x<imgSize[DIM_X]; x++) {
			fread(&buf,1, 1, fh_in);
			img(x,y,0) = buf;
		} // for
	} // for

	// get U and V plane
	for (int32 y=0; y<(imgSize[DIM_Y]+1)/2; y++) {


		// get U plane
		for (int32 x=0; x<(imgSize[DIM_X]+1)/2; x++) {
			fread(&buf,1, 1, fh_in);
			int32 px = 2*x;
			int32 py = 2*y;
			img(px,  py,  1) = buf;
			img(px+1,py,  1) = buf;
			img(px,  py+1,1) = buf;
			img(px+1,py+1,1) = buf;
		} // for

		// get V plane
		for (int32 x=0; x<(imgSize[DIM_X]+1)/2; x++) {
			fread(&buf,1, 1, fh_in);
			int32 px = 2*x;
			int32 py = 2*y;
			img(px,  py,  2) = buf;
			img(px+1,py,  2) = buf;
			img(px,  py+1,2) = buf;
			img(px+1,py+1,2) = buf;
		} // for
	
	} // for

	return;
} // nextImageYuv



// --------------------------------------------------


// read image
Image<uint8>& 
VideoDecoder::getImage(int32 frameNumber) 
{
	if (frameNumber < 0) {
		cerr << "VideoDecoder: the frame " << frameNumber 
			 << " does not exist, frame number 0 is used instead." << endl;
		frameNumber=0;
	} // if
	if (frameNumber >= index.videoSize[DIM_Z]) {
		cerr << "VideoDecoder: the frame " << frameNumber 
			 << " does not exist, using frame " << index.videoSize[DIM_Z]-1 << " instead." << endl;
		frameNumber=index.videoSize[DIM_Z]-1;
	} // if

	int32 frame = frameNumber;	
		
	// get image in cache
	int32 hit = imageInCache (frame);
	if (hit == -1) {
		cerr << "VideoDecoder: could not decode frame " << frame << endl;
		hit = 0;
	} // if

	return images[hit];
} // getImage

int32
VideoDecoder::cacheHitNo(int32 frame)
{
	int32 hit = -1;
	for (int32 i=0; i<cacheSize; i++) {
		if (cache(i) == frame) {
			hit = i;
			break;
		} // if
	} // for

	return hit;
} // cacheHit

/*
int 
VideoDecoder::clip (int x) 
{ 
	//if (x<0)   cerr << "clip-" << endl;
	//if (x>255) cerr << "clip+" << endl;
	if (x<0) return 0; 
	if (x>255) return 255; 
	return x; 
} // clip
*/

int32
VideoDecoder::imageInCache(int32 frame)
{

	// decoder has not been initialized
	if (fh_in == NULL) {
		initializeDecoder();
	} // if

	// video has been decoded and images are not available in cache
	int32 hit = cacheHitNo(frame);
	if (hit == -1) {
		if (reachedEof) {
			cerr << "VideoDecoder: frame " << frame << " has been decoded and is not avaliable in cache." << endl;
			cerr << "            the decoder is initialized again" << endl << endl;
			initializeDecoder();
		} // if
	} // if

	// read next frame
	hit = cacheHitNo(frame);
	if (hit == -1) {
		while (lastDecodedFrame < frame) {
			if (lastDecodedFrame >= index.videoSize[DIM_Z]) {
				reachedEof = true;
				break;
			} // if

			// read next
			nextImage (images[oldestCachePosition]);

			cache(oldestCachePosition) = lastDecodedFrame;
			oldestCachePosition++;
			oldestCachePosition%=cacheSize;				
		} // while
	} // if

	hit = cacheHitNo(frame);

	return hit;
} // imageInCache

