/*
    VideoIndex.cpp     index for videos (length, resolution, ...)
	Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <VideoIndex.h>



// operator<< VideoIndec
std::ostream& operator<< (ostream& os, const VideoIndexData &vi)
{
	os << endl;
	os << "video parameters" << endl;
	os << "version           : " << vi.version<< endl;
	os << "videoFormat       : " << vi.videoFormat << endl;
	os << "bitrate           : " << vi.bitrate << endl;
	os << "width             : " << vi.width << endl;
	os << "height            : " << vi.height << endl;
	os << "fps               : " << vi.fps << endl;
	os << "videoAspect       : " << vi.videoAspect << endl;
	os << endl;

	if (vi.audio) {
		os << "audio parameters" << endl;
		os << vi.audioData << endl;
	} else {
		os << "no audio available!" << endl << endl;
	} // else

	os << "additional parameters" << endl;
	os << "version           : " << vi.version << endl;
	os << "videoSize[DIM_X]  : " << vi.videoSize[DIM_X] << endl;
	os << "videoSize[DIM_Y]  : " << vi.videoSize[DIM_Y] << endl;
	os << "videoSize[DIM_Z]  : " << vi.videoSize[DIM_Z] << endl;
	os << "videoSize[DIM_B]  : " << vi.videoSize[DIM_BAND] << endl;
	os << endl;

	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const VideoIndexData &vi)
{

	// video parameters
    TagFormat::writeNodeValue(os, "videoFormat",  vi.videoFormat);
    TagFormat::writeNodeValue(os, "bitrate",      vi.bitrate);
    TagFormat::writeNodeValue(os, "width",        vi.width);
    TagFormat::writeNodeValue(os, "height",       vi.height);
    TagFormat::writeNodeValue(os, "fps",          vi.fps);
    TagFormat::writeNodeValue(os, "videoAspect",  vi.videoAspect);

	// audio parameters
    TagFormat::writeNodeValue(os, "audio",   vi.audio);
	os << vi.audioData;

	// additional parameters
    TagFormat::writeNodeValue(os, "filename",     vi.filename);
    TagFormat::writeNodeValue(os, "version",        vi.version);
    TagFormat::writeNodeValue(os, "size[DIM_X]",    vi.videoSize[DIM_X]);
    TagFormat::writeNodeValue(os, "size[DIM_Y]",    vi.videoSize[DIM_Y]);
    TagFormat::writeNodeValue(os, "size[DIM_Z]",    vi.videoSize[DIM_Z]);
    TagFormat::writeNodeValue(os, "size[DIM_BAND]", vi.videoSize[DIM_BAND]);
	os << endl;
	
    return os;
} // operator<< 

ITagStream& operator>> (ITagStream& is, VideoIndexData& vi)
{
    char c;
    bool failFlag = false;
	
	if (is >> c) {
		is.putback(c);
		
		// video parameters
		TagFormat::readNodeValue(is, "videoFormat", vi.videoFormat);
		TagFormat::readNodeValue(is, "bitrate",     vi.bitrate);
		TagFormat::readNodeValue(is, "width",       vi.width);
		TagFormat::readNodeValue(is, "height",      vi.height);
		TagFormat::readNodeValue(is, "fps",         vi.fps);
		TagFormat::readNodeValue(is, "videoAspect", vi.videoAspect);

		// audio parameters
		TagFormat::readNodeValue(is, "audio", vi.audio);
		is >> vi.audioData;
			
		// additional parameters
		TagFormat::readNodeValue(is, "filename",    vi.filename);
		TagFormat::readNodeValue(is, "version", vi.version);
		TagFormat::readNodeValue(is, "size[DIM_X]",    vi.videoSize[DIM_X]);
		TagFormat::readNodeValue(is, "size[DIM_Y]",    vi.videoSize[DIM_Y]);
		TagFormat::readNodeValue(is, "size[DIM_Z]",    vi.videoSize[DIM_Z]);
		TagFormat::readNodeValue(is, "size[DIM_BAND]", vi.videoSize[DIM_BAND]);
	} // if
	if (is.rdstate() & ios::failbit) {
		failFlag = true;
	} // if
	
    if (failFlag) {
		throw MocaException ("VideoIndex : Illegal format of index file!");
	} else {
        is.clear();
    } 
    return is;
} // operator>>




VideoIndex::VideoIndex (const string &pathFilename, bool silent, bool fastOpen, bool debugMsg)
{
	// modify filename, if spaces in filename/path exist
	this->pathFilename = pathFilename;
	this->silent = silent;
	this->fastOpen = fastOpen;
	this->debugMsg  = debugMsg;
	CreateMocaPath pathCheck (this->pathFilename, false);
	pathCheck.validatePath (this->pathFilename);
	
	videodecodingExecutable = getVideoDecoderCmd();
	fh_in = NULL;


#ifdef COMPILER_MSVC
	//  Switch Win32 standard input to binary mode, so that piping the
	//  raw image data works. 
	int result = _setmode( _fileno(stdin), _O_BINARY );
	if (result == -1) {
		cerr << "Cannot set standard input to binary mode.\n";
		exit(1);
	}
#endif

#ifdef CYGWIN
	//  Switch Cygwin standard input to binary mode, so that piping the
	//  raw image data works. 
	int result = setmode( fileno(stdin), O_BINARY );
	if (result == -1) {
		cerr << "Cannot set standard input to binary mode.\n";
		exit(1);
	}
#endif


}


void VideoIndex::close()
{
  if (debugMsg) cerr << "closing file handler..." << ends;
  try {
    if (fh_in != NULL) {
      int error = pclose(fh_in);
      if (error == -1) {
	cerr << "error while closing process: "<< videodecodingExecutable << endl;
      } // if
    } // if
  } catch (...) {
    cerr << "error while closing process: "<< videodecodingExecutable << endl;
  } // catch
  fh_in = NULL;
  if (debugMsg) cerr << "\rclosing file handler...done" << endl;
  return;
}


void VideoIndex::open(string &cmd)
{
  if (debugMsg) cerr << "opening file handler..." << ends;
  
  if (fh_in != NULL) {
    close();
  } // if
  fh_in = NULL;
  
  // open process
  OpenProcess::open(&fh_in, cmd, P_READ, true);
  if (debugMsg) cerr << "\ropening file handler...done" << endl;
  return;	
}


void VideoIndex::process(VideoIndexData &data)
{
	// read data
	ReadMocaData<VideoIndexData> reader (pathFilename, VIDEO_INDEX_FILENAME, true);
	bool dataValid = reader.process (VIDEO_INDEX_VERSION, data); 

	if (!fastOpen) {
		if (dataValid) {
			// check image size (e.g., if the video was replaced)
			VideoIndexData tmpData;
			getVideoParameters (tmpData);
			if ((data.height != tmpData.height) ||  (data.width != tmpData.width)) {
				dataValid = false;
			} // if
		} // if
	} // if

	if (!dataValid) {
		
		// create new index
		if (!silent) {
			cerr << "\rVideoIndex: Creating index file ...          " << endl;
		} // if
		readValues(data);
		
		// save index file
		WriteMocaData<VideoIndexData> writer (pathFilename, VIDEO_INDEX_FILENAME);
		writer.process(data);
		if (!silent) {
			cerr << "\rVideoIndex: Creating index file ... done          " << endl;
		} // if
	} // if

	return;
}



bool VideoIndex::readTextLines (vector<string> &textLines, string &cmd)
{
	ProcessReader reader (cmd);
	reader.convertToText (textLines);
	bool error = false;
	if (textLines.size()<1) error = true;

	if (debugMsg) {
	  cerr << "text data: start" << endl;
	  for (int32 i=0; i<textLines.size(); i++) {
	    if (i<10) cerr << "0";
	    if (i<100) cerr << "0";
	    cerr << i << ": " << textLines[i] << endl;
	  } // for
	  cerr << "text data: end" << endl;
	} // if


#if 0
	textLines.clear();
	bool error = true;
	int count = 0;

	try {
		char buf[2];
		if (fh_in != NULL) {
			error=false;	
			
			// read text lines
			bool eofFound=false;
			string s = "";
			while (!feof(fh_in)) {
				
				count = fread (buf, 1, 1, fh_in);
				if (buf[0]==10 || buf[0]==13) {
					if (s.length()>0) textLines.push_back(s);
					s = "";
				} else {
					s += buf[0];
				} // else
			} // while
			
		}
	}  catch (...) {
		error=true;
	} // catch
#endif
	return error;
}



// get video parameters (not audio!)
void VideoIndex::getVideoParameters (VideoIndexData &data)
{
	// read parameters as text line
	vector<string> textLines;
	bool error=true;	
	try {
		// TODO: it is not required to decode the audio, but 
	        // the parameter -nosound does not give any status information (-identify)
		string cmd = videodecodingExecutable + " -noaspect -vo null -ao null " +
                             "-vc null -speed 100 -nosound -identify -quiet " + pathFilename;
		open(cmd);

		error = readTextLines (textLines, cmd);

	}  catch (...) {
		error=true;
	} // catch
	if (error) {
		cerr << "VideoIndex: could not read video parameters!" << endl << endl;
		exit (1);		
	} // if


	// analyse parameters
	{
		data.audio = false;
		for (int32 i=0; i<textLines.size(); i++) {
			string s = textLines[i];
			if (s.substr(0,3) == "ID_") {
				
				s = s.substr(3);
				if (s.substr(0,6) == "VIDEO_") {
					
					// get video parameters
					s = s.substr(6);
					
					if (s.substr(0,7) == "FORMAT=") {
						data.videoFormat = s.substr(7);
					} else if (s.substr(0,8) == "BITRATE=") {
						s = s.substr(8);
						data.bitrate = atol(s.c_str());
					} else if (s.substr(0,6) == "WIDTH=") {
						s = s.substr(6);
						data.width = atol(s.c_str());
					} else if (s.substr(0,7) == "HEIGHT=") {
						s = s.substr(7);
						data.height = atol(s.c_str());
					} else if (s.substr(0,4) == "FPS=") {
						s = s.substr(4);
						data.fps = atof(s.c_str());
					} else if (s.substr(0,7) == "ASPECT=") {
						s = s.substr(7);
						data.videoAspect = atof(s.c_str());
					} // else 
					
				//} else if (s.substr(0,6) == "AUDIO_") {
				//	
				//	// get audio parameters
				//	s = s.substr(6);
				//	
				//	if (s.substr(0,5) == "RATE=") {
				//		s = s.substr(5);
				//		data.audioData.samplesPerSec = atol(s.c_str()); 
				//	}  else if (s.substr(0,4) == "NCH=") {
				//		s = s.substr(4);
				//		data.audioData.channels = atol(s.c_str());
				//	} // else 
					
				} // if
				
			} // if
		} // for

//		if (audioRate*audioChannels > 0) {
//			data.audio = true;
//		} // if
	}

	// set video size
	data.videoSize[DIM_X] = data.width;
	data.videoSize[DIM_Y] = data.height;
	data.videoSize[DIM_Z] = 0;
	data.videoSize[DIM_BAND] = 3;

	return;
} // getVideoParameters 



void VideoIndex::readValues(VideoIndexData &data)
{
	// set version
	data.clear();
	data.version = VIDEO_INDEX_VERSION;

	// store short filename
	CreateMocaPath mocaPath (pathFilename, false); 
	data.filename = mocaPath.file;

	// mplayer available?
	bool error=true;
	string cmd = "";
	string s = "";
	size_t count=0; 
	try {
		cmd = videodecodingExecutable + " -help";
		open(cmd);

		if (fh_in != NULL) {

			// read values
			char buf[20];
			count = fread(&buf,1,10,fh_in); 
			if (count > 0) {
			    buf[9]='\0';
			    s = buf;
			    if (s.find("MPlayer") != string::npos) {
					error=false;
			    }
			} // if
		} // if
	}  catch (...) {
		error=true;
	} // catch
	if (error) {
		cerr << "VideoIndex: could not open video decoder " << videodecodingExecutable << endl << endl;
		exit (1);		
	} // if
	if (!silent) {
		cerr << "...videodecoder found!" << endl;
	} // if



	// check if video exists and is valid
	bool fileExists =false;
	try
	{
		FileHandling::convertToValidUnixFilename (pathFilename);
		fileExists = FileHandling::fileExists (pathFilename);
	}  catch (...) {
		error=true;
	} // catch

	if (error || !fileExists){
		cerr << "VideoIndex: could not open video " << pathFilename << endl << endl;
		exit (1);		
	} // if



	// get video parameters (not audio!)
	getVideoParameters (data);
	if (!silent) {
		cerr << "...video parameters found!" << endl;
	} // if



	// decode video to get length (number of frames)
	vector<string> textLines;
	if (!silent) {
		cerr << "...analyze frames in video..." << ends;
	} // if
	try {
		
		int32 cnt=0;
		cmd = videodecodingExecutable + " -nosound -noaspect -vo stdout:status -speed 100 " + pathFilename;
		open(cmd);

		error = readTextLines (textLines, cmd);

	}  catch (...) {
		error=true;
	} // catch
	if (error) {
		cerr << endl << endl << "VideoIndex: could not open video decoder " << videodecodingExecutable << endl << endl;
		exit (1);		
	} // if
	if (!silent) {
		cerr << "done!" << endl;
	} // if

	
	// analyse parameters
	int32 numberOfFrames = 0;
	{
		for (int32 i=0; i<textLines.size(); i++) {
			string s = textLines[i];
			if (s.substr(0,8) == "FRAMES: ") {
				
				s = s.substr(8);
				numberOfFrames = atol (s.c_str());
				break;
			} // if
		} // for

	}
	data.videoSize[DIM_Z] = numberOfFrames;


	


/*
	-ao pcm | 	null | 
	-dumpaudio
	-nowaveheader

	-vc dummy -vo null
	-really-quiet
	-ao pcm -aofile -
*/

	// extract audio parameters
	if (!silent) {
		cerr << "...analyse audio parameters..." << ends;
	} // if

	data.audio = true;
	if (data.audio) {
		try {
			error=true;	
			cmd = videodecodingExecutable + " -vc dummy -vo null -ao pcm -aofile - -really-quiet " + pathFilename;
			open(cmd);
			
			// read WAV header, PCM raw data, and WAV header again (updated with length)
			char buf[1024];
			if (fh_in != NULL) {
				error=false;	
				while (!feof(fh_in)) {
					count = fread(&buf,1,1024,fh_in); 
				} // if
			} // if
			
			// copy second (updated) WAV header
			char header [44];
			for (int32 i=0; i<44; i++) {
				int32 bufP = count-44+i;
				if (bufP<0) bufP+=1024;
				header[i] = buf[bufP];
			} // for

			// store audio data
			data.audioData.setHeader((unsigned char *)header, true);
			if ((data.audioData.channels<=0) || 
				(data.audioData.dataLength<=0) || (data.audioData.bitsPerSample<=0)) {
				error=true;
			} // if
			
		}  catch (...) {
			error=true;
		} // catch
	} // if

	if (error) {
		data.audio=false;
		data.audioData.setHeader (0, 0, 0, 0);
	} // if

	if (!silent) {
		if (data.audio) {
			cerr << "done!" << endl;
		} else {
			cerr << "no audio!" << endl;
		} // else
	} // if

	return;
} // readValues

