/*
    XmlFormat.cpp     handle XML files
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <XmlFormat.h>

std::ostream& operator<< (ostream &os, const XmlItem &item)
{
	// print status
	int status = item.status;
	if ((status<0) || (status>9))
		os << "status=" << status << "  ";
	else
		os << "status=" << status << "   ";

	// print tag
	//string s="<"+item.tag+">";
	string s=item.tag;
	while(s.length()<25) s+=" ";
	os << s + " ";

	// print xml value
	os << item.data << endl;

	return os;
}



std::ostream& operator<< (ostream &os, const list<XmlItem> &data)
{
	list<XmlItem>::const_iterator it;

	for (it=data.begin(); it!=data.end(); it++) {
		os << (*it);
	} // for

	return os;
}



XmlFormat::XmlFormat(const string &file, bool compareUpperCase)
{
	// initialize parameters
	xmlValidFormat=false;
	filename = file;
	this->compareUpperCase=compareUpperCase;
	FileHandling::convertToValidFilename (filename);
	xmlData.clear();

	// open file and analyse metadata
	analyseFile();
}

void
XmlFormat::analyseFile() 
{
	bool fileError = false;
	try {
		// open file
		ifstream ifs;
		if (openFile(ifs)) {
			
			// read data
			char c;
			
			if (ifs >> c) {
				ifs.putback(c);
				
				// read data to xmlList
				createList (ifs, xmlData);
				
				if (compareUpperCase) {
					list<XmlItem>::iterator it;
					for (it=xmlData.begin(); it!=xmlData.end(); it++) {
						for (string::size_type k=0; k<(*it).tag.size(); k++)
							(*it).tag[k]=toupper((*it).tag[k]);
						for (string::size_type k=0; k<(*it).data.size(); k++)
							(*it).data[k]=toupper((*it).data[k]);
					} // for
				} // if


				// close file
				ifs.close();
				
				// aggregate xmlList
				aggregateList (xmlData);
			} // if
		} // if
	} catch (...) {
		fileError = true;
	} // catch

	if (xmlData.size() > 5) xmlValidFormat=true;
	if (fileError) xmlValidFormat=false;

	return;
}

bool 
XmlFormat::openFile(ifstream &ifs) 
{
	bool error = true;
	try {
		ifs.open (filename.c_str(), std::ios::in | std::ios::binary); 
		error = !ifs.is_open();
	} catch (MocaException ex) {
		error = true;
	} // catch
	if (error) {
		cerr << "could not open file: " << filename << endl;
		return false;
	} // if
	ifs.unsetf(ios::skipws);

	return true;
} 

void 
XmlFormat::createList (ifstream &ifs, list<XmlItem> &xmlData)
{
	bool eofError=false;
    char c;
    //char c1;
	string s1;
	string s2;
	int8 status; // -1: unknown, 0 : data, 1: startTag, 2 : endTag

	while (!eofError) {
		c  = 'A';
		//c1 = 'A';
		s1 = "";
		s2 = "";
		status = -1;
		
		// find next '<'
		while ((ifs >> c) && (c != '<')) {
			if ((c!=9) && (c!=10) && (c!=13)) s1+=c;
		} // while
		if (c!='<') {
			eofError=true; 
		} else {
			status = 1;
			if (!(ifs>>c)) {
				eofError=true; 
			} else {
				if (c=='/') {
					status = 2;
				} else {
					ifs.putback(c);
				} // else
			} // else
		} // else
		
		// store data
		if (s1.length()>0) {
			XmlItem item;
			item.data   = s1;
			item.status = 0;
			xmlData.push_back (item);
		} // if
		
		// find next '>'
		while ((ifs >> c) && (c != '>')) {
			if ((c!=9) && (c!=10) && (c!=13)) s2+=c;
		} // while
		if (c!='>') {
			eofError=true; 
		} // if
		
		// store tag
		if (s2.length()>0) {
			XmlItem item;
			item.data   = s2;
			item.status = status;
			xmlData.push_back (item);
		} // if
	} // while

	return;
}


void 
XmlFormat::aggregateList (list<XmlItem> &xmlData)
{
	list<XmlItem>::iterator it;
	list<XmlItem>::iterator nextIt1;
	list<XmlItem>::iterator nextIt2;
	for (it=xmlData.begin(); it!=xmlData.end(); it++) {
		nextIt1=it;
		if (nextIt1!=xmlData.end()) nextIt1++;
		nextIt2=nextIt1;
		if (nextIt2!=xmlData.end()) nextIt2++;
#if 0
		XmlItem t1 = (*it);
		XmlItem t2 = t1;
		if (nextIt1 != xmlData.end()) t2 = (*nextIt1);
		XmlItem t3 = t2;
		if (nextIt2 != xmlData.end()) t3 = (*nextIt2);
#endif

		if (nextIt1!=xmlData.end()) {
			if (((*it).status==1) && ((*nextIt1).status==2)) {
				// start and end tag?
				string s1 = (*it).data;
				string s2 = (*nextIt1).data;
				if (s1.compare (s2) == 0) {
					(*it).tag = s1;
					(*it).data = "";
					(*it).status = 4; // valid tag without data
					xmlData.erase (nextIt1);
					if (it!=xmlData.begin()) it--;
				} // if
			} // if
		} // if
		

		if (nextIt2!=xmlData.end()) {

			if (((*it).status==1) && ((*nextIt2).status==2)) {
				// start and end tag?
				string s1 = (*it).data;
				string s2 = (*nextIt2).data;
				if (s1.compare (s2) == 0) {
					(*nextIt1).tag = s1;
					(*nextIt1).status = 3; // valid tag with data
					xmlData.erase (it);
					xmlData.erase (nextIt2);
					it = nextIt1;
					if (it!=xmlData.begin()) it--;
				} // if
			} // if
		} // if
	} // for

	// remove spaces in tags and data fields
	for (it=xmlData.begin(); it!=xmlData.end(); it++) {
		string s = (*it).tag;
		while ((s.length()>0) && (s[0] == ' ')) s = s.substr(1);
		while ((s.length()>0) && (s[s.length()-1] == ' ')) s = s.substr(0, s.length()-1);
		(*it).tag = s;

		s = (*it).data;
		while ((s.length()>0) && (s[0] == ' ')) s = s.substr(1);
		while ((s.length()>0) && (s[s.length()-1] == ' ')) s = s.substr(0, s.length()-1);
		(*it).data = s;
	} // for


	// remove and aggregate empty tags <>
	for (it=xmlData.begin(); it!=xmlData.end(); it++) {
		if ((*it).tag == "") {
			string::size_type p = (*it).data.find_first_of(" ");
			if (p == string::npos) {
				(*it).tag = (*it).data;
				(*it).data = "";
			} else {
				(*it).tag = (*it).data.substr(0,p);
				(*it).data = (*it).data.substr(p+1);
			} // else
		} // if
	} // for



	return;
} // aggregateList

void 
XmlFormat::findData (string &tagName, list<XmlItem> &xmlData, XmlItem &data)
{
	if (compareUpperCase) {
		for (string::size_type k=0; k<tagName.size(); k++)
			tagName[k]=toupper(tagName[k]);
	} // if

	data.data   = "";
	data.status = -1;
	data.tag    = "";
	list<XmlItem>::iterator it;
	for (it=xmlData.begin(); it!=xmlData.end(); it++) {
		string s = (*it).tag;
		if (s.compare(tagName) == 0) {
			data=(*it);
			return;
		} // if
	} // for

	return;
}


void 
XmlFormat::process (list<XmlItem> &data)
{
	data = xmlData;

	return;
}

