/*
    SampleBlockBase.cpp     basic class for all data types in MoCA
    Copyright (C) 2000 Gerald Kuehne

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <SampleBlockBase.h>

SampleBlockBase::
SampleBlockBase (sizeType dataSize,
                 const SampleBlockSize& size, const SampleBlockSize& tileSize,
                 BorderMode borderMode)
  : dataSize         (dataSize),
    size             (size),
    capacity         (size),
    tileSize         (tileSize),
    numberOfTiles    (0, 0, 0, 0),
    borderMode       (borderMode),

    sampleIndexX     (NULL),
    sampleIndexY     (NULL),
    sampleIndexZ     (NULL),
    sampleIndexBand  (NULL),
    sampleOffsetX    (NULL),
    sampleOffsetY    (NULL),
    sampleOffsetZ    (NULL),
    sampleOffsetBand (NULL),

    tileIndexX       (NULL),
    tileIndexY       (NULL),
    tileIndexZ       (NULL),
    tileIndexBand    (NULL),
    tileOffsetX      (0),
    tileOffsetY      (0),
    tileOffsetZ      (0),
    tileOffsetBand   (0),

    data             (NULL)

{
  for (sizeType d = 0; d < DIM_N; d++) {
    sampleIndex[d] = sampleOffset[d]  = NULL;
    tileIndex[d]   = NULL;
    tileOffset[d]  = 0;
  }

  realloc (dataSize, size, tileSize, borderMode);
}

SampleBlockBase::
SampleBlockBase (const SampleBlockBase& other)
  : dataSize         (0),
    size             (0, 0, 0, 0),
    capacity         (0, 0, 0, 0),
    tileSize         (0, 0, 0, 0),
    numberOfTiles    (0, 0, 0, 0),
    borderMode       (BORDER_REPLICATE),

    sampleIndexX     (NULL),
    sampleIndexY     (NULL),
    sampleIndexZ     (NULL),
    sampleIndexBand  (NULL),
    sampleOffsetX    (NULL),
    sampleOffsetY    (NULL),
    sampleOffsetZ    (NULL),
    sampleOffsetBand (NULL),

    tileIndexX       (NULL),
    tileIndexY       (NULL),
    tileIndexZ       (NULL),
    tileIndexBand    (NULL),
    tileOffsetX      (0),
    tileOffsetY      (0),
    tileOffsetZ      (0),
    tileOffsetBand   (0),

    data             (NULL)
{
  realloc (other.dataSize, other.size, other.tileSize, other.borderMode);
  memcpy  (data, other.data,
	   dataSize * capacity[DIM_X] * capacity[DIM_Y] *
           capacity[DIM_Z] * capacity[DIM_BAND]);
}

SampleBlockBase::
~SampleBlockBase ()
{
  remove ();
}


SampleBlockBase&
SampleBlockBase::
operator= (const SampleBlockBase& other)
{
  if (this != &other) {
    realloc (other.dataSize, other.size, other.tileSize, other.borderMode);
    memcpy  (data, other.data,
             dataSize * capacity[DIM_X] * capacity[DIM_Y] *
             capacity[DIM_Z] * capacity[DIM_BAND]);
  }
  return *this;
}


void
SampleBlockBase::
remove ()
{
  if (data) {
    /* clear sample structures (index/offset) */
    delete [] (sampleIndexX - DEFAULT_BORDER_SIZE);
	sizeType d;
    for (d = 1; d < DIM_N; d++) {
      delete [] (sampleIndex[d] - DEFAULT_BORDER_SIZE);
      sampleIndex[d] = NULL;
    }
    sampleIndexX = NULL;
    sampleIndexY = sampleIndexZ = sampleIndexBand = NULL;

    for (d = 0; d < DIM_N; d++) {
      delete [] (sampleOffset[d] - DEFAULT_BORDER_SIZE);
      sampleOffset[d] = NULL;
    }
    sampleOffsetX = sampleOffsetY = sampleOffsetZ = sampleOffsetBand = NULL;

    /* clear tile structures (index) */
    delete [] (tileIndexX);
    for (d = 1; d < DIM_N; d++) {
      delete [] tileIndex[d];
      tileIndex[d] = NULL;
    }
    tileIndexX = NULL;
    tileIndexY = tileIndexZ = tileIndexBand = NULL;

    /* clear sample date */
    delete [] data;
    data = NULL;

    size          = SampleBlockSize (0, 0, 0, 0);
    capacity      = size;
    tileSize      = SampleBlockSize (0, 0, 0, 0);
    numberOfTiles = SampleBlockSize (0, 0, 0, 0);
  }
}

void
SampleBlockBase::
allocateSampleMemory ()
{
  uint32 memoryToAllocate = dataSize;
  for (sizeType d = 0; d < DIM_N; d++) {
    uint32 remainder = size[d] % tileSize[d];
    if (remainder > 0)
      capacity[d] += tileSize[d] - remainder;

    memoryToAllocate *= capacity[d];
  }

  data = new uint8 [memoryToAllocate];
  checkMemory (data);
}

void
SampleBlockBase::
allocateTileStructures ()
{
  sizeType d;
  for (d = 0; d < DIM_N; d++)
    numberOfTiles[d] = size[d] / tileSize[d] + ((size[d] % tileSize[d] > 0) ? 1 : 0);

  /* allocate tileIndex structures */
  tileIndexX  = new uint8* [numberOfTiles[DIM_X]];
  checkMemory (tileIndexX);

  for (d = 1; d < DIM_N; d++) {
    tileIndex[d]     = new offsetType [numberOfTiles[d]];
    checkMemory (tileIndex[d]);
  }
  tileIndexY    = tileIndex[DIM_Y];
  tileIndexZ    = tileIndex[DIM_Z];
  tileIndexBand = tileIndex[DIM_BAND];
}


void
SampleBlockBase::
initializeTileStructures ()
{
  tileOffset[DIM_X] =
    tileSize[DIM_X] * tileSize[DIM_Y] * tileSize[DIM_Z] * tileSize[DIM_BAND];
  tileOffsetX           = tileOffset[DIM_X];
  tileOffset[DIM_Y]    = numberOfTiles[DIM_X] * tileOffset[DIM_X];
  tileOffsetY           = tileOffset[DIM_Y];
  tileOffset[DIM_Z]    = numberOfTiles[DIM_Y] * tileOffset[DIM_Y];
  tileOffsetZ           = tileOffset[DIM_Z];
  tileOffset[DIM_BAND] = numberOfTiles[DIM_Z] * tileOffset[DIM_Z];
  tileOffsetBand        = tileOffset[DIM_BAND];

  tileIndexX[0] = data;
  for (sizeType i = 1; i < numberOfTiles[DIM_X]; i++)
    tileIndexX[i] = tileIndexX[i-1] + tileOffset[DIM_X] * dataSize;

  for (sizeType d = 1; d < DIM_N; d++) {
    tileIndex[d][0] = 0;
    for (sizeType i = 1; i < numberOfTiles[d]; i++)
      tileIndex[d][i] = tileIndex[d][i-1] + tileOffset[d] * dataSize;
  }
}


void
SampleBlockBase::
allocateSampleStructures ()
{
  /* allocate sampleIndex structures */
  sampleIndexX = new uint8* [size[DIM_X] + 2 * DEFAULT_BORDER_SIZE];
  checkMemory (sampleIndexX);
  sampleIndexX += DEFAULT_BORDER_SIZE;

  sizeType d;
  for (d = 1; d < DIM_N; d++) {
    sampleIndex[d]     = new offsetType [size[d] + 2 * DEFAULT_BORDER_SIZE];
    checkMemory (sampleIndex[d]);
    sampleIndex[d]    += DEFAULT_BORDER_SIZE;
  }
  sampleIndexY    = sampleIndex[DIM_Y];
  sampleIndexZ    = sampleIndex[DIM_Z];
  sampleIndexBand = sampleIndex[DIM_BAND];

  /* allocate sampleOffset structures */
  for (d = 0; d < DIM_N; d++) {
    sampleOffset[d]   = new offsetType [size[d] + 2 * DEFAULT_BORDER_SIZE];
    checkMemory (sampleOffset[d]);
    sampleOffset[d]  += DEFAULT_BORDER_SIZE;
  }
  sampleOffsetX    = sampleOffset[DIM_X];
  sampleOffsetY    = sampleOffset[DIM_Y];
  sampleOffsetZ    = sampleOffset[DIM_Z];
  sampleOffsetBand = sampleOffset[DIM_BAND];
}

void
SampleBlockBase::
initializeSampleStructures (BorderMode borderMode)
{
  offsetType delta[DIM_N];
  delta[DIM_X]    = 1;
  delta[DIM_Y]    = tileSize[DIM_X];
  delta[DIM_Z]    = tileSize[DIM_X] * tileSize[DIM_Y];
  delta[DIM_BAND] = tileSize[DIM_X] * tileSize[DIM_Y] * tileSize[DIM_Z];

  /* calculate difference between neighboring sample in different tiles */
  offsetType deltaBorder[DIM_N];
  deltaBorder[DIM_X] = tileOffset[DIM_X] - (tileSize[DIM_X] - 1);
  deltaBorder[DIM_Y] = tileOffset[DIM_Y] - (tileSize[DIM_X] * (tileSize[DIM_Y] - 1));
  deltaBorder[DIM_Z] = tileOffset[DIM_Z] -
    (tileSize[DIM_X] * tileSize[DIM_Y] * (tileSize[DIM_Z] - 1));
  deltaBorder[DIM_BAND] = tileOffset[DIM_BAND] -
    (tileSize[DIM_X] * tileSize[DIM_Y] * tileSize[DIM_BAND] * (tileSize[DIM_BAND]-1));

  /* calculate offsets */
  sizeType d;
  for (d = 0; d < DIM_N; d++) {
    for (sizeType i = 0; i < size[d]-1; i++)
      sampleOffset[d][i] = ( ((i+1) % tileSize[d]) == 0 ) ? deltaBorder[d] : delta[d];
    sampleOffset[d][size[d]-1] = 0;
  }

  /* calculate indices */
  sampleIndexX[0] = data;
  for (sizeType i = 1; i < size[DIM_X]; i++)
    sampleIndexX[i] = sampleIndexX[i-1] + sampleOffset[DIM_X][i-1] * dataSize;

  for (d = 1; d < DIM_N; d++) {
    sampleIndex[d][0] = 0;
    for (sizeType i = 1; i < size[d]; i++)
      sampleIndex[d][i] = sampleIndex[d][i-1] + sampleOffset[d][i-1] * dataSize;
  }

  setBorderMode (borderMode);
}


void
SampleBlockBase::
realloc (sizeType dataSize, const SampleBlockSize& size, const SampleBlockSize& tileSize,
         BorderMode borderMode)
{
  if (this->data         != NULL &&
      this->dataSize     == dataSize &&
      this->size         == size &&
      this->tileSize     == tileSize) {
    if (this->borderMode != borderMode)
      setBorderMode (borderMode);
    return;
  }

  remove ();

  this->dataSize   = dataSize;
  this->size       = size;
  this->capacity   = size;
  this->tileSize   = (tileSize[DIM_X] * tileSize[DIM_Y] * tileSize[DIM_Z] * tileSize[DIM_BAND] > 0) ?
    tileSize : size;

  if (size[DIM_X] * size[DIM_Y] * size[DIM_Z] * size[DIM_BAND] > 0) {
    allocateSampleMemory ();
    allocateTileStructures ();
    initializeTileStructures ();
    allocateSampleStructures ();
    initializeSampleStructures (borderMode);
  }
}

void
SampleBlockBase::
initBorderReplicate ()
{
  for (indexType i = 0; i < DEFAULT_BORDER_SIZE; i++) {
    sampleIndexX[-(1 + i)]        = sampleIndexX[0];
    sampleIndexX[size[DIM_X] + i] = sampleIndexX[size[DIM_X] - 1];
  }

  for (sizeType d = 1; d < DIM_N; d++) {
    for (indexType i = 0; i < DEFAULT_BORDER_SIZE; i++) {
      sampleIndex[d][-(1 + i)]    = sampleIndex[d][0];
      sampleIndex[d][size[d] + i] = sampleIndex[d][size[d] - 1];
    }
  }

}

void
SampleBlockBase::
initBorderReflecting ()
{
  indexType direction = 1;
  indexType position  = size[DIM_X] - 1;
  indexType i;
  for (i = 0; i < DEFAULT_BORDER_SIZE; i++) {
    if (i % size[DIM_X] == 0)
      direction *= -1;
    else
      position += direction;
    sampleIndexX[size[DIM_X] + i] = sampleIndexX[position];
  }

  direction = -1;
  position  = 0;
  for (i = 0; i < DEFAULT_BORDER_SIZE; i++) {
    if (i % size[DIM_X] == 0)
      direction *= -1;
    else
      position += direction;
    sampleIndexX[-(1 + i)] = sampleIndexX[position];
  }


  for (sizeType d = 1; d < DIM_N; d++) {
    direction = 1;
    position  = size[d] - 1;
	indexType i;
    for (i = 0; i < DEFAULT_BORDER_SIZE; i++) {
      if (i % size[d] == 0)
        direction *= -1;
      else
        position += direction;
      sampleIndex[d][size[d] + i] = sampleIndex[d][position];
    }

    direction = -1;
    position  = 0;
    for (i = 0; i < DEFAULT_BORDER_SIZE; i++) {
      if (i % size[d] == 0)
        direction *= -1;
      else
        position += direction;
      sampleIndex[d][-(1 + i)] = sampleIndex[d][position];
    }
  }
}

void
SampleBlockBase::
initBorderPeriodic ()
{
  for (indexType i = 0; i < DEFAULT_BORDER_SIZE; i++) {
    sampleIndexX[-(1 + i)] = sampleIndexX[size[DIM_X]-1 - (i % size[DIM_X])];
    sampleIndexX[size[DIM_X] + i] = sampleIndexX[i % size[DIM_X]];
  }

  for (sizeType d = 1; d < DIM_N; d++) {
    for (indexType i = 0; i < DEFAULT_BORDER_SIZE; i++) {
      sampleIndex[d][-(1 + i)] = sampleIndex[d][size[d]-1 - (i % size[d])];
      sampleIndex[d][size[d] + i] = sampleIndex[d][i % size[d]];
    }
  }

}


void
SampleBlockBase::
setBorderMode (BorderMode borderMode)
{
  this->borderMode = borderMode;

  initBorderReplicate ();

  // set indices
  switch (borderMode) {
  case BORDER_REPLICATE:
    initBorderReplicate ();
    break;
  case BORDER_REFLECTING:
    initBorderReflecting ();
    break;
  case BORDER_PERIODIC:
    initBorderPeriodic ();
    break;
  default:
    checkValue (false, "borderMode");
    break;
  }

  // set offsets
  indexType i;
  for (i = -DEFAULT_BORDER_SIZE; i < 0; i++)
    sampleOffsetX[i] = (sampleIndexX[i+1] - sampleIndexX[i]) / dataSize;
  for (i = -1; i < DEFAULT_BORDER_SIZE - 1; i++)
    sampleOffsetX[size[DIM_X] + i] = (sampleIndexX[size[DIM_X] + i+1] -
      sampleIndexX[size[DIM_X] + i]) / dataSize;
  sampleOffsetX[size[DIM_X] + DEFAULT_BORDER_SIZE - 1] = 0;

  for (sizeType d = 1; d < DIM_N; d++) {
	indexType i;
    for (i = -DEFAULT_BORDER_SIZE; i < 0; i++)
      sampleOffset[d][i] = (sampleIndex[d][i+1] - sampleIndex[d][i]) / dataSize;
    for (i = -1; i < DEFAULT_BORDER_SIZE - 1; i++)
      sampleOffset[d][size[d] + i] = (sampleIndex[d][size[d] + i+1] -
        sampleIndex[d][size[d] + i]) / dataSize;
    sampleOffset[d][size[d] + DEFAULT_BORDER_SIZE - 1] = 0;
  }
}


