/*
   This file is part of the MoCA video library.
   Name: SampleBlockBase.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef SAMPLE_BLOCK_BASE_H
#define SAMPLE_BLOCK_BASE_H

// C++ includes
#include <cstring>

// MoCA includes
#include <MocaTypes.h>
#include <MocaError.h>
#include <SampleBlockSize.h>
#include <MocaMacros.h>

template<class T> class SampleBlock;
template<class T> class SubSampleBlock;

class SampleBlockBase
{
#ifdef COMPILER_MSVC
	friend class SampleBlock;
	friend class SubSampleBlock;
#else
	template<class T> friend class SampleBlock;
	template<class T> friend class SubSampleBlock;
#endif
protected:
    enum { DEFAULT_BORDER_SIZE = 64 };

    // Construction/Destruction
    explicit SampleBlockBase (sizeType dataSize,
                              const SampleBlockSize& size     = SampleBlockSize (0,0,0,0),
                              const SampleBlockSize& tileSize = SampleBlockSize (0,0,0,0),
                              BorderMode borderMode           = BORDER_REPLICATE);
    SampleBlockBase (const SampleBlockBase&);

    virtual ~SampleBlockBase ();

    SampleBlockBase& operator= (const SampleBlockBase&);


    // Interface
    void resize  (const SampleBlockSize& size);
    void realloc (sizeType dataSize,
                  const SampleBlockSize& size, const SampleBlockSize& tileSize,
                  BorderMode borderMode);

    SampleBlockSize getSize () const;
    SampleBlockSize getTileSize () const;
    SampleBlockSize getNumberOfTiles () const;

    BorderMode getBorderMode () const;
    void setBorderMode (BorderMode borderMode);

    void* getSamplePtr (indexType x, indexType y, indexType z, indexType band);
    const void* getSamplePtr (indexType x, indexType y, indexType z, indexType band) const;

    void* getTilePtr   (indexType x, indexType y, indexType z, indexType band) const;

    const offsetType& getSampleOffset     (sizeType dimension, indexType i) const;
    const offsetType& getSampleOffsetX    (indexType i) const;
    const offsetType& getSampleOffsetY    (indexType i) const;
    const offsetType& getSampleOffsetZ    (indexType i) const;
    const offsetType& getSampleOffsetBand (indexType i) const;

    const offsetType& getTileOffset     (sizeType dimension) const;
    const offsetType& getTileOffsetX    () const;
    const offsetType& getTileOffsetY    () const;
    const offsetType& getTileOffsetZ    () const;
    const offsetType& getTileOffsetBand () const;


protected:
    void allocateSampleMemory ();
    void allocateSampleStructures ();
    void allocateTileStructures ();
    void initializeSampleStructures (BorderMode borderMode);
    void initializeTileStructures ();

    void initBorderReplicate  ();
    void initBorderReflecting ();
    void initBorderPeriodic   ();

    void remove ();

    sizeType dataSize;

    SampleBlockSize size;
    SampleBlockSize capacity;
    SampleBlockSize tileSize;
    SampleBlockSize numberOfTiles;

    BorderMode borderMode;

    uint8**     sampleIndexX;
    offsetType* sampleIndexY;
    offsetType* sampleIndexZ;
    offsetType* sampleIndexBand;
    offsetType* sampleIndex[DIM_N];

    offsetType* sampleOffsetX;
    offsetType* sampleOffsetY;
    offsetType* sampleOffsetZ;
    offsetType* sampleOffsetBand;
    offsetType* sampleOffset[DIM_N];

    uint8**     tileIndexX;
    offsetType* tileIndexY;
    offsetType* tileIndexZ;
    offsetType* tileIndexBand;
    offsetType* tileIndex[DIM_N];

    offsetType  tileOffsetX;
    offsetType  tileOffsetY;
    offsetType  tileOffsetZ;
    offsetType  tileOffsetBand;
    offsetType  tileOffset[DIM_N];

    uint8* data;

};


inline
void
SampleBlockBase::
resize (const SampleBlockSize& size)
{
    realloc (dataSize, size, tileSize, borderMode);
}

inline
SampleBlockSize
SampleBlockBase::
getSize () const
{
    return size;
}

inline
SampleBlockSize
SampleBlockBase::
getTileSize () const
{
    return tileSize;
}

inline
SampleBlockSize
SampleBlockBase::
getNumberOfTiles () const
{
    return numberOfTiles;
}

inline
BorderMode
SampleBlockBase::
getBorderMode () const
{
    return borderMode;
}

inline
void*
SampleBlockBase::
getSamplePtr (indexType x, indexType y, indexType z, indexType band)
{
    return sampleIndexBand[band] + sampleIndexX[x] + sampleIndexY[y] + sampleIndexZ[z];
}

inline
const void*
SampleBlockBase::
getSamplePtr (indexType x, indexType y, indexType z, indexType band) const
{
    return sampleIndexBand[band] + sampleIndexX[x] + sampleIndexY[y] + sampleIndexZ[z];
}

inline
void*
SampleBlockBase::
getTilePtr (indexType x, indexType y, indexType z, indexType band) const
{
    return tileIndexBand[band] + tileIndexX[x] + tileIndexY[y] + tileIndexZ[z];
}

inline
const offsetType&
SampleBlockBase::
getSampleOffset (sizeType dimension, indexType i) const
{
    return sampleOffset[dimension][i];
}

inline
const offsetType&
SampleBlockBase::
getSampleOffsetX (indexType i) const
{
    return sampleOffsetX[i];
}

inline
const offsetType&
SampleBlockBase::
getSampleOffsetY (indexType i) const
{
    return sampleOffsetY[i];
}

inline
const offsetType&
SampleBlockBase::
getSampleOffsetZ (indexType i) const
{
    return sampleOffsetZ[i];
}

inline
const offsetType&
SampleBlockBase::
getSampleOffsetBand (indexType i) const
{
    return sampleOffsetBand[i];
}

inline
const offsetType&
SampleBlockBase::
getTileOffset (sizeType dimension) const
{
    return tileOffset[dimension];
}

inline
const offsetType&
SampleBlockBase::
getTileOffsetX () const
{
    return tileOffsetX;
}

inline
const offsetType&
SampleBlockBase::
getTileOffsetY () const
{
    return tileOffsetY;
}

inline
const offsetType&
SampleBlockBase::
getTileOffsetZ () const
{
    return tileOffsetZ;
}

inline
const offsetType&
SampleBlockBase::
getTileOffsetBand () const
{
    return tileOffsetBand;
}


#endif
